@@ -1,6 +1,4 @@
 /*
- * $Id: rlimits.c,v 3.1.2.3 1997/06/01 06:13:15 hzoli Exp $
- *
  * rlimits.c - resource limit builtins
  *
  * This file is part of zsh, the Z shell.
@@ -29,17 +27,17 @@
  *
  */
 
-#include "zsh.h"
+#include "rlimits.mdh"
 #include "rlimits.pro"
 
-#ifdef HAVE_GETRLIMIT
+#if defined(HAVE_GETRLIMIT) && defined(RLIM_INFINITY)
 
 /* Generated rec array containing limits required for the limit builtin.     *
  * They must appear in this array in numerical order of the RLIMIT_* macros. */
 
-#include "rlimits.h"
+# include "rlimits.h"
 
-#if defined(RLIM_T_IS_QUAD_T) || defined(RLIM_T_IS_UNSIGNED)
+# if defined(RLIM_T_IS_QUAD_T) || defined(RLIM_T_IS_UNSIGNED)
 static rlim_t
 zstrtorlimt(const char *s, char **t, int base)
 {
@@ -64,9 +62,9 @@ zstrtorlimt(const char *s, char **t, int base)
 	*t = (char *)s;
     return ret;
 }
-#else
-# define zstrtorlimt(a, b, c)	zstrtol((a), (b), (c))
-#endif
+# else /* !RLIM_T_IS_QUAD_T && !RLIM_T_IS_UNSIGNED */
+#  define zstrtorlimt(a, b, c)	zstrtol((a), (b), (c))
+# endif /* !RLIM_T_IS_QUAD_T && !RLIM_T_IS_UNSIGNED */
 
 /* Display resource limits.  hard indicates whether `hard' or `soft'  *
  * limits should be displayed.  lim specifies the limit, or may be -1 *
@@ -215,7 +213,6 @@ printulimit(int lim, int hard, int head)
     else
 	printf("%ld\n", (long)limit);
 }
-#endif /* HAVE_GETRLIMIT */
 
 /* limit: set or show resource limits.  The variable hard indicates *
  * whether `hard' or `soft' resource limits are being set/shown.    */
@@ -224,11 +221,6 @@ printulimit(int lim, int hard, int head)
 static int
 bin_limit(char *nam, char **argv, char *ops, int func)
 {
-#ifndef HAVE_GETRLIMIT
-    /* limit builtin not appropriate to this system */
-    zwarnnam(nam, "not available on this system", NULL, 0);
-    return 1;
-#else
     char *s;
     int hard, limnum, lim;
     rlim_t val;
@@ -330,7 +322,6 @@ bin_limit(char *nam, char **argv, char *ops, int func)
 	    ret++;
     }
     return ret;
-#endif /* HAVE_GETRLIMIT */
 }
 
 /* unlimit: remove resource limits.  Much of this code is the same as *
@@ -340,11 +331,6 @@ bin_limit(char *nam, char **argv, char *ops, int func)
 static int
 bin_unlimit(char *nam, char **argv, char *ops, int func)
 {
-#ifndef HAVE_GETRLIMIT
-    /* unlimit builtin not appropriate to this system */
-    zwarnnam(nam, "not available on this system", NULL, 0);
-    return 1;
-#else
     int hard, limnum, lim;
     int ret = 0;
     uid_t euid = geteuid();
@@ -400,7 +386,6 @@ bin_unlimit(char *nam, char **argv, char *ops, int func)
 	}
     }
     return ret;
-#endif /* HAVE_GETRLIMIT */
 }
 
 /* ulimit: set or display resource limits */
@@ -409,11 +394,6 @@ bin_unlimit(char *nam, char **argv, char *ops, int func)
 static int
 bin_ulimit(char *name, char **argv, char *ops, int func)
 {
-#ifndef HAVE_GETRLIMIT
-    /* builtin not appropriate */
-    zwarnnam(name, "not available on this system", NULL, 0);
-    return 1;
-#else
     int res, resmask = 0, hard = 0, soft = 0, nres = 0;
     char *options;
 
@@ -578,13 +558,20 @@ bin_ulimit(char *name, char **argv, char *ops, int func)
 	if (resmask & 1)
 	    printulimit(res, hard, nres > 1);
     return 0;
-#endif /* HAVE_GETRLIMIT */
 }
 
-static struct binlist bintab[] = {
-    {"limit", 0, bin_limit, 0, -1, 0, "sh", NULL, 0},
-    {"ulimit", 0, bin_ulimit, 0, -1, 0, NULL, NULL, 0},
-    {"unlimit", 0, bin_unlimit, 0, -1, 0, "hs", NULL, 0},
+#else /* !HAVE_GETRLIMIT || !RLIM_INFINITY */
+
+# define bin_limit   bin_notavail
+# define bin_ulimit  bin_notavail
+# define bin_unlimit bin_notavail
+
+#endif /* !HAVE_GETRLIMIT || !RLIM_INFINITY */
+
+static struct builtin bintab[] = {
+    BUILTIN("limit",   0, bin_limit,   0, -1, 0, "sh", NULL),
+    BUILTIN("ulimit",  0, bin_ulimit,  0, -1, 0, NULL, NULL),
+    BUILTIN("unlimit", 0, bin_unlimit, 0, -1, 0, "hs", NULL),
 };
 
 /**/
