@@ -1,6 +1,4 @@
 /*
- * $Id: cap.c,v 3.1.2.1 1997/05/06 05:56:36 hzoli Exp $
- *
  * cap.c - POSIX.1e (POSIX.6) capability set manipulation
  *
  * This file is part of zsh, the Z shell.
@@ -29,8 +27,7 @@
  *
  */
 
-#include "zsh.h"
-
+#include "cap.mdh"
 #include "cap.pro"
 
 #ifdef HAVE_CAP_INIT
@@ -111,13 +108,6 @@ bin_setcap(char *nam, char **argv, char *ops, int func)
 
 #else /* !HAVE_CAP_INIT */
 
-static int
-bin_notavail(char *nam, char **argv, char *ops, int func)
-{
-    zwarnnam(nam, "not available on this system", NULL, 0);
-    return 1;
-}
-
 # define bin_cap    bin_notavail
 # define bin_getcap bin_notavail
 # define bin_setcap bin_notavail
@@ -126,10 +116,10 @@ bin_notavail(char *nam, char **argv, char *ops, int func)
 
 /* module paraphernalia */
 
-static struct binlist bintab[] = {
-    { "cap",    0, bin_cap,    0,  1, 0, NULL, NULL, 0 },
-    { "getcap", 0, bin_getcap, 1, -1, 0, NULL, NULL, 0 },
-    { "setcap", 0, bin_setcap, 2, -1, 0, NULL, NULL, 0 },
+static struct builtin bintab[] = {
+    BUILTIN("cap",    0, bin_cap,    0,  1, 0, NULL, NULL),
+    BUILTIN("getcap", 0, bin_getcap, 1, -1, 0, NULL, NULL),
+    BUILTIN("setcap", 0, bin_setcap, 2, -1, 0, NULL, NULL),
 };
 
 /**/
