@@ -1,6 +1,4 @@
 /*
- * $Id: clone.c,v 3.1.2.1 1997/06/01 06:13:15 hzoli Exp $
- *
  * clone.c - start a forked instance of the current shell on a new terminal
  *
  * This file is part of zsh, the Z shell.
@@ -38,7 +36,7 @@
  *
  */
 
-#include "zsh.h"
+#include "clone.mdh"
 #include "clone.pro"
 
 /**/
@@ -94,8 +92,8 @@ bin_clone(char *nam, char **args, char *ops, int func)
     return 0;
 }
 
-static struct binlist bintab[] = {
-    { "clone", 0, bin_clone, 1, 1, 0, NULL, NULL, 0 },
+static struct builtin bintab[] = {
+    BUILTIN("clone", 0, bin_clone, 1, 1, 0, NULL, NULL),
 };
 
 /**/
