@@ -1,6 +1,4 @@
 /*
- * $Id: files.c,v 3.1.2.2 1997/04/25 04:53:16 hzoli Exp $
- *
  * files.c - file operation builtins
  *
  * This file is part of zsh, the Z shell.
@@ -29,7 +27,7 @@
  *
  */
 
-#include "zsh.h"
+#include "files.mdh"
 
 typedef int (*MoveFunc) _((char const *, char const *));
 
@@ -292,15 +290,13 @@ domove(char *nam, MoveFunc move, char *p, char *q, int flags)
 		return 0;
 	    doit = 1;
 	} else if((flags & MV_ASKNW) &&
-#ifdef S_ISLNK
 		!S_ISLNK(st.st_mode) &&
-#endif
 		access(qbuf, W_OK)) {
 	    nicezputs(nam, stderr);
 	    fputs(": replace `", stderr);
 	    nicezputs(q, stderr);
-	    fprintf(stderr, "', overriding mode %04lo? ",
-		(unsigned long) (st.st_mode & 07777));
+	    fprintf(stderr, "', overriding mode %04o? ",
+		mode_to_octal(st.st_mode));
 	    fflush(stderr);
 	    if(!ask())
 		return 0;
@@ -403,15 +399,13 @@ dorm(char *nam, char *arg, char *rp, char *ops, struct dirsav *ds, int first)
 	    if(!ask())
 		return 0;
 	} else if(!ops['f'] &&
-#ifdef S_ISLNK
 		!S_ISLNK(st.st_mode) &&
-#endif
 	    	access(rp, W_OK)) {
 	    nicezputs(nam, stderr);
 	    fputs(": remove `", stderr);
 	    nicezputs(arg, stderr);
-	    fprintf(stderr, "', overriding mode %04lo? ",
-		(unsigned long) (st.st_mode & 07777));
+	    fprintf(stderr, "', overriding mode %04o? ",
+		mode_to_octal(st.st_mode));
 	    fflush(stderr);
 	    if(!ask())
 		return 0;
@@ -500,17 +494,19 @@ dormr(char *nam, char *arg, char *rp, char *ops, struct dirsav *ds, int first)
 
 /* module paraphernalia */
 
-static struct binlist bintab[] = {
 #ifdef HAVE_LSTAT
-    { "ln",    0, bin_ln,    1, -1, BIN_LN, "dfis",  NULL, 0 },
+# define LN_OPTS "dfis"
 #else
-    { "ln",    0, bin_ln,    1, -1, BIN_LN, "dfi",   NULL, 0 },
+# define LN_OPTS "dfi"
 #endif
-    { "mkdir", 0, bin_mkdir, 1, -1, 0,      "pm",    NULL, 0 },
-    { "mv",    0, bin_ln,    2, -1, BIN_MV, "fi",    NULL, 0 },
-    { "rm",    0, bin_rm,    1, -1, 0,      "dfirs", NULL, 0 },
-    { "rmdir", 0, bin_rmdir, 1, -1, 0,      NULL,    NULL, 0 },
-    { "sync",  0, bin_sync,  0,  0, 0,      NULL,    NULL, 0 },
+
+static struct builtin bintab[] = {
+    BUILTIN("ln",    0, bin_ln,    1, -1, BIN_LN, LN_OPTS, NULL),
+    BUILTIN("mkdir", 0, bin_mkdir, 1, -1, 0,      "pm",    NULL),
+    BUILTIN("mv",    0, bin_ln,    2, -1, BIN_MV, "fi",    NULL),
+    BUILTIN("rm",    0, bin_rm,    1, -1, 0,      "dfirs", NULL),
+    BUILTIN("rmdir", 0, bin_rmdir, 1, -1, 0,      NULL,    NULL),
+    BUILTIN("sync",  0, bin_sync,  0,  0, 0,      NULL,    NULL),
 };
 
 /**/
