@@ -1,6 +1,4 @@
 /*
- * $Id: stat.c,v 3.1.2.2 1997/04/25 04:51:37 hzoli Exp $
- *
  * stat.c - stat builtin interface to system call
  *
  * This file is part of zsh, the Z shell.
@@ -29,8 +27,7 @@
  *
  */
 
-#include "zsh.h"
-
+#include "stat.mdh"
 #include "stat.pro"
 
 enum statnum { ST_DEV, ST_INO, ST_MODE, ST_NLINK, ST_UID, ST_GID,
@@ -53,57 +50,55 @@ statmodeprint(mode_t mode, char *outbuf, int flags)
 	    strcat(outbuf, " (");
     }
     if (flags & STF_STRING) {
-	int ioff, itest;
 	char pm[11] = "?---------";
 
-#ifdef S_ISBLK
 	if (S_ISBLK(mode))
 	    *pm = 'b';
-#endif
-#ifdef S_ISCHR
-	if (S_ISCHR(mode))
+	else if (S_ISCHR(mode))
 	    *pm = 'c';
-#endif
-#ifdef S_ISREG
-	if (S_ISREG(mode))
-	    *pm = '-';
-#endif
-#ifdef S_ISDIR
-	if (S_ISDIR(mode))
+	else if (S_ISDIR(mode))
 	    *pm = 'd';
-#endif
-#ifdef S_ISFIFO
-	if (S_ISFIFO(mode))
+	else if (S_ISFIFO(mode))
 	    *pm = 'p';
-#endif
-#ifdef S_ISLNK
-	if (S_ISLNK(mode))
+	else if (S_ISLNK(mode))
 	    *pm = 'l';
-#endif
-#ifdef S_ISSOCK
-	if (S_ISSOCK(mode))
+	else if (S_ISMPC(mode))
+	    *pm = 'm';
+	else if (S_ISNWK(mode))
+	    *pm = 'n';
+	else if (S_ISOFD(mode))
+	    *pm = 'M';
+	else if (S_ISOFL(mode))
+	    *pm = 'M';
+	else if (S_ISREG(mode))
+	    *pm = '-';
+	else if (S_ISSOCK(mode))
 	    *pm = 's';
-#endif
 
-	/* too much hassle using macros for the following which should be
-	 * standardish.
-	 */
-	ioff = 1;
-	for (itest = 64; itest > 0; itest /= 8) {
-	    if (mode & (4*itest))
-		pm[ioff] = 'r';
-	    if (mode & (2*itest))
-		pm[ioff+1] = 'w';
-	    if (mode & itest)
-		pm[ioff+2] = 'x';
-	    ioff += 3;
-	}
-	if (mode & 04000)	/* setuid */
-	    pm[3] = (pm[3] == 'x') ? 's' : 'S';
-	if (mode & 02000)	/* setgid */
-	    pm[6] = (pm[6] == 'x') ? 's' : 'S';
-	if (mode & 01000)	/* sticky bit */
-	    pm[9] = (pm[9] == 'x') ? 't' : 'T';
+	if(mode & S_IRUSR)
+	    pm[1] = 'r';
+	if(mode & S_IWUSR)
+	    pm[2] = 'w';
+	if(mode & S_IXUSR)
+	    pm[3] = 'x';
+	if(mode & S_IRGRP)
+	    pm[4] = 'r';
+	if(mode & S_IWGRP)
+	    pm[5] = 'w';
+	if(mode & S_IXGRP)
+	    pm[6] = 'x';
+	if(mode & S_IROTH)
+	    pm[7] = 'r';
+	if(mode & S_IWOTH)
+	    pm[8] = 'w';
+	if(mode & S_IXOTH)
+	    pm[9] = 'x';
+	if (mode & S_ISUID)
+	    pm[3] = (mode & S_IXUSR) ? 's' : 'S';
+	if (mode & S_ISGID)
+	    pm[6] = (mode & S_IXGRP) ? 's' : 'S';
+	if (mode & S_ISVTX)
+	    pm[9] = (mode & S_IXOTH) ? 't' : 'T';
 
 	strcat(outbuf, pm);
 	if (flags & STF_RAW)
@@ -183,14 +178,13 @@ static void
 statlinkprint(struct stat *sbuf, char *outbuf, char *fname)
 {
     int num;
-#ifdef S_ISLNK
+
     /* fname is NULL if we are looking at an fd */
     if (fname && S_ISLNK(sbuf->st_mode) &&
  	(num = readlink(fname, outbuf, PATH_MAX)) > 0) {
 	/* readlink doesn't terminate the buffer itself */
 	outbuf[num] = '\0';
     }
-#endif
 }
 
 
@@ -515,11 +509,15 @@ bin_stat(char *name, char **args, char *ops, int func)
     return ret;
 }
 
+static struct builtin bintab[] = {
+    BUILTIN("stat", 0, bin_stat, 0, -1, 0, NULL, NULL),
+};
+
 /**/
 int
 boot_stat(Module m)
 {
-    return addbuiltin("stat", 0, bin_stat, 0, -1, 0, NULL, NULL);
+    return !addbuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab));
 }
 
 #ifdef MODULE
@@ -528,7 +526,8 @@ boot_stat(Module m)
 int
 cleanup_stat(Module m)
 {
-    return deletebuiltin("stat");
+    deletebuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab));
+    return 0;
 }
 
 #endif
