@@ -1,6 +1,4 @@
 /*
- * $Id: compctl.c,v 3.1.2.3 1997/06/01 06:13:15 hzoli Exp $
- *
  * compctl.c - the compctl builtin
  *
  * This file is part of zsh, the Z shell.
@@ -29,8 +27,7 @@
  *
  */
 
-#define IN_COMP
-#include "comp.h"
+#include "compctl.mdh"
 #include "compctl.pro"
 
 #define COMP_LIST	(1<<0)	/* -L */
@@ -176,6 +173,9 @@ get_compctl(char *name, char ***av, Compctl cc, int first, int isdef)
 	    case 'Q':
 		cct.mask |= CC_QUOTEFLAG;
 		break;
+	    case '/':
+		cct.mask |= CC_DIRS;
+		break;
 	    case 'k':
 		if ((*argv)[1]) {
 		    cct.keyvar = (*argv) + 1;
@@ -202,7 +202,12 @@ get_compctl(char *name, char ***av, Compctl cc, int first, int isdef)
 		    *argv = "" - 1;
 		}
 		break;
+	    case 'Y':
+		cct.mask |= CC_EXPANDEXPL;
+		goto expl;
 	    case 'X':
+		cct.mask &= ~CC_EXPANDEXPL;
+	    expl:
 		if ((*argv)[1]) {
 		    cct.explain = (*argv) + 1;
 		    *argv = "" - 1;
@@ -214,6 +219,18 @@ get_compctl(char *name, char ***av, Compctl cc, int first, int isdef)
 		    *argv = "" - 1;
 		}
 		break;
+	    case 'y':
+		if ((*argv)[1]) {
+		    cct.ylist = (*argv) + 1;
+		    *argv = "" - 1;
+		} else if (!argv[1]) {
+		    zwarnnam(name, "function/variable expected after -%c",
+			     NULL, **argv);
+		} else {
+		    cct.ylist = *++argv;
+		    *argv = "" - 1;
+		}
+		break;
 	    case 'P':
 		if ((*argv)[1]) {
 		    cct.prefix = (*argv) + 1;
@@ -277,6 +294,19 @@ get_compctl(char *name, char ***av, Compctl cc, int first, int isdef)
 		    *argv = "" - 1;
 		}
 		break;
+	    case 'W':
+		if ((*argv)[1]) {
+		    cct.withd = (*argv) + 1;
+		    *argv = "" - 1;
+		} else if (!argv[1]) {
+		    zwarnnam(name, "path expected after -%c", NULL,
+			    **argv);
+		    return 1;
+		} else {
+		    cct.withd = *++argv;
+		    *argv = "" - 1;
+		}
+		break;
 	    case 'H':
 		if ((*argv)[1])
 		    cct.hnum = atoi((*argv) + 1);
@@ -671,7 +701,8 @@ cc_assign(char *name, Compctl *ccptr, Compctl cct, int reass)
     Compctl cc;
 
     if (cct->subcmd && (cct->keyvar || cct->glob || cct->str ||
-			cct->func || cct->explain || cct->prefix)) {
+			cct->func || cct->explain || cct->ylist ||
+			cct->prefix)) {
 	zwarnnam(name, "illegal combination of options", NULL, 0);
 	return 1;
     }
@@ -709,9 +740,11 @@ cc_assign(char *name, Compctl *ccptr, Compctl cct, int reass)
     zsfree(cc->str);
     zsfree(cc->func);
     zsfree(cc->explain);
+    zsfree(cc->ylist);
     zsfree(cc->prefix);
     zsfree(cc->suffix);
     zsfree(cc->subcmd);
+    zsfree(cc->withd);
     zsfree(cc->hpat);
     
     /* and copy over the new stuff, (permanently) allocating
@@ -723,9 +756,11 @@ cc_assign(char *name, Compctl *ccptr, Compctl cct, int reass)
     cc->str = ztrdup(cct->str);
     cc->func = ztrdup(cct->func);
     cc->explain = ztrdup(cct->explain);
+    cc->ylist = ztrdup(cct->ylist);
     cc->prefix = ztrdup(cct->prefix);
     cc->suffix = ztrdup(cct->suffix);
     cc->subcmd = ztrdup(cct->subcmd);
+    cc->withd = ztrdup(cct->withd);
     cc->hpat = ztrdup(cct->hpat);
     cc->hnum = cct->hnum;
 
@@ -787,7 +822,7 @@ static void
 printcompctl(char *s, Compctl cc, int printflags)
 {
     Compctl cc2;
-    char *css = "fcqovbAIFpEjrzBRGudeNOZUnQmw";
+    char *css = "fcqovbAIFpEjrzBRGudeNOZUnQmw/";
     char *mss = " pcCwWsSnNmrR";
     unsigned long t = 0x7fffffff;
     unsigned long flags = cc->mask;
@@ -838,12 +873,14 @@ printcompctl(char *s, Compctl cc, int printflags)
     flags = cc->mask;
     printif(cc->keyvar, 'k');
     printif(cc->func, 'K');
-    printif(cc->explain, 'X');
+    printif(cc->explain, (cc->mask & CC_EXPANDEXPL) ? 'Y' : 'X');
+    printif(cc->ylist, 'y');
     printif(cc->prefix, 'P');
     printif(cc->suffix, 'S');
     printif(cc->glob, 'g');
     printif(cc->str, 's');
     printif(cc->subcmd, 'l');
+    printif(cc->withd, 'W');
     if (cc->hpat) {
 	printf(" -H %d ", cc->hnum);
 	quotedzputs(cc->hpat, stdout);
@@ -1021,8 +1058,8 @@ bin_compctl(char *name, char **argv, char *ops, int func)
     return ret;
 }
 
-static struct binlist bintab[] = {
-    { "compctl", 0, bin_compctl, 0, -1, 0, NULL, NULL, 0 },
+static struct builtin bintab[] = {
+    BUILTIN("compctl", 0, bin_compctl, 0, -1, 0, NULL, NULL),
 };
 
 /**/
