@@ -1,6 +1,4 @@
 /*
- * $Id: zle_main.c,v 3.1.2.13 1997/06/01 06:13:15 hzoli Exp $
- *
  * zle_main.c - main routines for line editor
  *
  * This file is part of zsh, the Z shell.
@@ -29,15 +27,87 @@
  *
  */
 
-#define ZLEGLOBALS
-#define IN_ZLE
-#define IN_COMP
-#include "zle.h"
+#include "zle.mdh"
 #include "zle_main.pro"
 
+/* != 0 if we're done editing */
+
+/**/
+int done;
+
+/* location of mark */
+
+/**/
+int mark;
+
+/* last character pressed */
+
+/**/
+int c;
+
+/* the binding for this key */
+
+/**/
+Thingy bindk;
+
+/* insert mode/overwrite mode flag */
+
+/**/
+int insmode;
+
 static int eofchar, eofsent;
 static long keytimeout;
 
+#ifdef HAVE_SELECT
+/* Terminal baud rate */
+
+static int baud;
+#endif
+
+/* flags associated with last command */
+
+/**/
+int lastcmd;
+
+/* the status line, and its length */
+
+/**/
+char *statusline;
+/**/
+int statusll;
+
+/* The current history line and cursor position for the top line *
+ * on the buffer stack.                                          */
+
+/**/
+int stackhist, stackcs;
+
+/* != 0 if we are making undo records */
+
+/**/
+int undoing;
+
+/* current modifier status */
+
+/**/
+struct modifier zmod;
+
+/* Current command prefix status.  This is normally 0.  Prefixes set *
+ * this to 1.  Each time round the main loop, this is checked: if it *
+ * is 0, the modifier status is reset; if it is 1, the modifier      *
+ * status is left unchanged, and this flag is reset to 0.  The       *
+ * effect is that several prefix commands can be executed, and have  *
+ * cumulative effect, but any other command execution will clear the *
+ * modifiers.                                                        */
+
+/**/
+int prefixflag;
+
+/* != 0 if there is a pending beep (usually indicating an error) */
+
+/**/
+int feepflag;
+
 /* set up terminal */
 
 /**/
@@ -345,7 +415,7 @@ zleread(char *lp, char *rp, int ha)
 	char *pptbuf;
 	int pptlen;
 
-	pptbuf = putprompt(lp, &pptlen, NULL, 1);
+	pptbuf = unmetafy(promptexpand(lp, 0, NULL, NULL), &pptlen);
 	write(2, (WRITE_ARG_2_T)pptbuf, pptlen);
 	free(pptbuf);
 	return (unsigned char *)shingetline();
@@ -371,8 +441,10 @@ zleread(char *lp, char *rp, int ha)
     insmode = unset(OVERSTRIKE);
     eofsent = 0;
     resetneeded = 0;
-    lpmpt = lp;
-    rpmpt = rp;
+    lpptbuf = promptexpand(lp, 1, NULL, NULL);
+    pmpt_attr = txtchange;
+    rpptbuf = promptexpand(rp, 1, NULL, NULL);
+    rpmpt_attr = txtchange;
     histallowed = ha;
     PERMALLOC {
 	histline = curhist;
@@ -383,9 +455,11 @@ zleread(char *lp, char *rp, int ha)
 	line = (unsigned char *)zalloc((linesz = 256) + 2);
 	virangeflag = lastcmd = done = cs = ll = mark = 0;
 	curhistline = NULL;
-	addedsuffix = complexpect = vichgflag = 0;
+	vichgflag = 0;
 	viinsbegin = 0;
 	statusline = NULL;
+	selectkeymap("main", 1);
+	fixsuffix();
 	if ((s = (unsigned char *)getlinknode(bufstack))) {
 	    setline((char *)s);
 	    zsfree((char *)s);
@@ -412,7 +486,6 @@ zleread(char *lp, char *rp, int ha)
 	initmodifier(&zmod);
 	prefixflag = 0;
 	feepflag = 0;
-	selectkeymap("main", 1);
 	refresh();
 	while (!done && !errflag) {
 
@@ -454,6 +527,8 @@ zleread(char *lp, char *rp, int ha)
 	statusline = NULL;
 	invalidatelist();
 	trashzle();
+	free(lpptbuf);
+	free(rpptbuf);
 	zleactive = 0;
 	alarm(0);
     } LASTALLOC;
@@ -494,8 +569,8 @@ execzlefunc(Thingy func)
 	if(!(wflags & ZLE_KEEPSUFFIX))
 	    removesuffix();
 	if(!(wflags & ZLE_MENUCMP)) {
+	    fixsuffix();
 	    invalidatelist();
-	    addedsuffix = complexpect = 0;
 	}
 	if (wflags & ZLE_LINEMOVE)
 	    vilinerange = 1;
@@ -525,18 +600,6 @@ execzlefunc(Thingy func)
     }
 }
 
-/**/
-void
-removesuffix(void)
-{
-    if(addedsuffix) {
-	backdel(addedsuffix);
-	addedsuffix = 0;
-	invalidatelist();
-    }
-    complexpect = 0;
-}
-
 /* initialise command modifiers */
 
 /**/
@@ -773,10 +836,10 @@ trashzle(void)
 	kungetct = 0;
 }
 
-static struct binlist bintab[] = {
-    { "bindkey", 0, bin_bindkey, 0, -1, 0, "evaMldDANmrsLR",   NULL, 0 },
-    { "vared",   0, bin_vared,   1,  7, 0, NULL,               NULL, 0 },
-    { "zle",     0, bin_zle,     0, -1, 0, "lDANL",            NULL, 0 },
+static struct builtin bintab[] = {
+    BUILTIN("bindkey", 0, bin_bindkey, 0, -1, 0, "evaMldDANmrsLR", NULL),
+    BUILTIN("vared",   0, bin_vared,   1,  7, 0, NULL,             NULL),
+    BUILTIN("zle",     0, bin_zle,     0, -1, 0, "lDANL",          NULL),
 };
 
 /**/
@@ -795,7 +858,6 @@ boot_zle(Module m)
 
     /* miscellaneous initialisations */
     stackhist = stackcs = -1;
-    lastnamed = refthingy(t_undefinedkey);
     kungetbuf = (char *) zalloc(kungetsz = 32);
 
     /* initialise the keymap system */
@@ -821,7 +883,6 @@ cleanup_zle(Module m)
 
     deletebuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab));
     cleanup_keymaps();
-    unrefthingy(lastnamed);
     deletehashtable(thingytab);
 
     zfree(vichgbuf, vichgbufsz);
