@@ -1,6 +1,4 @@
 /*
- * $Id: zle_refresh.c,v 3.1.2.6 1997/06/02 06:56:09 hzoli Exp $
- *
  * zle_refresh.c - screen update
  *
  * This file is part of zsh, the Z shell.
@@ -29,19 +27,51 @@
  *
  */
 
-#define IN_ZLE
-#define IN_COMP
-#include "zle.h"
+#include "zle.mdh"
 #include "zle_refresh.pro"
 
+/* Expanded prompts */
+
+/**/
+char *lpptbuf, *rpptbuf;
+
+/* Text attributes after displaying prompts */
+
+/**/
+unsigned pmpt_attr, rpmpt_attr;
+
+/* number of lines displayed */
+
+/**/
+int nlnct;
+
+/* Most lines of the buffer we've shown at once with the current list *
+ * showing.  == 0 if there is no list.  == -1 if a new list has just  *
+ * been put on the screen.  == -2 if refresh() needs to put up a new  *
+ * list.                                                              */
+
+/**/
+int showinglist;
+
+/* Non-zero if ALWAYS_LAST_PROMPT has been used, meaning that the *
+ * screen below the buffer display should not be cleared by       *
+ * refresh(), but should be by trashzle().                        */
+
+/**/
+int clearflag;
+
 #ifdef HAVE_SELECT
-#define SELECT_ADD_COST(X)	cost += X
-#define zputc(a, b)		putc(a, b), cost++
-#define zwrite(a, b, c, d)	fwrite(a, b, c, d), cost += (b * c)
+/* cost of last update */
+/**/
+int cost;
+
+# define SELECT_ADD_COST(X)	cost += X
+# define zputc(a, b)		putc(a, b), cost++
+# define zwrite(a, b, c, d)	fwrite(a, b, c, d), cost += (b * c)
 #else
-#define SELECT_ADD_COST(X)
-#define zputc(a, b)		putc(a, b)
-#define zwrite(a, b, c, d)	fwrite(a, b, c, d)
+# define SELECT_ADD_COST(X)
+# define zputc(a, b)		putc(a, b)
+# define zwrite(a, b, c, d)	fwrite(a, b, c, d)
 #endif
 
 /* Oct/Nov 94: <mason> some code savagely redesigned to fix several bugs -
@@ -49,21 +79,19 @@
    I've put my fingers into just about every routine in here -
    any queries about updates to mason@werple.net.au */
 
-char **nbuf = NULL,		/* new video buffer line-by-line char array */
+static char **nbuf = NULL,	/* new video buffer line-by-line char array */
     **obuf = NULL;		/* old video buffer line-by-line char array */
-static char *lpptbuf, *rpptbuf; /* prompt buffers			    */
 static int more_start,		/* more text before start of screen?	    */
     more_end,			/* more stuff after end of screen?	    */
+    lppth,			/* lines taken up by the prompt		    */
     olnct,			/* previous number of lines		    */
     ovln,			/* previous video cursor position line	    */
-    lpptlen, rpptlen,           /* length of prompt buffers                 */
     pptw, rpw,                  /* prompt widths on screen                  */
+    rppth,			/* right prompt height                      */
     vcs, vln,			/* video cursor position column & line	    */
     vmaxln,			/* video maximum number of lines	    */
     winw, winh, rwinh,		/* window width & height		    */
     winpos;			/* singlelinezle: line's position in window */
-static unsigned pmpt_attr = 0,	/* text attributes after displaying prompt  */
-    rpmpt_attr = 0;		/* text attributes after displaying rprompt */
 
 /**/
 void
@@ -187,9 +215,9 @@ scrollwindow(int tline)
     sen = s + winw;					\
 }
 
-int cleareol,			/* clear to end-of-line (if can't cleareod) */
-    clearf = 0;			/* alwayslastprompt used immediately before */
-static int put_rpmpt,		/* whether we should display right-prompt   */
+static int cleareol,		/* clear to end-of-line (if can't cleareod) */
+    clearf,			/* alwayslastprompt used immediately before */
+    put_rpmpt,			/* whether we should display right-prompt   */
     oput_rpmpt,			/* whether displayed right-prompt last time */
     oxtabs,			/* oxtabs - tabs expand to spaces if set    */
     numscrolls, onumscrolls;
@@ -261,8 +289,8 @@ refresh(void)
             olnct = t0;
         if (termflags & TERM_SHORT)
             vcs = 0;
-        else if (!clearflag && lpptlen)
-            zwrite(lpptbuf, lpptlen, 1, shout);
+        else if (!clearflag && lpptbuf[0])
+            zputs(lpptbuf, shout);
 	if (clearflag) {
 	    zputc('\r', shout);
 	    vcs = 0;
@@ -397,7 +425,8 @@ refresh(void)
 
 /* determine whether the right-prompt exists and can fit on the screen */
     if (!more_start)
-	put_rpmpt = rpptlen && (int)strlen(nbuf[0]) + rpw < winw - 1;
+	put_rpmpt = rppth == 1 && rpptbuf[0] && !strchr(rpptbuf, '\t') &&
+	    (int)strlen(nbuf[0]) + rpw < winw - 1;
     else {
 /* insert >.... on first line if there is more text before start of screen */
 	memset(nbuf[0], ' ', pptw);
@@ -449,7 +478,7 @@ refresh(void)
     /* output the right-prompt if appropriate */
 	if (put_rpmpt && !ln && !oput_rpmpt) {
 	    moveto(0, winw - 1 - rpw);
-	    zwrite(rpptbuf, rpptlen, 1, shout);
+	    zputs(rpptbuf, shout);
 	    vcs = winw - 1;
 	/* reset character attributes to that set by the main prompt */
 	    txtchange = pmpt_attr;
@@ -571,9 +600,10 @@ refreshline(int ln)
 	if (nllen)
 	    strncpy(p1, nl, nllen);
 	memset(p1 + nllen, ' ', winw - nllen);
-	p1[winw] = p1[winw + 1] = '\0';
+	p1[winw] = '\0';
+	p1[winw + 1] = (nllen < winw) ? '\0' : nl[winw + 1];
 	if (ln && nbuf[ln])
-	    strncpy(nl, p1, winw + 2);	/* next time obuf will be up-to-date */
+	    memcpy(nl, p1, winw + 2);	/* next time obuf will be up-to-date */
 	else
 	    nl = p1;		/* don't keep the padding for prompt line */
 	nllen = winw;
@@ -754,6 +784,8 @@ refreshline(int ln)
 void
 moveto(int ln, int cl)
 {
+    int c;
+
     if (vcs == winw) {
 	vln++, vcs = 0;
 	if (!hasam) {
@@ -761,12 +793,13 @@ moveto(int ln, int cl)
 	    zputc('\n', shout);
 	} else {
 	    if ((vln < nlnct) && nbuf[vln] && *nbuf[vln])
-		zputc(*nbuf[vln], shout);
+		c = *nbuf[vln];
 	    else
-		zputc(' ', shout);
+		c = ' ';
+	    zputc(c, shout);
 	    zputc('\r', shout);
 	    if ((vln < olnct) && obuf[vln] && *obuf[vln])
-		*obuf[vln] = *nbuf[vln];
+		*obuf[vln] = c;
 	}
     }
 
@@ -860,18 +893,19 @@ tc_rightcurs(int cl)
 
 /* otherwise _carefully_ write the contents of the video buffer.
    if we're anywhere in the prompt, goto the left column and write the whole
-   prompt out unless lpptlen == pptw : we can cheat then */
+   prompt out unless ztrlen(lpptbuf) == pptw : we can cheat then */
     if (vln == 0 && i < pptw) {
-	if (lpptlen == pptw)
-	    zwrite(lpptbuf + i, lpptlen - i, 1, shout);
-	else if (tccan(TCRIGHT) && (tclen[TCRIGHT] * ct <= lpptlen))
+	if (strlen(lpptbuf) == pptw)
+	    fputs(lpptbuf + i, shout);
+	else if (tccan(TCRIGHT) && (tclen[TCRIGHT] * ct <= ztrlen(lpptbuf)))
 	    /* it is cheaper to send TCRIGHT than reprint the whole prompt */
 	    for (ct = pptw - i; ct--; )
 		tcout(TCRIGHT);
         else {
 	    if (i != 0)
 		zputc('\r', shout);
-	    zwrite(lpptbuf, lpptlen, 1, shout);
+	    tc_upcurs(lppth - 1);
+	    zputs(lpptbuf, shout);
 	}
 	i = pptw;
 	ct = cl - i;
@@ -968,7 +1002,7 @@ singlerefresh(void)
 	cs = 0;
     }
 
-    memcpy(vbuf, lpptbuf + lpptlen - pptw, pptw); /* only use last part of prompt */
+    memcpy(vbuf, strchr(lpptbuf, 0) - pptw, pptw); /* only use last part of prompt */
     vbuf[pptw] = '\0';
     vp = vbuf + pptw;
 
@@ -1071,16 +1105,12 @@ singmoveto(int pos)
     }
 }
 
-/* generate left and right prompts */
+/* recheck size of prompts */
 
 /**/
 static void
 genprompts(void)
 {
-    zfree(lpptbuf, lpptlen);
-    zfree(rpptbuf, rpptlen);
-    lpptbuf = putprompt(lpmpt, &lpptlen, &pptw, 1);
-    pmpt_attr = txtchange;
-    rpptbuf = putprompt(rpmpt, &rpptlen, &rpw, 1);
-    rpmpt_attr = txtchange;
+    countprompt(lpptbuf, &pptw, &lppth);
+    countprompt(rpptbuf, &rpw, &rppth);
 }
