@@ -1,6 +1,4 @@
 /*
- * $Id: zle_thingy.c,v 3.1.2.7 1997/06/01 06:13:15 hzoli Exp $
- *
  * zle_thingy.c - thingies
  *
  * This file is part of zsh, the Z shell.
@@ -29,9 +27,7 @@
  *
  */
 
-#define IN_ZLE
-#define IN_COMP
-#include "zle.h"
+#include "zle.mdh"
 #include "zle_thingy.pro"
 
 /*
@@ -50,6 +46,11 @@
  * a table of widgets.
  */
 
+/* Hashtable of thingies.  Enabled nodes are those that refer to widgets. */
+
+/**/
+HashTable thingytab;
+
 /**********************************/
 /* hashtable management functions */
 /**********************************/
