@@ -1,6 +1,4 @@
 /*
- * $Id: zle_utils.c,v 3.1.2.5 1997/06/01 06:13:15 hzoli Exp $
- *
  * zle_utils.c - miscellaneous line editor utilities
  *
  * This file is part of zsh, the Z shell.
@@ -29,11 +27,39 @@
  *
  */
 
-#define IN_ZLE
-#define IN_COMP
-#include "zle.h"
+#include "zle.mdh"
 #include "zle_utils.pro"
 
+/* Primary cut buffer */
+
+/**/
+struct cutbuffer cutbuf;
+
+/* Emacs-style kill buffer ring */
+
+/**/
+struct cutbuffer kring[KRINGCT];
+/**/
+int kringnum;
+
+/* Vi named cut buffers.  0-25 are the named buffers "a to "z, and *
+ * 26-34 are the numbered buffer stack "1 to "9.                   */
+
+/**/
+struct cutbuffer vibuf[35];
+
+/* the line before last mod (for undo purposes) */
+
+/**/
+char *lastline;
+/**/
+int lastlinesz, lastll;
+
+/* size of line buffer */
+
+/**/
+int linesz;
+
 /* make sure that the line buffer has at least sz chars */
 
 /**/
@@ -75,7 +101,7 @@ shiftchars(int to, int cnt)
 	line[to] = line[to + cnt];
 	to++;
     }
-    ll = to;
+    line[ll = to] = '\0';
 }
 
 /**/
@@ -398,7 +424,17 @@ handlefeep(void)
     feepflag = 0;
 }
 
+/***************/
 /* undo system */
+/***************/
+
+/* head of the undo list, and the current position */
+
+static struct change *changes, *curchange;
+
+/* list of pending changes, not yet in the undo system */
+
+static struct change *nextchanges, *endnextchanges;
 
 /**/
 void
