@@ -1,6 +1,4 @@
 /*
- * $Id: zle_vi.c,v 3.1.2.4 1997/06/01 06:13:15 hzoli Exp $
- *
  * zle_vi.c - vi-specific functions
  *
  * This file is part of zsh, the Z shell.
@@ -29,11 +27,37 @@
  *
  */
 
-#define IN_ZLE
-#define IN_COMP
-#include "zle.h"
+#include "zle.mdh"
 #include "zle_vi.pro"
 
+/* != 0 if we're getting a vi range */
+
+/**/
+int virangeflag;
+
+/* kludge to get cw and dw to work right */
+
+/**/
+int wordflag;
+
+/* != 0 if we're killing lines into a buffer, vi-style */
+
+/**/
+int vilinerange;
+
+/* last vi change buffer, for vi change repetition */
+
+/**/
+int vichgbufsz, vichgbufptr, vichgflag;
+
+/**/
+char *vichgbuf;
+
+/* point where vi insert mode was last entered */
+
+/**/
+int viinsbegin;
+
 static struct modifier lastmod;
 static int inrepeat, vichgrepeat;
 
@@ -894,8 +918,8 @@ vidigitorbeginningofline(void)
     if(zmod.flags & MOD_TMULT)
 	digitargument();
     else {
+	removesuffix();
 	invalidatelist();
-	addedsuffix = complexpect = 0;
 	vibeginningofline();
     }
 }
