@@ -1,6 +1,4 @@
 /*
- * $Id: builtin.c,v 3.1.2.16 1997/06/01 06:13:15 hzoli Exp $
- *
  * builtin.c - builtin commands
  *
  * This file is part of zsh, the Z shell.
@@ -29,9 +27,168 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "builtin.pro"
 
+/* Builtins in the main executable */
+
+static struct builtin builtins[] =
+{
+    BIN_PREFIX("-", BINF_DASH),
+    BIN_PREFIX("builtin", BINF_BUILTIN),
+    BIN_PREFIX("command", BINF_COMMAND),
+    BIN_PREFIX("exec", BINF_EXEC),
+    BIN_PREFIX("noglob", BINF_NOGLOB),
+    BUILTIN("[", 0, bin_test, 0, -1, BIN_BRACKET, NULL, NULL),
+    BUILTIN(".", BINF_PSPECIAL, bin_dot, 1, -1, 0, NULL, NULL),
+    BUILTIN(":", BINF_PSPECIAL, bin_true, 0, -1, 0, NULL, NULL),
+    BUILTIN("alias", BINF_MAGICEQUALS, bin_alias, 0, -1, 0, "Lgmr", NULL),
+    BUILTIN("autoload", BINF_TYPEOPTS, bin_functions, 0, -1, 0, "t", "u"),
+    BUILTIN("bg", 0, bin_fg, 0, -1, BIN_BG, NULL, NULL),
+    BUILTIN("break", BINF_PSPECIAL, bin_break, 0, 1, BIN_BREAK, NULL, NULL),
+    BUILTIN("bye", 0, bin_break, 0, 1, BIN_EXIT, NULL, NULL),
+    BUILTIN("cd", 0, bin_cd, 0, 2, BIN_CD, NULL, NULL),
+    BUILTIN("chdir", 0, bin_cd, 0, 2, BIN_CD, NULL, NULL),
+    BUILTIN("continue", BINF_PSPECIAL, bin_break, 0, 1, BIN_CONTINUE, NULL, NULL),
+    BUILTIN("declare", BINF_TYPEOPTS | BINF_MAGICEQUALS | BINF_PSPECIAL, bin_typeset, 0, -1, 0, "LRUZfilrtux", NULL),
+    BUILTIN("dirs", 0, bin_dirs, 0, -1, 0, "v", NULL),
+    BUILTIN("disable", 0, bin_enable, 0, -1, BIN_DISABLE, "afmr", NULL),
+    BUILTIN("disown", 0, bin_fg, 0, -1, BIN_DISOWN, NULL, NULL),
+    BUILTIN("echo", BINF_PRINTOPTS | BINF_ECHOPTS, bin_print, 0, -1, BIN_ECHO, "neE", "-"),
+    BUILTIN("echotc", 0, bin_echotc, 1, -1, 0, NULL, NULL),
+    BUILTIN("emulate", 0, bin_emulate, 1, 1, 0, "R", NULL),
+    BUILTIN("enable", 0, bin_enable, 0, -1, BIN_ENABLE, "afmr", NULL),
+    BUILTIN("eval", BINF_PSPECIAL, bin_eval, 0, -1, BIN_EVAL, NULL, NULL),
+    BUILTIN("exit", BINF_PSPECIAL, bin_break, 0, 1, BIN_EXIT, NULL, NULL),
+    BUILTIN("export", BINF_TYPEOPTS | BINF_MAGICEQUALS | BINF_PSPECIAL, bin_typeset, 0, -1, BIN_EXPORT, "LRUZfilrtu", "x"),
+    BUILTIN("false", 0, bin_false, 0, -1, 0, NULL, NULL),
+    BUILTIN("fc", BINF_FCOPTS, bin_fc, 0, -1, BIN_FC, "nlreIRWAdDfEim", NULL),
+    BUILTIN("fg", 0, bin_fg, 0, -1, BIN_FG, NULL, NULL),
+    BUILTIN("functions", BINF_TYPEOPTS, bin_functions, 0, -1, 0, "mtu", NULL),
+    BUILTIN("getln", 0, bin_read, 0, -1, 0, "ecnAlE", "zr"),
+    BUILTIN("getopts", 0, bin_getopts, 2, -1, 0, NULL, NULL),
+    BUILTIN("hash", BINF_MAGICEQUALS, bin_hash, 0, -1, 0, "dfmrv", NULL),
+
+#ifdef ZSH_HASH_DEBUG
+    BUILTIN("hashinfo", 0, bin_hashinfo, 0, 0, 0, NULL, NULL),
+#endif
+
+    BUILTIN("history", 0, bin_fc, 0, -1, BIN_FC, "nrdDfEim", "l"),
+    BUILTIN("integer", BINF_TYPEOPTS | BINF_MAGICEQUALS | BINF_PSPECIAL, bin_typeset, 0, -1, 0, "lrtux", "i"),
+    BUILTIN("jobs", 0, bin_fg, 0, -1, BIN_JOBS, "dlpZrs", NULL),
+    BUILTIN("kill", 0, bin_kill, 0, -1, 0, NULL, NULL),
+    BUILTIN("let", 0, bin_let, 1, -1, 0, NULL, NULL),
+    BUILTIN("local", BINF_TYPEOPTS | BINF_MAGICEQUALS | BINF_PSPECIAL, bin_typeset, 0, -1, 0, "LRUZilrtu", NULL),
+    BUILTIN("log", 0, bin_log, 0, 0, 0, NULL, NULL),
+    BUILTIN("logout", 0, bin_break, 0, 1, BIN_LOGOUT, NULL, NULL),
+
+#if defined(ZSH_MEM) & defined(ZSH_MEM_DEBUG)
+    BUILTIN("mem", 0, bin_mem, 0, 0, 0, "v", NULL),
+#endif
+
+    BUILTIN("popd", 0, bin_cd, 0, 2, BIN_POPD, NULL, NULL),
+    BUILTIN("print", BINF_PRINTOPTS, bin_print, 0, -1, BIN_PRINT, "RDPnrslzNu0123456789pioOcm-", NULL),
+    BUILTIN("pushd", 0, bin_cd, 0, 2, BIN_PUSHD, NULL, NULL),
+    BUILTIN("pushln", BINF_PRINTOPTS, bin_print, 0, -1, BIN_PRINT, NULL, "-nz"),
+    BUILTIN("pwd", 0, bin_pwd, 0, 0, 0, "rLP", NULL),
+    BUILTIN("r", BINF_R, bin_fc, 0, -1, BIN_FC, "nrl", NULL),
+    BUILTIN("read", 0, bin_read, 0, -1, 0, "rzu0123456789pkqecnAlE", NULL),
+    BUILTIN("readonly", BINF_TYPEOPTS | BINF_MAGICEQUALS | BINF_PSPECIAL, bin_typeset, 0, -1, 0, "LRUZfiltux", "r"),
+    BUILTIN("rehash", 0, bin_hash, 0, 0, 0, "dfv", "r"),
+    BUILTIN("return", BINF_PSPECIAL, bin_break, 0, 1, BIN_RETURN, NULL, NULL),
+    BUILTIN("set", BINF_PSPECIAL, bin_set, 0, -1, 0, NULL, NULL),
+    BUILTIN("setopt", 0, bin_setopt, 0, -1, BIN_SETOPT, NULL, NULL),
+    BUILTIN("shift", BINF_PSPECIAL, bin_shift, 0, -1, 0, NULL, NULL),
+    BUILTIN("source", BINF_PSPECIAL, bin_dot, 1, -1, 0, NULL, NULL),
+    BUILTIN("suspend", 0, bin_suspend, 0, 0, 0, "f", NULL),
+    BUILTIN("test", 0, bin_test, 0, -1, BIN_TEST, NULL, NULL),
+    BUILTIN("ttyctl", 0, bin_ttyctl, 0, 0, 0, "fu", NULL),
+    BUILTIN("times", BINF_PSPECIAL, bin_times, 0, 0, 0, NULL, NULL),
+    BUILTIN("trap", BINF_PSPECIAL, bin_trap, 0, -1, 0, NULL, NULL),
+    BUILTIN("true", 0, bin_true, 0, -1, 0, NULL, NULL),
+    BUILTIN("type", 0, bin_whence, 0, -1, 0, "ampfs", "v"),
+    BUILTIN("typeset", BINF_TYPEOPTS | BINF_MAGICEQUALS | BINF_PSPECIAL, bin_typeset, 0, -1, 0, "LRUZfilrtuxm", NULL),
+    BUILTIN("umask", 0, bin_umask, 0, 1, 0, "S", NULL),
+    BUILTIN("unalias", 0, bin_unhash, 1, -1, 0, "m", "a"),
+    BUILTIN("unfunction", 0, bin_unhash, 1, -1, 0, "m", "f"),
+    BUILTIN("unhash", 0, bin_unhash, 1, -1, 0, "adfm", NULL),
+    BUILTIN("unset", BINF_PSPECIAL, bin_unset, 1, -1, 0, "fm", NULL),
+    BUILTIN("unsetopt", 0, bin_setopt, 0, -1, BIN_UNSETOPT, NULL, NULL),
+    BUILTIN("wait", 0, bin_fg, 0, -1, BIN_WAIT, NULL, NULL),
+    BUILTIN("whence", 0, bin_whence, 0, -1, 0, "acmpvfs", NULL),
+    BUILTIN("where", 0, bin_whence, 0, -1, 0, "pms", "ca"),
+    BUILTIN("which", 0, bin_whence, 0, -1, 0, "amps", "c"),
+
+#ifdef DYNAMIC
+    BUILTIN("zmodload", 0, bin_zmodload, 0, -1, 0, "Laudi", NULL),
+#endif
+};
+
+/****************************************/
+/* Builtin Command Hash Table Functions */
+/****************************************/
+
+/* hash table containing builtin commands */
+
+/**/
+HashTable builtintab;
+ 
+/**/
+void
+createbuiltintable(void)
+{
+    builtintab = newhashtable(85, "builtintab", NULL);
+
+    builtintab->hash        = hasher;
+    builtintab->emptytable  = NULL;
+    builtintab->filltable   = NULL;
+    builtintab->addnode     = addhashnode;
+    builtintab->getnode     = gethashnode;
+    builtintab->getnode2    = gethashnode2;
+    builtintab->removenode  = removehashnode;
+    builtintab->disablenode = disablehashnode;
+    builtintab->enablenode  = enablehashnode;
+    builtintab->freenode    = freebuiltinnode;
+    builtintab->printnode   = printbuiltinnode;
+
+    addbuiltins("zsh", builtins, sizeof(builtins)/sizeof(*builtins));
+}
+
+/* Print a builtin */
+
+/**/
+static void
+printbuiltinnode(HashNode hn, int printflags)
+{
+    Builtin bn = (Builtin) hn;
+
+    if (printflags & PRINT_WHENCE_CSH) {
+	printf("%s: shell built-in command\n", bn->nam);
+	return;
+    }
+
+    if (printflags & PRINT_WHENCE_VERBOSE) {
+	printf("%s is a shell builtin\n", bn->nam);
+	return;
+    }
+
+    /* default is name only */
+    printf("%s\n", bn->nam);
+}
+
+/**/
+static void
+freebuiltinnode(HashNode hn)
+{
+    Builtin bn = (Builtin) hn;
+ 
+    if(!(bn->flags & BINF_ADDED)) {
+	zsfree(bn->nam);
+	zsfree(bn->optstr);
+	zfree(bn, sizeof(struct builtin));
+    }
+}
+
 static char *auxdata;
 static int auxlen;
 
@@ -384,6 +541,7 @@ bin_set(char *nam, char **args, char *ops, int func)
 
 /**** directory-handling builtins ****/
 
+/**/
 int doprintdir = 0;		/* set in exec.c (for autocd) */
 
 /* pwd: display the name of the current directory */
@@ -401,6 +559,11 @@ bin_pwd(char *name, char **argv, char *ops, int func)
     return 0;
 }
 
+/* the directory stack */
+ 
+/**/
+LinkList dirstack;
+ 
 /* dirs: list the directory stack, or replace it with a provided list */
 
 /**/
@@ -914,7 +1077,7 @@ printif(char *str, int c)
 int
 bin_fc(char *nam, char **argv, char *ops, int func)
 {
-    int first = -1, last = -1, retval, delayrem, minflag = 0;
+    int first = -1, last = -1, retval, minflag = 0;
     char *s;
     struct asgment *asgf = NULL, *asgl = NULL;
     Comp com = NULL;
@@ -933,10 +1096,6 @@ bin_fc(char *nam, char **argv, char *ops, int func)
 	    return 1;
 	}
     }
-    delayrem = 0;
-    if (!(ops['l'] && unset(HISTNOSTORE)) &&
-	!(ops['R'] || ops['W'] || ops['A']))
-	delayrem = 1;
     if (ops['R']) {
 	/* read history from a file */
 	readhistfile(*argv ? *argv : getsparam("HISTFILE"), 1);
@@ -954,6 +1113,8 @@ bin_fc(char *nam, char **argv, char *ops, int func)
 		     (ops['I'] ? 3 : 1));
 	return 0;
     }
+    if (!(ops['l'] && unset(HISTNOSTORE)))
+	remhist();
     /* put foo=bar type arguments into the substitution list */
     while (*argv && equalsplit(*argv, &s)) {
 	Asgment a = (Asgment) alloc(sizeof *a);
@@ -972,29 +1133,21 @@ bin_fc(char *nam, char **argv, char *ops, int func)
     if (*argv) {
 	minflag = **argv == '-';
 	first = fcgetcomm(*argv);
-	if (first == -1) {
-	    if (delayrem)
-		remhist();
+	if (first == -1)
 	    return 1;
-	}
 	argv++;
     }
     /* interpret and check second history line specifier */
     if (*argv) {
 	last = fcgetcomm(*argv);
-	if (last == -1) {
-	    if (delayrem)
-		remhist();
+	if (last == -1)
 	    return 1;
-	}
 	argv++;
     }
     /* There is a maximum of two history specifiers.  At least, there *
      * will be as long as the history list is one-dimensional.        */
     if (*argv) {
 	zwarnnam("fc", "too many arguments", NULL, 0);
-	if (delayrem)
-	    remhist();
 	return 1;
     }
     /* default values of first and last, and range checking */
@@ -1035,14 +1188,14 @@ bin_fc(char *nam, char **argv, char *ops, int func)
 		if (fcedit(editor, fil))
 		    if (stuff(fil))
 			zwarnnam("fc", "%e: %s", s, errno);
-		    else
+		    else {
+			loop(0,1);
 			retval = lastval;
+		    }
 	    }
 	}
 	unlink(fil);
     }
-    if (delayrem)
-	remhist();
     return retval;
 }
 
@@ -2071,6 +2224,11 @@ bin_false(char *name, char **argv, char *ops, int func)
     return 1;
 }
 
+/* the zle buffer stack */
+ 
+/**/
+LinkList bufstack;
+
 /* echo, print, pushln */
 
 /**/
@@ -2109,14 +2267,9 @@ bin_print(char *name, char **args, char *ops, int func)
 	    args[n] = getkeystring(args[n], &len[n],
 				    func != BIN_ECHO && !ops['e'], &nnl);
 	/* -P option -- interpret as a prompt sequence */
-	if(ops['P']) {
-	    char *arg = putprompt(metafy(args[n], len[n], META_NOALLOC),
-				  &len[n], NULL, 0);
-	    args[n] = (char *)alloc(len[n] + 1);
-	    memcpy(args[n], arg, len[n]);
-	    args[n][len[n]] = 0;
-	    free(arg);
-	}
+	if(ops['P'])
+	    args[n] = unmetafy(promptexpand(metafy(args[n], len[n],
+		META_NOALLOC), 0, NULL, NULL), &len[n]);
 	/* -D option -- interpret as a directory, and use ~ */
 	if(ops['D']) {
 	    Nameddir d = finddir(args[n]);
@@ -2372,87 +2525,88 @@ bin_shift(char *name, char **argv, char *ops, int func)
 int
 bin_getopts(char *name, char **argv, char *ops, int func)
 {
-    int lenstr, lenoptstr, i;
+    int lenstr, lenoptstr, quiet, lenoptbuf;
     char *optstr = unmetafy(*argv++, &lenoptstr), *var = *argv++;
     char **args = (*argv) ? argv : pparams;
-    static int optcind = 1, quiet;
-    char *str, optbuf[2], *opch = optbuf + 1;
+    static int optcind = 0;
+    char *str, optbuf[2], *p, opch;
 
-    /* zoptind keeps count of the current argument number */
-    if (zoptind < 1)
+    /* zoptind keeps count of the current argument number.  The *
+     * user can set it to zero to start a new option parse.     */
+    if (zoptind < 1) {
 	/* first call */
 	zoptind = 1;
-    if (zoptind == 1)
-	quiet = 0;
-    optbuf[0] = '+';
-    zsfree(zoptarg);
-    zoptarg = ztrdup("");
-    setsparam(var, ztrdup(""));
-    if (*optstr == ':') {
-	quiet = 1;
-	optstr++;
+	optcind = 0;
     }
-    if (zoptind > arrlen(args))
+    if(zoptind > arrlen(args))
+	/* no more options */
 	return 1;
-    str = unmetafy(args[zoptind - 1], &lenstr);
-    if ((*str != '+' && *str != '-') || optcind >= lenstr ||
-	(lenstr == 2 && str[0] == '-' && str[1] == '-')) {
-	/* current argument doesn't contain options, or optcind is impossibly
-	large */
-	if (*str == '+' || *str == '-')
+
+    /* leading ':' in optstr means don't print an error message */
+    quiet = *optstr == ':';
+    optstr += quiet;
+    lenoptstr -= quiet;
+
+    /* find place in relevant argument */
+    str = unmetafy(dupstring(args[zoptind - 1]), &lenstr);
+    if(optcind >= lenstr) {
+	if(!args[zoptind++])
+	    return 1;
+	str = unmetafy(dupstring(args[zoptind - 1]), &lenstr);
+	optcind = 0;
+    }
+    if(!optcind) {
+	if(lenstr < 2 || (*str != '-' && *str != '+'))
+	    return 1;
+	if(lenstr == 2 && str[0] == '-' && str[1] == '-') {
 	    zoptind++;
-	optcind = 0;
-	return 1;
+	    return 1;
+	}
+	optcind++;
     }
-    /* Get the option character.  optcind records the current position within
-    the argument. */
-    if (!optcind)
-	optcind = 1;
-    *opch = str[optcind++];
-    if (optcind == lenstr) {
-	if(args[zoptind++])
-	    str = unmetafy(args[zoptind - 1], &lenstr);
-	optcind = 0;
-    }
-    /* look for option in the provided optstr */
-    for (i = 0; i != lenoptstr; i++)
-	if (*opch == optstr[i])
-	    break;
-    if (i == lenoptstr) {
-	/* not a valid option */
-	setsparam(var, ztrdup("?"));
-	if (quiet) {
+    opch = str[optcind++];
+    if(str[0] == '+') {
+	optbuf[0] = '+';
+	lenoptbuf = 2;
+    } else
+	lenoptbuf = 1;
+    optbuf[lenoptbuf - 1] = opch;
+
+    /* check for legality */
+    if(opch == ':' || !(p = memchr(optstr, opch, lenoptstr))) {
+	p = "?";
+err:
+	if(quiet) {
+	    setsparam(var, ztrdup(p));
 	    zsfree(zoptarg);
-	    zoptarg = metafy(opch, 1, META_DUP);
-	    return 0;
+	    zoptarg = metafy(optbuf, lenoptbuf, META_DUP);
+	} else {
+	    zerr(*p == '?' ? "bad option: -%c" :
+		"argument expected after -%c option", NULL, opch);
+	    errflag = 0;
 	}
-	zerr("bad option: -%c", NULL, *opch);
-	errflag = 0;
 	return 0;
     }
-    /* copy option into specified parameter, with + if required */
-    setsparam(var, metafy(opch - (*str == '+'), 1 + (*str == '+'), META_DUP));
-    /* handle case of an expected extra argument */
-    if (optstr[i + 1] == ':') {
-	if (!args[zoptind - 1]) {
-	    /* no extra argument was provided */
-	    if (quiet) {
-		zsfree(zoptarg);
-		zoptarg = metafy(opch, 1, META_DUP);
-		setsparam(var, ztrdup(":"));
-		return 0;
+
+    /* check for required argument */
+    if(p[1] == ':') {
+	if(optcind == lenstr) {
+	    if(!args[zoptind]) {
+		p = ":";
+		goto err;
 	    }
-	    setsparam(var, ztrdup("?"));
-	    zerr("argument expected after -%c option", NULL, *opch);
-	    errflag = 0;
-	    return 0;
-	}
-	/* skip over the extra argument */
+	    p = ztrdup(args[zoptind++]);
+	} else
+	    p = metafy(str+optcind, lenstr-optcind, META_DUP);
+	optcind = ztrlen(args[zoptind - 1]);
 	zsfree(zoptarg);
-	zoptarg = metafy(str + optcind, lenstr - optcind, META_DUP);
-	zoptind++;
-	optcind = 0;
+	zoptarg = p;
+    } else {
+	zsfree(zoptarg);
+	zoptarg = ztrdup("");
     }
+
+    setsparam(var, metafy(optbuf, lenoptbuf, META_DUP));
     return 0;
 }
 
@@ -2511,6 +2665,11 @@ bin_break(char *name, char **argv, char *ops, int func)
     return 0;
 }
 
+/* we have printed a 'you have stopped (running) jobs.' message */
+ 
+/**/
+int stopmsg;
+ 
 /* check to see if user has jobs running/stopped */
 
 /**/
@@ -2704,8 +2863,6 @@ static int readfd;
 /* Read a character from readfd, or from the buffer zbuf.  Return EOF on end of
 file/buffer. */
 
-extern int cs;
-
 /* read: get a line of input, or (for compctl functions) return some *
  * useful data about the state of the editing line.  The -E and -e   *
  * options mean that the result should be sent to stdout.  -e means, *
@@ -2849,7 +3006,7 @@ bin_read(char *name, char **args, char *ops, int func)
 	readbuf[1] = '\0';
 
 	/* get, and store, reply */
-	readbuf[0] = ((char)getquery(NULL)) == 'y' ? 'y' : 'n';
+	readbuf[0] = ((char)getquery(NULL, 0)) == 'y' ? 'y' : 'n';
 
 	/* dispose of result appropriately, etc. */
 	if (haso) {
@@ -3046,10 +3203,8 @@ zread(void)
     return STOUC(cc);
 }
 
-/* holds lexer for par_cond():  normally yylex(), testlex() for bin_test() */
-extern void (*condlex) _((void));
-
 /* holds arguments for testlex() */
+/**/
 char **testargs;
 
 /* test, [: the old-style general purpose logical expression builtin */
@@ -3368,3 +3523,13 @@ bin_umask(char *nam, char **args, char *ops, int func)
     umask(um);
     return 0;
 }
+
+/* Generic builtin for facilities not available on this OS */
+
+/**/
+int
+bin_notavail(char *nam, char **argv, char *ops, int func)
+{
+    zwarnnam(nam, "not available on this system", NULL, 0);
+    return 1;
+}
