@@ -1,6 +1,4 @@
 /*
- * $Id: compat.c,v 3.1.2.2 1997/06/01 06:13:15 hzoli Exp $
- *
  * compat.c - compatibiltiy routines for the deprived
  *
  * This file is part of zsh, the Z shell.
@@ -29,7 +27,7 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "compat.pro"
 
 /* Return pointer to first occurence of string t *
@@ -67,11 +65,12 @@ gethostname(char *name, int namelen)
 
 
 #ifndef HAVE_GETTIMEOFDAY
-void
+int
 gettimeofday(struct timeval *tv, struct timezone *tz)
 {
     tv->tv_usec = 0;
     tv->tv_sec = (long)time((time_t) 0);
+    return 0;
 }
 #endif
 
