@@ -1,6 +1,4 @@
 /*
- * $Id: cond.c,v 3.1.2.3 1997/06/01 06:13:15 hzoli Exp $
- *
  * cond.c - evaluate conditional expressions
  *
  * This file is part of zsh, the Z shell.
@@ -29,7 +27,7 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "cond.pro"
 
 /**/
@@ -88,19 +86,15 @@ evalcond(Cond c)
     case 's':
 	return ((st = getstat(c->left)) && !!(st->st_size));
     case 'S':
-#ifdef S_ISSOCK
 	return (S_ISSOCK(dostat(c->left)));
-#else
-	return 0;   /* some versions of SCO are missing S_ISSOCK */
-#endif
     case 'u':
 	return (!!(dostat(c->left) & S_ISUID));
     case 'w':
 	return (doaccess(c->left, W_OK));
     case 'x':
 	if (privasserted()) {
-	    unsigned short mode = dostat(c->left);
-	    return (mode & 0111) || S_ISDIR(mode);
+	    mode_t mode = dostat(c->left);
+	    return (mode & S_IXUGO) || S_ISDIR(mode);
 	}
 	return doaccess(c->left, X_OK);
     case 'z':
@@ -189,7 +183,7 @@ getstat(char *s)
 
 
 /**/
-static unsigned short
+static mode_t
 dostat(char *s)
 {
     struct stat *statp;
@@ -203,7 +197,7 @@ dostat(char *s)
 /* pem@aaii.oz; needed since dostat now uses "stat" */
 
 /**/
-static unsigned short
+static mode_t
 dolstat(char *s)
 {
     if (lstat(unmeta(s), &st) < 0)
