@@ -1,6 +1,4 @@
 /*
- * $Id: exec.c,v 3.1.2.11 1997/06/01 06:13:15 hzoli Exp $
- *
  * exec.c - command execution
  *
  * This file is part of zsh, the Z shell.
@@ -29,11 +27,99 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "exec.pro"
 
+/* used to suppress ERREXIT and trapping of SIGZERR, SIGEXIT. */
+
+/**/
+int noerrexit;
+
+/* suppress error messages */
+ 
+/**/
+int noerrs;
+ 
+/* do not save history on exec and exit */
+
+/**/
+int nohistsave;
+ 
+/* error/break flag */
+ 
+/**/
+int errflag;
+ 
+/* Status of return from a trap */
+ 
+/**/
+int trapreturn;
+ 
+/* != 0 if this is a subshell */
+ 
+/**/
+int subsh;
+ 
+/* != 0 if we have a return pending */
+ 
+/**/
+int retflag;
+
+/**/
+long lastval2;
+ 
+/* The table of file descriptors.  A table element is zero if the  *
+ * corresponding fd is not used by the shell.  It is greater than  *
+ * 1 if the fd is used by a <(...) or >(...) substitution and 1 if *
+ * it is an internal file descriptor which must be closed before   *
+ * executing an external command.  The first ten elements of the   *
+ * table is not used.  A table element is set by movefd and cleard *
+ * by zclose.                                                      */
+
+/**/
+char *fdtable;
+
+/* The allocated size of fdtable */
+
+/**/
+int fdtable_size;
+
+/* The highest fd that marked with nonzero in fdtable */
+
+/**/
+int max_zsh_fd;
+
+/* input fd from the coprocess */
+
+/**/
+int coprocin;
+
+/* output fd from the coprocess */
+
+/**/
+int coprocout;
+
+/* != 0 if the line editor is active */
+
+/**/
+int zleactive;
+
+/* pid of process undergoing 'process substitution' */
+ 
+/**/
+pid_t cmdoutpid;
+ 
+/* exit status of process undergoing 'process substitution' */
+ 
+/**/
+int cmdoutval;
+ 
+/* Stack to save some variables before executing a signal handler function */
+
+/**/
+struct execstack *exstack;
+
 #define execerr() if (!forked) { lastval = 1; return; } else _exit(1)
-	
 
 static LinkList args;
 static int doneps4;
@@ -58,6 +144,12 @@ parse_string(char *s)
 }
 
 #ifdef HAVE_GETRLIMIT
+
+/* the resource limits for the shell and its children */
+
+/**/
+struct rlimit current_limits[RLIM_NLIMITS], limits[RLIM_NLIMITS];
+ 
 /**/
 int
 zsetlimit(int limnum, char *nam)
@@ -86,7 +178,8 @@ setlimits(char *nam)
 	    ret++;
     return ret;
 }
-#endif
+
+#endif /* HAVE_GETRLIMIT */
 
 /* fork and set limits */
 
@@ -114,6 +207,7 @@ zfork(void)
 }
 
 
+/**/
 int list_pipe = 0, simple_pline = 0;
 
 static pid_t list_pipe_pid;
@@ -654,6 +748,7 @@ execpline(Sublist l, int how, int last1)
     pline_level--;
     if (how & Z_ASYNC) {
 	lastwj = newjob;
+	jobtab[thisjob].stat |= STAT_NOSTTY;
 	if (l->flags & PFLAG_COPROC)
 	    zclose(ipipe[1]);
 	if (how & Z_DISOWN) {
@@ -783,7 +878,7 @@ execpline(Sublist l, int how, int last1)
     return lastval;
 }
 
-int subsh_close = -1;
+static int subsh_close = -1;
 
 /* execute pipeline.  This function assumes the `pline' is not NULL. */
 
@@ -910,18 +1005,26 @@ static int
 clobber_open(struct redir *f)
 {
     struct stat buf;
-    int fd;
+    int fd, oerrno;
 
+    /* If clobbering, just open. */
     if (isset(CLOBBER) || IS_CLOBBER_REDIR(f->type))
 	return open(unmeta(f->name), O_WRONLY | O_CREAT | O_TRUNC, 0666);
+
+    /* If not clobbering, attempt to create file exclusively. */
     if ((fd = open(unmeta(f->name), O_WRONLY | O_CREAT | O_EXCL, 0666)) >= 0)
 	return fd;
-    if ((fd = open(unmeta(f->name), O_WRONLY)) < 0)
-	return fd;
-    if (!fstat(fd, &buf) && !S_ISREG(buf.st_mode))
-	return fd;
-    close(fd);
-    errno = EEXIST;
+
+    /* If that fails, we are still allowed to open non-regular files. *
+     * Try opening, and if it's a regular file then close it again    *
+     * because we weren't supposed to open it.                        */
+    oerrno = errno;
+    if ((fd = open(unmeta(f->name), O_WRONLY)) != -1) {
+	if(!fstat(fd, &buf) && !S_ISREG(buf.st_mode))
+	    return fd;
+	close(fd);
+    }
+    errno = oerrno;
     return -1;
 }
 
@@ -1207,8 +1310,10 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 		is_builtin = 1;
 #ifdef DYNAMIC
 		/* autoload the builtin if necessary */
-		if (!((Builtin) hn)->handlerfunc)
+		if (!((Builtin) hn)->handlerfunc) {
 		    load_module(((Builtin) hn)->optstr);
+		    hn = builtintab->getnode(builtintab, cmdarg);
+		}
 #endif
 		assign = (hn->flags & BINF_MAGICEQUALS);
 		break;
@@ -2452,6 +2557,8 @@ execfuncdef(Cmd cmd)
 	    shfunctab->addnode(shfunctab, ztrdup(s), shf);
 	}
     } LASTALLOC;
+    if(isset(HISTNOFUNCTIONS))
+	remhist();
     return 0;
 }
 
@@ -2719,8 +2826,6 @@ stripkshdef(List l, char *name)
 
 /* check to see if AUTOCD applies here */
 
-extern int doprintdir;
-
 /**/
 static char *
 cancd(char *s)
