@@ -1,6 +1,4 @@
 /*
- * $Id: glob.c,v 3.1.2.2 1997/06/02 04:19:48 hzoli Exp $
- *
  * glob.c - filename generation
  *
  * This file is part of zsh, the Z shell.
@@ -29,9 +27,14 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "glob.pro"
 
+/* flag for CSHNULLGLOB */
+ 
+/**/
+int badcshglob;
+ 
 static int mode;		/* != 0 if we are parsing glob patterns */
 static int pathpos;		/* position in pathbuf                  */
 static int matchsz;		/* size of matchbuf                     */
@@ -77,9 +80,11 @@ static struct qual *quals;
 /* Other state values for current pattern */
 static int qualct, qualorct;
 static int range, amc, units;
-static int gf_nullglob, gf_markdirs, gf_noglobdots, gf_listtypes;
+static int gf_nullglob, gf_markdirs, gf_noglobdots, gf_listtypes, gf_follow;
 
 /* Prefix, suffix for doing zle trickery */
+
+/**/
 char *glob_pre, *glob_suf;
 
 /* pathname component in filename patterns */
@@ -100,17 +105,25 @@ struct comp {
  * pattern or path component, a piece of path to be added.    */
 #define C_ONEHASH	1
 #define C_TWOHASH	2
-#define C_CLOSURE	(C_ONEHASH|C_TWOHASH)
-#define C_LAST		4
-#define C_PATHADD	8
+#define C_OPTIONAL	4
+#define C_STAR		8    
+#define C_CLOSURE	(C_ONEHASH|C_TWOHASH|C_OPTIONAL|C_STAR)
+#define C_LAST		16
+#define C_PATHADD	32
 
 /* Test macros for the above */
 #define CLOSUREP(c)	(c->stat & C_CLOSURE)
-#define ONEHASHP(c)	(c->stat & C_ONEHASH)
+#define ONEHASHP(c)	(c->stat & (C_ONEHASH|C_STAR))
 #define TWOHASHP(c)	(c->stat & C_TWOHASH)
+#define OPTIONALP(c)	(c->stat & C_OPTIONAL)
+#define STARP(c)	(c->stat & C_STAR)
 #define LASTP(c)	(c->stat & C_LAST)
 #define PATHADDP(c)	(c->stat & C_PATHADD)
 
+/* Flags passed down to guts when compiling */
+#define GF_PATHADD	1	/* file glob, adding path components */
+#define GF_TOPLEV	2	/* outside (), so ~ ends main match */
+
 static char *pptr;		/* current place in string being matched */
 static Comp tail = 0;
 static int first;		/* are leading dots special? */
@@ -121,7 +134,7 @@ exists(char *s)
 {
     char *us = unmeta(s);
 
-    return access(us,0) == 0 || readlink(us,NULL,0) == 0;
+    return access(us, F_OK) == 0 || readlink(us,NULL,0) == 0;
 }
 
 /* Add a component to pathbuf: This keeps track of how    *
@@ -168,7 +181,8 @@ insert(char *s, int checked)
     if (gf_listtypes || gf_markdirs) {
 	/* Add the type marker to the end of the filename */
 	checked = statted = 1;
-	if (lstat(unmetafy(getfullpath(s), NULL), &buf))
+	if (gf_follow ? stat(unmetafy(getfullpath(s), NULL), &buf)
+	    : lstat(unmetafy(getfullpath(s), NULL), &buf))
 	    return;
 	if (gf_listtypes || S_ISDIR(buf.st_mode)) {
 	    char *t;
@@ -234,65 +248,42 @@ insert(char *s, int checked)
     }
 }
 
-/* check to see if str is eligible for filename generation
- * It returns NULL if no wilds or modifiers found.
- * If a leading % is immediately followed by a ?, that single
- * ? is not treated as a wildcard.
- * If str has wilds it returns a pointer to the first wildcard.
- * If str has no wilds but ends in a (:...) type modifier it returns
- * a pointer to the colon.
- * If str has no wilds but ends in (...:...) it returns a pointer
- * to the terminating null character of str.
- */
+/* Check to see if str is eligible for filename generation. */
 
 /**/
-static char *
+int
 haswilds(char *str)
 {
-    char *mod = NULL;
-    int parlev = 0;
-
+    /* `[' and `]' are legal even if bad patterns are usually not. */
     if ((*str == Inbrack || *str == Outbrack) && !str[1])
-	return NULL;
+	return 0;
 
     /* If % is immediately followed by ?, then that ? is     *
      * not treated as a wildcard.  This is so you don't have *
      * to escape job references such as %?foo.               */
     if (str[0] == '%' && str[1] == Quest)
 	str[1] = '?';
-    for (; *str; str++)
+
+    if(*str == Tilde)
+	str++;
+    for (; *str; str++) {
 	switch (*str) {
-	case Inpar:
-	    parlev++;
-	    break;
-	case Outpar:
-	    if (! --parlev && str[1])
-		mod = NULL;
-	    break;
-	case ':':
-	    if (parlev == 1 && !mod)
-		mod = str;
-	    break;
-	case Bar:
-	    if (!parlev) {
-		*str = '|';
+	    case Inpar:
+	    case Bar:
+	    case Star:
+	    case Inbrack:
+	    case Inang:
+	    case Quest:
+		return 1;
+	    case Pound:
+	    case Hat:
+	    case Tilde:
+		if (isset(EXTENDEDGLOB))
+		    return 1;
 		break;
-	    } /* else fall through */
-	case Star:
-	case Inbrack:
-	case Inang:
-	case Quest:
-	    return str;
-	case Pound:
-	case Hat:
-	    if (isset(EXTENDEDGLOB))
-		return str;
 	}
-    if (!mod || parlev)
-	return NULL;
-    if (mod[-1] == Inpar)
-	return mod;
-    return str;
+    }
+    return 0;
 }
 
 /* Do the globbing:  scanner is called recursively *
@@ -364,7 +355,7 @@ scanner(Complist q)
 	    struct stat st;
 	    stat(fn, &st);
 	    /* a directory with subdirectories has link count greater than 2 */
-	    if (!S_ISDIR(st.st_mode) || st.st_nlink <= 2)
+	    if (!S_ISDIR(st.st_mode) || st.st_nlink == 2)
 		return;
 	}
 	lock = opendir(fn);
@@ -447,12 +438,17 @@ scanner(Complist q)
 
 /* Parse a series of path components pointed to by pptr */
 
+/* enum used with ksh-like patterns, @(...) etc. */
+
+enum { KF_NONE, KF_AT, KF_QUEST, KF_STAR, KF_PLUS, KF_NOT };
+
 /* parse lowest level pattern */
 
 /**/
 static Comp
-parsecomp(void)
+parsecomp(int gflag)
 {
+    int kshfunc;
     Comp c = (Comp) alloc(sizeof *c), c1, c2;
     char *cstr, *ls = NULL;
 
@@ -469,31 +465,56 @@ parsecomp(void)
 	 */
 	if (*pptr == Hat && isset(EXTENDEDGLOB)) {
 	    /* negate remaining pattern */
-	    pptr++;
-	    c->str = dupstrpfx(cstr, pptr - cstr);
-	    if (!(c->next = parsecomp()))
-		return NULL;
-	    return c;
-	}
-	if (*pptr == Star && pptr[1] &&
-	    (unset(EXTENDEDGLOB) || pptr[1] != Tilde || !pptr[2] ||
-	     pptr[2] == Bar ||
-	     pptr[2] == Outpar) && (mode || pptr[1] != '/')) {
-	    /* Star followed by other patterns is treated like a closure
-	     * (zero or more repetitions) of the single character pattern
-	     * operator `?'.
-	     */
+	    Comp stail = tail;
+	    tail = NULL;
 	    c->str = dupstrpfx(cstr, pptr - cstr);
 	    pptr++;
+
 	    c1 = (Comp) alloc(sizeof *c1);
-	    *(c1->str = dupstring("?")) = Quest;
-	    c1->stat |= C_ONEHASH;
-	    if (!(c2 = parsecomp()))
+	    c1->stat |= C_STAR;
+
+	    c2 = (Comp) alloc(sizeof *c2);
+	    if (!(c2->exclude = parsecomp(gflag)))
 		return NULL;
-	    c1->next = c2;
-	    c->next = c1;
+	    if (!*pptr || *pptr == '/')
+		c2->stat |= C_LAST;
+	    c2->left = c1;
+	    c2->next = stail;
+	    c->next = c2;
+	    tail = stail;
 	    return c;
 	}
+
+	/* Ksh-type globs */
+	kshfunc = KF_NONE;
+	if (isset(KSHGLOB) && *pptr && pptr[1] == Inpar) {
+	    switch (*pptr) {
+	    case '@':		/* just do paren as usual */
+		kshfunc = KF_AT;
+		break;
+
+	    case Quest:
+	    case '?':		/* matched optionally, treat as (...|) */
+		kshfunc = KF_QUEST;
+		break;
+
+	    case Star:
+	    case '*':		/* treat as (...)# */
+		kshfunc = KF_STAR;
+		break;
+
+	    case '+':		/* treat as (...)## */
+		kshfunc = KF_PLUS;
+		break;
+
+	    case '!':		/* treat as (*~...) */
+		kshfunc = KF_NOT;
+		break;
+	    }
+	    if (kshfunc != KF_NONE)
+		pptr++;
+	}
+
 	if (*pptr == Inpar) {
 	    /* Found a group (...) */
 	    char *startp = pptr, *endp;
@@ -505,36 +526,74 @@ parsecomp(void)
 		errflag = 1;
 		return NULL;
 	    }
+	    if (kshfunc == KF_STAR)
+		dpnd = 1;
+	    else if (kshfunc == KF_PLUS)
+		dpnd = 2;
+	    else if (kshfunc == KF_QUEST)
+		dpnd = 3;
 	    if (*pptr == Pound && isset(EXTENDEDGLOB)) {
 		/* Zero (or one) or more repetitions of group */
-		dpnd = 1;
 		pptr++;
-		if (*pptr == Pound) {
+		if(*pptr == Pound) {
 		    pptr++;
-		    dpnd = 2;
-		}
+		    if(dpnd == 0)
+			dpnd = 2;
+		    else if(dpnd == 3)
+			dpnd = 1;
+		} else
+		    dpnd = 1;
 	    }
 	    /* Parse the remaining pattern following the group... */
-	    if (!(c1 = parsecomp()))
+	    if (!(c1 = parsecomp(gflag)))
 		return NULL;
 	    /* ...remembering what comes after it... */
-	    tail = dpnd ? NULL : c1;
+	    tail = (dpnd || kshfunc == KF_NOT) ? NULL : c1;
 	    /* ...before going back and parsing inside the group. */
 	    endp = pptr;
 	    pptr = startp;
-	    c->str = dupstrpfx(cstr, pptr - cstr);
+	    c->str = dupstrpfx(cstr, (pptr - cstr) - (kshfunc != KF_NONE));
 	    pptr++;
-	    c->next = (Comp) alloc(sizeof *c);
-	    if (!(c->next->left = parsecompsw(0)))
+	    c2 = (Comp) alloc(sizeof *c);
+	    c->next = c2;
+	    c2->next = (dpnd || kshfunc == KF_NOT) ?
+		c1 : (Comp) alloc(sizeof *c);
+	    if (!(c2->left = parsecompsw(0)))
 		return NULL;
+	    if (kshfunc == KF_NOT) {
+		/* we'd actually rather it didn't match.  Instead, match *
+		 * a star and put the parsed pattern into exclude.       */
+		Comp c3 = (Comp) alloc(sizeof *c3);
+		c3->stat |= C_STAR;
+
+		c2->exclude = c2->left;
+		c2->left = c3;
+	    }
 	    /* Remember closures for group. */
 	    if (dpnd)
-		c->next->stat |= (dpnd == 2) ? C_TWOHASH : C_ONEHASH;
-	    c->next->next = dpnd ? c1 : (Comp) alloc(sizeof *c);
+		c2->stat |= (dpnd == 3) ? C_OPTIONAL
+		    : (dpnd == 2) ? C_TWOHASH : C_ONEHASH;
 	    pptr = endp;
 	    tail = stail;
 	    return c;
 	}
+	if (*pptr == Star && pptr[1] &&
+	    (unset(EXTENDEDGLOB) || !(gflag & GF_TOPLEV) ||
+	     pptr[1] != Tilde || !pptr[2] || pptr[2] == Bar ||
+	     pptr[2] == Outpar) && (mode || pptr[1] != '/')) {
+	    /* Star followed by other patterns is now treated as a special
+	     * type of closure in doesmatch().
+	     */
+	    c->str = dupstrpfx(cstr, pptr - cstr);
+	    pptr++;
+	    c1 = (Comp) alloc(sizeof *c1);
+	    c1->stat |= C_STAR;
+	    if (!(c2 = parsecomp(gflag)))
+		return NULL;
+	    c1->next = c2;
+	    c->next = c1;
+	    return c;
+	}
 	if (*pptr == Pound && isset(EXTENDEDGLOB)) {
 	    /* repeat whatever we've just had (ls) zero or more times */
 	    if (!ls)
@@ -552,7 +611,7 @@ parsecomp(void)
 	    c1->next = c2;
 	    c2->stat |= C_ONEHASH;
 	    /* parse the rest of the pattern and return. */
-	    c2->next = parsecomp();
+	    c2->next = parsecomp(gflag);
 	    if (!c2->next)
 		return NULL;
 	    c->str = dupstrpfx(cstr, ls - cstr);
@@ -584,7 +643,8 @@ parsecomp(void)
 	pptr++;
     }
     /* mark if last pattern component in path component or pattern */
-    if (*pptr == '/' || !*pptr)
+    if (*pptr == '/' || !*pptr ||
+	(isset(EXTENDEDGLOB) && *pptr == Tilde && (gflag & GF_TOPLEV)))
 	c->stat |= C_LAST;
     c->str = dupstrpfx(cstr, pptr - cstr);
     return c;
@@ -594,11 +654,33 @@ parsecomp(void)
 
 /**/
 static Comp
-parsecompsw(int pathadd)
+parsecompsw(int gflag)
 {
-    Comp c1, c2, c3, excl = NULL;
+    Comp c1, c2, c3, excl = NULL, stail = tail;
+    char *sptr;
 
-    c1 = parsecomp();
+    /*
+     * Check for a tilde in the expression.  We need to know this in 
+     * advance so as to be able to treat the whole a~b expression by
+     * backtracking:  see exclusion code in doesmatch().
+     */
+    if (isset(EXTENDEDGLOB)) {
+	int pct = 0;
+	for (sptr = pptr; *sptr; sptr++) {
+	    if (*sptr == Inpar)
+		pct++;
+	    else if (*sptr == Outpar && --pct < 0)
+		break;
+	    else if (*sptr == Bar && !pct)
+		break;
+	    else if (*sptr == Tilde && !pct) {
+		tail = NULL;
+		break;
+	    }
+	}
+    }
+
+    c1 = parsecomp(gflag);
     if (!c1)
 	return NULL;
     if (isset(EXTENDEDGLOB) && *pptr == Tilde) {
@@ -607,31 +689,32 @@ parsecompsw(int pathadd)
 
 	mode = 1;
 	pptr++;
-	excl = parsecomp();
+	excl = parsecomp(gflag);
 	mode = oldmode;
 	if (!excl)
 	    return NULL;
     }
+    tail = stail;
     if (*pptr == Bar || excl) {
 	/* found an alternative or something to exclude */
 	c2 = (Comp) alloc(sizeof *c2);
 	if (*pptr == Bar) {
 	    /* get the next alternative after the | */
 	    pptr++;
-	    c3 = parsecompsw(pathadd);
+	    c3 = parsecompsw(gflag);
 	    if (!c3)
 		return NULL;
-	} else {
-	    /* mark if end of pattern or path component */
-	    if (!*pptr || *pptr == '/')
-		c2->stat |= C_LAST;
+	} else
 	    c3 = NULL;
-	}
+	/* mark if end of pattern or path component */
+	if (!*pptr || *pptr == '/')
+	    c1->stat |= c2->stat = C_LAST;
 	c2->str = dupstring("");
 	c2->left = c1;
 	c2->right = c3;
-	c2->exclude = excl;
-	if (pathadd)
+	if ((c2->exclude = excl))
+	    c2->next = stail;
+	if (gflag & GF_PATHADD)
 	    c2->stat |= C_PATHADD;
 	return c2;
     }
@@ -695,7 +778,7 @@ parsecomplist(void)
 	}
     } else {
 	/* parse single path component */
-	if (!(c1 = parsecompsw(1)))
+	if (!(c1 = parsecompsw(GF_PATHADD|GF_TOPLEV)))
 	    return NULL;
 	/* then do the remaining path compoents */
 	if (*pptr == '/' || !*pptr) {
@@ -774,15 +857,10 @@ glob(LinkList list, LinkNode np)
 					/* chops it up                   */
 
     MUSTUSEHEAP("glob");
-    if (unset(GLOBOPT) || !(str = haswilds(ostr))) {
+    if (unset(GLOBOPT) || !haswilds(ostr)) {
 	untokenize(ostr);
 	return;
     }
-    if (*str == ':' && str[-1] == Inpar) {
-	str[-1] = '\0';
-	modify((char **) &getdata(np), &str);
-	return;
-    }
     str = dupstring(ostr);
     sl = strlen(str);
     uremnode(list, np);
@@ -793,26 +871,29 @@ glob(LinkList list, LinkNode np)
     colonmod = NULL;
     gf_nullglob = isset(NULLGLOB);
     gf_markdirs = isset(MARKDIRS);
-    gf_listtypes = 0;
+    gf_listtypes = gf_follow = 0;
     gf_noglobdots = unset(GLOBDOTS);
-    if (str[sl - 1] == Outpar) {	/* check for qualifiers */
-	char *s;
-	int sense = 0;			/* bit 0 for match (0)/don't match (1)   */
-					/* bit 1 for follow links (2), don't (0) */
-	long data = 0;			/* Any numerical argument required       */
 
-	int (*func) _((Statptr, long));
+    /* Check for qualifiers */
+    if (isset(BAREGLOBQUAL) && str[sl - 1] == Outpar) {
+	char *s;
 
 	/* Check these are really qualifiers, not a set of *
 	 * alternatives or exclusions                      */
-	for (s = str + sl - 2; s != str; s--)
-	    if (*s == Bar || *s == Outpar || *s == Inpar
-		|| (isset(EXTENDEDGLOB) && *s == Tilde))
+	for (s = str + sl - 2; *s != Inpar; s--)
+	    if (*s == Bar || *s == Outpar ||
+		(isset(EXTENDEDGLOB) && *s == Tilde))
 		break;
 	if (*s == Inpar) {
 	    /* Real qualifiers found. */
-	    *s++ = '\0';
-	    while (*s != Outpar && !colonmod) {
+	    int sense = 0;	/* bit 0 for match (0)/don't match (1)   */
+				/* bit 1 for follow links (2), don't (0) */
+	    long data = 0;	/* Any numerical argument required       */
+	    int (*func) _((Statptr, long));
+
+	    str[sl-1] = 0;
+	    *s++ = 0;
+	    while (*s && !colonmod) {
 		func = (int (*) _((Statptr, long)))0;
 		if (idigit(*s)) {
 		    /* Store numeric argument for qualifier */
@@ -838,6 +919,7 @@ glob(LinkList list, LinkNode np)
 			/* Remaining arguments are history-type     *
 			 * colon substitutions, handled separately. */
 			colonmod = s - 1;
+			untokenize(colonmod);
 			break;
 		    case Hat:
 		    case '^':
@@ -849,37 +931,26 @@ glob(LinkList list, LinkNode np)
 			/* Toggle matching of symbolic links */
 			sense ^= 2;
 			break;
-#ifdef S_IFLNK
 		    case '@':
 			/* Match symbolic links */
-			func = qualmode;
-			data = S_IFLNK;
+			func = qualislnk;
 			break;
-#endif
-#ifdef S_IFSOCK
 		    case Equals:
 		    case '=':
 			/* Match sockets */
-			func = qualmode;
-			data = S_IFSOCK;
+			func = qualissock;
 			break;
-#endif
-#ifdef S_IFIFO
 		    case 'p':
 			/* Match named pipes */
-			func = qualmode;
-			data = S_IFIFO;
+			func = qualisfifo;
 			break;
-#endif
 		    case '/':
 			/* Match directories */
-			func = qualmode;
-			data = S_IFDIR;
+			func = qualisdir;
 			break;
 		    case '.':
 			/* Match regular files */
-			func = qualmode;
-			data = S_IFREG;
+			func = qualisreg;
 			break;
 		    case '%':
 			/* Match special files: block, *
@@ -887,7 +958,7 @@ glob(LinkList list, LinkNode np)
 			if (*s == 'b')
 			    s++, func = qualisblk;
 			else if (*s == 'c')
-			    s++, func = qualischar;
+			    s++, func = qualischr;
 			else
 			    func = qualisdev;
 			break;
@@ -999,7 +1070,7 @@ glob(LinkList list, LinkNode np)
 				    zerr("unknown user", NULL, 0);
 				    data = 0;
 				}
-				if ((*tt = sav) != Outpar)
+				if ((*tt = sav))
 				    s = tt + 1;
 				else
 				    s = tt;
@@ -1032,7 +1103,7 @@ glob(LinkList list, LinkNode np)
 				    zerr("unknown group", NULL, 0);
 				    data = 0;
 				}
-				if ((*tt = sav) != Outpar)
+				if ((*tt = sav))
 				    s = tt + 1;
 				else
 				    s = tt;
@@ -1047,11 +1118,13 @@ glob(LinkList list, LinkNode np)
 			break;
 		    case 'M':
 			/* Mark directories with a / */
-			gf_markdirs = !(sense & 1);
+			if ((gf_markdirs = !(sense & 1)))
+			    gf_follow = sense & 2;
 			break;
 		    case 'T':
 			/* Mark types in a `ls -F' type fashion */
-			gf_listtypes = !(sense & 1);
+			if ((gf_listtypes = !(sense & 1)))
+			    gf_follow = sense & 2;
 			break;
 		    case 'N':
 			/* Nullglob:  remove unmatched patterns. */
@@ -1236,31 +1309,22 @@ notstrcmp(char **a, char **b)
 char
 file_type(mode_t filemode)
 {
-    switch (filemode & S_IFMT) {/* screw POSIX */
-    case S_IFDIR:
-	return '/';
-#ifdef S_IFIFO
-    case S_IFIFO:
-	return '|';
-#endif
-    case S_IFCHR:
-	return '%';
-    case S_IFBLK:
+    if(S_ISBLK(filemode))
 	return '#';
-#ifdef S_IFLNK
-    case S_IFLNK:
-	return /* (access(pbuf, F_OK) == -1) ? '&' :*/ '@';
-#endif
-#ifdef S_IFSOCK
-    case S_IFSOCK:
+    else if(S_ISCHR(filemode))
+	return '%';
+    else if(S_ISDIR(filemode))
+	return '/';
+    else if(S_ISFIFO(filemode))
+	return '|';
+    else if(S_ISLNK(filemode))
+	return '@';
+    else if(S_ISREG(filemode))
+	return (filemode & S_IXUGO) ? '*' : ' ';
+    else if(S_ISSOCK(filemode))
 	return '=';
-#endif
-    default:
-	if (filemode & 0111)
-	    return '*';
-	else
-	    return ' ';
-    }
+    else
+	return '?';
 }
 
 /* check to see if str is eligible for brace expansion */
@@ -1603,23 +1667,13 @@ xpandbraces(LinkList list, LinkNode *np)
 int
 matchpat(char *a, char *b)
 {
-    Comp c;
-    int val, len;
-    char *b2;
+    Comp c = parsereg(b);
 
-    len = strlen(b);
-    b2 = (char *)alloc(len + 3);
-    strcpy(b2 + 1, b);
-    b2[0] = Inpar;
-    b2[len + 1] = Outpar;
-    b2[len + 2] = '\0';
-    c = parsereg(b2);
     if (!c) {
 	zerr("bad pattern: %s", b, 0);
 	return 0;
     }
-    val = domatch(a, c, 0);
-    return val;
+    return domatch(a, c, 0);
 }
 
 /* do the ${foo%%bar}, ${foo#bar} stuff */
@@ -1908,11 +1962,15 @@ getmatch(char **sp, char *pat, int fl, int n)
 int
 domatch(char *str, Comp c, int fist)
 {
+    int ret;
     pptr = str;
     first = fist;
     if (*pptr == Nularg)
 	pptr++;
-    return doesmatch(c);
+    PERMALLOC {
+	ret = doesmatch(c);
+    } LASTALLOC;
+    return ret;
 }
 
 #define untok(C)  (itok(C) ? ztokens[(C) - Pound] : (C))
@@ -1938,6 +1996,8 @@ excluded(Comp c, char *eptr)
 	pptr = eptr;
 	ret = doesmatch(c->exclude);
     }
+    if (*pptr)
+	ret = 0;
 
     pptr = saves;
     first = savei;
@@ -1945,28 +2005,194 @@ excluded(Comp c, char *eptr)
     return ret;
 }
 
+struct gclose {
+    char *start;
+    char *end;
+};
+typedef struct gclose *Gclose;
+
+static int inclosure;		/* see comment in doesmatch() */
+
+/* Add a list of matches that fit the closure.  trystring is a string of
+ * the same length as the target string; a non-zero in that string
+ * indicates that we have already tried to match the patterns following
+ * the closure (c->next) at that point and failed.  This means that not
+ * only should we not bother using the corresponding match, we should
+ * also not bother going any further, since the first time we got to
+ * that position (when it was marked), we must already have failed on
+ * and backtracked over any further closure matches beyond that point.
+ */
+
+/**/
+static void
+addclosures(Comp c, LinkList closlist, int *pdone, char *trystring)
+{
+    Gclose gcnode;
+    char *opptr = pptr;
+
+    while (*pptr) {
+	if (STARP(c)) {
+	    if (trystring[(pptr+1)-opptr])
+		break;
+	    gcnode = (Gclose)zalloc(sizeof(struct gclose));
+	    gcnode->start = pptr;
+	    gcnode->end = ++pptr;
+	} else {
+	    char *saves = pptr;
+	    if (OPTIONALP(c) && *pdone >= 1)
+		return;
+	    if (!matchonce(c) || saves == pptr ||
+		trystring[pptr-opptr]) {
+		pptr = saves;
+		break;
+	    }
+	    gcnode = (Gclose)zalloc(sizeof(struct gclose));
+	    gcnode->start = saves;
+	    gcnode->end = pptr;
+	}
+	pushnode(closlist, gcnode);
+	(*pdone)++;
+    }
+}
+
 /* see if current string in pptr matches c */
 
 /**/
 static int
 doesmatch(Comp c)
 {
-    char *pat = c->str;
-    int done = 0;
-
-  tailrec:
-    if (ONEHASHP(c) || (done && TWOHASHP(c))) {
-	/* Do multiple matches like (pat)# and (pat)## */
-	char *saves = pptr;
+    if (CLOSUREP(c)) {
+	int done, retflag = 0;
+	char *saves, *trystring, *opptr;
+	LinkList closlist;
+	Gclose gcnode;
 
 	if (first && *pptr == '.')
 	    return 0;
-	if (doesmatch(c->next))
-	    return 1;
-	pptr = saves;
-	first = 0;
-    }
-    done++;
+
+	if (!inclosure && !c->left) {
+	    /* We are not inside another closure, and the current
+	     * pattern is a simple string.  We handle this very common
+	     * case specially: otherwise, matches like *foo* are
+	     * extremely slow.  Here, instead of backtracking, we track
+	     * forward until we get a match.  At top level, we are bound
+	     * to get there eventually, so this is OK.
+	     */
+	    char looka;
+
+	    if (STARP(c) && c->next &&
+		!c->next->left && (looka = *c->next->str) &&
+		!itok(looka)) {
+		/* Another simple optimisation for a very common case:
+		 * we are processing a * and there is
+		 * an ordinary character match next.  We look ahead for
+		 * that character, taking care of Meta bytes.
+		 */
+		while (*pptr) {
+		    for (; *pptr; pptr++) {
+			if (*pptr == Meta)
+			    pptr++;
+			else if (*pptr == looka)
+			    break;
+		    }
+		    if (!*(saves = pptr))
+			break;
+		    if (doesmatch(c->next))
+			return 1;
+		    pptr = saves+1;
+		}
+	    } else {
+		/* Standard track-forward code */
+		for (done = 0; ; done++) {
+		    saves = pptr;
+		    if ((done || ONEHASHP(c) || OPTIONALP(c)) &&
+			((!c->next && (!LASTP(c) || !*pptr)) ||
+			 (c->next && doesmatch(c->next))))
+			return 1;
+		    if (done && OPTIONALP(c))
+			return 0;
+		    pptr = saves;
+		    first = 0;
+		    if (STARP(c)) {
+			if (!*pptr)
+			    return 0;
+			pptr++;
+		    } else if (!matchonce(c) || pptr == saves)
+			return 0;
+		}
+	    }
+	    return 0;
+	}
+	/* The full, gory backtracking code is now necessary. */
+	inclosure++;
+	closlist = newlinklist();
+	trystring = zcalloc(strlen(pptr)+1);
+	opptr = pptr;
+
+	/* Start by making a list where each match is as long
+	 * as possible.  We later have to take account of the
+	 * fact that earlier matches may be too long.
+	 */
+	done = 0;
+	addclosures(c, closlist, &done, trystring);
+	for (;;) {
+	    if (TWOHASHP(c) && !done)
+		break;
+	    saves = pptr;
+	    /* do we really want this LASTP here?? */
+	    if ((!c->next && (!LASTP(c) || !*pptr)) ||
+		(c->next && doesmatch(c->next))) {
+		retflag = 1;
+		break;
+	    }
+	    trystring[saves-opptr] = 1;
+	    /*
+	     * If we failed, the first thing to try is whether we can
+	     * shorten the match using the last pattern in the closure.
+	     */
+	    gcnode = firstnode(closlist) ? peekfirst(closlist) : NULL;
+	    if (gcnode && --gcnode->end > gcnode->start
+		&& (gcnode->end[-1] != Meta ||
+		    --gcnode->end > gcnode->start)) {
+		char savec = *gcnode->end;
+		*gcnode->end = '\0';
+		pptr = gcnode->start;
+		if (matchonce(c) && pptr != gcnode->start
+		    && !trystring[pptr-opptr]) {
+		    *gcnode->end = savec;
+		    gcnode->end = pptr;
+		    /* Try again to construct a list based on
+		     * this new position
+		     */
+		    addclosures(c, closlist, &done, trystring+(pptr-opptr));
+		    continue;
+		}
+		*gcnode->end = savec;
+	    }
+	    /* We've now exhausted the possibilities with that match,
+	     * backtrack to the previous.
+	     */
+	    if ((gcnode = (Gclose)getlinknode(closlist))) {
+		pptr = gcnode->start;
+		zfree(gcnode, sizeof(struct gclose));
+		done--;
+	    } else
+		break;
+	}
+	freelinklist(closlist, free);
+	zfree(trystring, strlen(opptr)+1);
+	inclosure--;
+
+	return retflag;
+    } else
+	return matchonce(c);
+}
+
+/**/
+static int
+matchonce(Comp c)
+{
+    char *pat = c->str;
     for (;;) {
 	/* loop until success or failure of pattern */
 	if (!pat || !*pat) {
@@ -1982,27 +2208,80 @@ doesmatch(Comp c)
 	    savei = first;
 	    /* Loop over alternatives with exclusions: (foo~bar|...). *
 	     * Exclusions apply to the pattern in c->left.            */
-	    if (c->left || c->right)
-		if (!doesmatch(c->left) ||
-		    (c->exclude && excluded(c, saves)))
-		    if (c->right) {
+	    if (c->left || c->right) {
+		int ret = 0, ret2 = 0;
+		if (c->exclude) {
+		    char *exclend = 0;
+
+		    /* We may need to back up on things like `(*~foo)'
+		     * if the `*' matched `foo' but needs to match `fo'.
+		     * exclend marks the end of the shortened text.  We
+		     * need to restore it to match the tail.
+		     * We set `inclosure' because we need the more
+		     * sophisticated code in doesmatch() for any nested
+		     * closures.
+		     */
+		    inclosure++;
+
+		    while (!exclend || exclend >= pptr) {
+			char exclsav = 0;
+			if (exclend) {
+			     exclsav = *exclend;
+			    *exclend = '\0';
+			 }
+			if ((ret = doesmatch(c->left))) {
+			    if (exclend)
+				*exclend = exclsav;
+			    exclsav = *(exclend = pptr);
+			    *exclend = '\0';
+			    ret2 = !excluded(c, saves);
+			}
+			if (exclend)
+			    *exclend = exclsav;
+
+			if (!ret)
+			    break;
+			if ((ret = ret2 &&
+			     ((!c->next && (!LASTP(c) || !*pptr))
+			      || (c->next && doesmatch(c->next)))) ||
+			    (!c->next && LASTP(c)))
+			    break;
+			/* Back up if necessary: exclend gives the position
+			 * of the end of the match we are excluding,
+			 * so only try to match to there.
+			 */
+			exclend--;
 			pptr = saves;
-			first = savei;
-			if (!doesmatch(c->right))
-			    return 0;
-		    } else
-			return 0;
-	    if (*pptr && CLOSUREP(c)) {
-		/* With a closure (#), need to keep trying */
-		pat = c->str;
-		goto tailrec;
+		    }
+		    inclosure--;
+		    if (ret)
+			return 1;
+		} else
+		    ret = doesmatch(c->left);
+		ret2 = 0;
+		if (c->right && (!ret || inclosure)) {
+		    /* If in a closure, we always want the longest match. */
+		    char *newpptr = pptr;
+		    pptr = saves;
+		    first = savei;
+		    ret2 = doesmatch(c->right);
+		    if (ret && (!ret2 || pptr < newpptr))
+			pptr = newpptr;
+		}
+		if (!ret && !ret2)
+		    return 0;
 	    }
+	    if (CLOSUREP(c))
+		return 1;
 	    if (!c->next)	/* no more patterns left */
 		return (!LASTP(c) || !*pptr);
-	    c = c->next;
-	    done = 0;
-	    pat = c->str;
-	    goto tailrec;
+	    /* optimisation when next pattern is not a closure */
+	    if (!CLOSUREP(c->next)) {
+		c = c->next;
+		pat = c->str;
+		continue;
+	    }
+	    return doesmatch(c->next);
 	}
 	/* Don't match leading dot if first is set */
 	if (first && *pptr == '.' && *pat != '.')
@@ -2021,8 +2300,6 @@ doesmatch(Comp c)
 	    pat++;
 	    continue;
 	}
-	if (*pat == Hat)	/* following pattern is negated */
-	    return 1 - doesmatch(c->next);
 	if (*pat == Inbrack) {
 	    /* Match groups of characters */
 #define PAT(X) (pat[X] == Meta ? pat[(X)+1] ^ 32 : untok(pat[X]))
@@ -2104,6 +2381,7 @@ doesmatch(Comp c)
 	    } else {
 		/* Flag that there is no upper limit */
 		int not3 = 0;
+		char *opptr = pptr;
 		/*
 		 * Form is <a-b>, where a or b are numbers or blank.
 		 * t1 = number supplied:  must be positive, so use
@@ -2123,6 +2401,15 @@ doesmatch(Comp c)
 		    t3 = (unsigned long)zstrtol(ptr + 1, &pat, 10);
 		DPUTS(*pat != Outang, "BUG: wrong internal range pattern");
 		pat++;
+		/*
+		 * If the number found is too large for the pattern,
+		 * try matching just the first part.  This way
+		 * we always get the longest possible match.
+		 */
+		while (!not3 && t1 > t3 && pptr > opptr+1) {
+		  pptr--;
+		  t1 /= 10;
+		}
 		if (t1 < t2 || (!not3 && t1 > t3))
 		    break;
 	    }
@@ -2148,7 +2435,7 @@ parsereg(char *str)
     remnulargs(str);
     mode = 1;			/* no path components */
     pptr = str;
-    return parsecompsw(0);
+    return parsecompsw(GF_TOPLEV);
 }
 
 /* blindly turn a string into a tokenised expression without lexing */
@@ -2293,8 +2580,7 @@ qualgid(struct stat *buf, long gid)
 static int
 qualisdev(struct stat *buf, long junk)
 {
-    junk = buf->st_mode & S_IFMT;
-    return junk == S_IFBLK || junk == S_IFCHR;
+    return S_ISBLK(buf->st_mode) || S_ISCHR(buf->st_mode);
 }
 
 /* block special file? */
@@ -2303,27 +2589,61 @@ qualisdev(struct stat *buf, long junk)
 static int
 qualisblk(struct stat *buf, long junk)
 {
-    junk = buf->st_mode & S_IFMT;
-    return junk == S_IFBLK;
+    return S_ISBLK(buf->st_mode);
 }
 
 /* character special file? */
 
 /**/
 static int
-qualischar(struct stat *buf, long junk)
+qualischr(struct stat *buf, long junk)
 {
-    junk = buf->st_mode & S_IFMT;
-    return junk == S_IFCHR;
+    return S_ISCHR(buf->st_mode);
 }
 
-/* file type is requested one */
+/* directory? */
 
 /**/
 static int
-qualmode(struct stat *buf, long mod)
+qualisdir(struct stat *buf, long junk)
 {
-    return (buf->st_mode & S_IFMT) == mod;
+    return S_ISDIR(buf->st_mode);
+}
+
+/* FIFO? */
+
+/**/
+static int
+qualisfifo(struct stat *buf, long junk)
+{
+    return S_ISFIFO(buf->st_mode);
+}
+
+/* symbolic link? */
+
+/**/
+static int
+qualislnk(struct stat *buf, long junk)
+{
+    return S_ISLNK(buf->st_mode);
+}
+
+/* regular file? */
+
+/**/
+static int
+qualisreg(struct stat *buf, long junk)
+{
+    return S_ISREG(buf->st_mode);
+}
+
+/* socket? */
+
+/**/
+static int
+qualissock(struct stat *buf, long junk)
+{
+    return S_ISSOCK(buf->st_mode);
 }
 
 /* given flag is set in mode */
@@ -2332,7 +2652,7 @@ qualmode(struct stat *buf, long mod)
 static int
 qualflags(struct stat *buf, long mod)
 {
-    return buf->st_mode & mod;
+    return mode_to_octal(buf->st_mode) & mod;
 }
 
 /* mode matches number supplied exactly  */
@@ -2341,7 +2661,7 @@ qualflags(struct stat *buf, long mod)
 static int
 qualeqflags(struct stat *buf, long mod)
 {
-    return (buf->st_mode & 07777) == mod;
+    return mode_to_octal(buf->st_mode) == mod;
 }
 
 /* regular executable file? */
@@ -2350,7 +2670,7 @@ qualeqflags(struct stat *buf, long mod)
 static int
 qualiscom(struct stat *buf, long mod)
 {
-    return (buf->st_mode & (S_IFMT | S_IEXEC)) == (S_IFREG | S_IEXEC);
+    return S_ISREG(buf->st_mode) && (buf->st_mode & S_IXUGO);
 }
 
 /* size in required range? */
