@@ -1,6 +1,4 @@
 /*
- * $Id: hist.c,v 3.1.2.7 1997/06/02 06:52:04 hzoli Exp $
- *
  * hist.c - history expansion
  *
  * This file is part of zsh, the Z shell.
@@ -29,10 +27,32 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "hist.pro"
 
+/* != 0 means history substitution is turned off */
+ 
+/**/
+int stophist;
+ 
+/* this line began with a space, so junk it if HISTIGNORESPACE is on */
+ 
+/**/
+int spaceflag;
+ 
+/* if != 0, we are expanding the current line */
+
+/**/
+int expanding;
+
+/* these are used to modify the cursor position during expansion */
+
+/**/
+int excs, exlast;
+
 /*
+ * Current history event number
+ *
  * Note on curhist: with history inactive, this points to the
  * last line actually added to the history list.  With history active,
  * the line does not get added to the list until hend(), if at all.
@@ -40,6 +60,35 @@
  * Thus if the line is not added to the list, curhist must be
  * decremented in hend().
  */
+ 
+/**/
+int curhist;
+
+/* number of history entries */
+ 
+/**/
+int histentct;
+ 
+/* array of history entries */
+ 
+/**/
+Histent histentarr;
+ 
+/* capacity of history lists */
+ 
+/**/
+int histsiz;
+ 
+/* if = 1, we have performed history substitution on the current line *
+ * if = 2, we have used the 'p' modifier                              */
+ 
+/**/
+int histdone;
+ 
+/* state of the history mechanism */
+ 
+/**/
+int histactive;
 
 /* Bits of histactive variable */
 #define HA_ACTIVE	(1<<0)	/* History mechanism is active */
@@ -47,15 +96,53 @@
 #define HA_JUNKED	(1<<2)	/* Last history line was already junked */
 #define HA_NOINC	(1<<3)	/* Don't store, curhist not incremented */
 
-extern int cs, ll;
-
 /* Array of word beginnings and endings in current history line. */
+
+/**/
 short *chwords;
+
 /* Max, actual position in chwords.
  * nwords = chwordpos/2 because we record beginning and end of words.
  */
+
+/**/
 int chwordlen, chwordpos;
 
+/* the last l for s/l/r/ history substitution */
+ 
+/**/
+char *hsubl;
+
+/* the last r for s/l/r/ history substitution */
+ 
+/**/
+char *hsubr;
+ 
+/* pointer into the history line */
+ 
+/**/
+char *hptr;
+ 
+/* the current history line */
+ 
+/**/
+char *chline;
+
+/* true if the last character returned by hgetc was an escaped bangchar *
+ * if it is set and NOBANGHIST is unset hwaddc escapes bangchars        */
+
+/**/
+int qbang;
+ 
+/* max size of histline */
+ 
+/**/
+int hlinesz;
+ 
+/* default event (usually curhist-1, that is, "!!") */
+ 
+static int defev;
+ 
 /* add a character to the current history word */
 
 /**/
@@ -565,6 +652,8 @@ hbegin(void)
     chwords = zalloc((chwordlen = 16)*sizeof(short));
     chwordpos = 0;
 
+    if (histactive & HA_JUNKED)
+	curhist--;
     curhistent = gethistent(curhist);
     if (!curhistent->ftim)
 	curhistent->ftim = time(NULL);
@@ -744,13 +833,15 @@ remhist(void)
 	    zsfree(he->text);
 	    he->text = NULL;
 	    histactive |= HA_JUNKED;
-	    curhist--;
+	    /* curhist-- is delayed until the next hbegin() */
 	}
     } else
 	histactive |= HA_NOSTORE;
 }
 
 /* Gives current expansion word if not last word before chwordpos. */
+
+/**/
 int hwgetword = -1;
 
 /* begin a word */
@@ -1481,7 +1572,7 @@ savehistfile(char *s, int err, int app)
     else
 	out = fdopen(open(unmeta(s), O_CREAT | O_WRONLY | O_TRUNC, 0600), "w");
     if (out) {
-	for (; ev <= curhist - !!histactive; ev++) {
+	for (; ev <= curhist - !!(histactive & HA_ACTIVE); ev++) {
 	    ent = gethistent(ev);
 	    if (app & 2) {
 		if (ent->flags & HIST_OLD)
