@@ -1,6 +1,4 @@
 /*
- * $Id: init.c,v 3.1.2.13 1997/06/01 06:13:15 hzoli Exp $
- *
  * init.c - main loop and initialization routines
  *
  * This file is part of zsh, the Z shell.
@@ -29,20 +27,57 @@
  *
  */
 
-#define GLOBALS
-#include "zsh.h"
+#include "zsh.mdh"
+#include "init.pro"
 
 #include "zshpaths.h"
 #include "zshxmods.h"
-#include "init.pro"
 
+/**/
 int noexitct = 0;
 
+/* what level of sourcing we are at */
+ 
+/**/
+int sourcelevel;
+
+/* the shell tty fd */
+
+/**/
+int SHTTY;
+
+/* the FILE attached to the shell tty */
+
+/**/
+FILE *shout;
+
+/* termcap strings */
+ 
+/**/
+char *tcstr[TC_COUNT];
+
+/* lengths of each termcap string */
+ 
+/**/
+int tclen[TC_COUNT];
+
+/* Values of the li, co and am entries */
+
+/**/
+int tclines, tccolumns, hasam;
+
+#ifdef DEBUG
+/* depth of allocation type stack */
+
+/**/
+int alloc_stackp;
+#endif
+
 /* keep executing lists until EOF found */
 
 /**/
 void
-loop(int toplevel)
+loop(int toplevel, int justonce)
 {
     List list;
 #ifdef DEBUG
@@ -60,13 +95,27 @@ loop(int toplevel)
 	lexinit();              /* initialize lexical state */
 	if (!(list = parse_event())) {	/* if we couldn't parse a list */
 	    hend();
-	    if (tok == ENDINPUT && !errflag)
+	    if ((tok == ENDINPUT && !errflag) || justonce)
 		break;
 	    continue;
 	}
 	if (hend()) {
 	    int toksav = tok;
+	    List prelist;
 
+	    if (toplevel && (prelist = getshfunc("preexec")) != &dummy_list) {
+		Histent he = gethistent(curhist);
+		LinkList args;
+		PERMALLOC {
+		    args = newlinklist();
+		    addlinknode(args, "preexec");
+		    if (he && he->text)
+			addlinknode(args, he->text);
+		} LASTALLOC;
+		doshfunc(prelist, args, 0, 1);
+		freelinklist(args, (FreeFunc) NULL);
+		errflag = 0;
+	    }
 	    if (stopmsg)	/* unset 'you have stopped jobs' flag */
 		stopmsg--;
 	    execlist(list, 0, 0);
@@ -92,6 +141,8 @@ loop(int toplevel)
 		dotrap(SIGEXIT);
 	    exit(lastval);
 	}
+	if (justonce)
+	    break;
     }
     popheap();
 }
@@ -308,6 +359,14 @@ init_shout(void)
 #endif
 }
 
+/* names of the termcap strings we want */
+
+static char *tccapnams[TC_COUNT] = {
+    "cl", "le", "LE", "nd", "RI", "up", "UP", "do",
+    "DO", "dc", "DC", "ic", "IC", "cd", "ce", "al", "dl", "ta",
+    "md", "so", "us", "me", "se", "ue"
+};
+
 /* Initialise termcap */
 
 /**/
@@ -449,12 +508,15 @@ setupvals(void)
 #endif
 
     /* Set default prompts */
-    if (opts[INTERACTIVE]) {
-	prompt  = ztrdup("%m%# ");
-	prompt2 = ztrdup("%_> ");
-    } else {
+    if(unset(INTERACTIVE)) {
 	prompt = ztrdup("");
 	prompt2 = ztrdup("");
+    } else if (emulation == EMULATE_KSH || emulation == EMULATE_SH) {
+	prompt  = ztrdup(privasserted() ? "# " : "$ ");
+	prompt2 = ztrdup("> ");
+    } else {
+	prompt  = ztrdup("%m%# ");
+	prompt2 = ztrdup("%_> ");
     }
     prompt3 = ztrdup("?# ");
     prompt4 = ztrdup("+ ");
@@ -719,7 +781,7 @@ source(char *s)
     scriptname = s;
 
     sourcelevel++;
-    loop(0);			/* loop through the file to be sourced        */
+    loop(0, 0);			/* loop through the file to be sourced        */
     sourcelevel--;
     fclose(bshin);
     fdtable[SHIN] = 0;
@@ -775,10 +837,15 @@ init_bltinmods(void)
 
 #ifdef LINKED_XMOD_zle
 
+/**/
 ZleVoidFn trashzleptr;
+/**/
 ZleVoidFn gotwordptr;
+/**/
 ZleVoidFn refreshptr;
+/**/
 ZleVoidIntFn spaceinlineptr;
+/**/
 ZleReadFn zlereadptr;
 
 #else /* !LINKED_XMOD_zle */
@@ -827,7 +894,7 @@ fallback_zleread(char *lp, char *rp, int ha)
     char *pptbuf;
     int pptlen;
 
-    pptbuf = putprompt(lp, &pptlen, NULL, 1);
+    pptbuf = unmetafy(promptexpand(lp, 0, NULL, NULL), &pptlen);
     write(2, (WRITE_ARG_2_T)pptbuf, pptlen);
     free(pptbuf);
     return (unsigned char *)shingetline();
@@ -839,6 +906,7 @@ fallback_zleread(char *lp, char *rp, int ha)
 
 #ifdef LINKED_XMOD_comp1
 
+/**/
 CompctlReadFn compctlreadptr;
 
 #else /* !LINKED_XMOD_comp1 */
