@@ -1,6 +1,4 @@
 /*
- * $Id: input.c,v 3.1.2.2 1997/06/01 06:13:15 hzoli Exp $
- *
  * input.c - read and store lines of input
  *
  * This file is part of zsh, the Z shell.
@@ -69,10 +67,34 @@
  * PWS 1996/12/10
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "input.pro"
 
-/* Input buffer variables:  inbufct and inbufflags are in globals.h. */
+/* the shell input fd */
+
+/**/
+int SHIN;
+
+/* buffered shell input for non-interactive shells */
+
+/**/
+FILE *bshin;
+
+/* != 0 means we are reading input from a string */
+ 
+/**/
+int strin;
+ 
+/* total # of characters waiting to be read. */
+
+/**/
+int inbufct;
+
+/* the flags controlling the input routines in input.c: see INP_* in zsh.h */
+
+/**/
+int inbufflags;
+
 static char *inbuf;		/* Current input buffer */
 static char *inbufptr;		/* Pointer into input buffer */
 static char *inbufpush;		/* Character at which to re-push alias */
@@ -194,16 +216,16 @@ ingetc(void)
 static int
 inputline(void)
 {
-    unsigned char *ingetcline, *ingetcpmptl = NULL, *ingetcpmptr = NULL;
+    char *ingetcline, *ingetcpmptl = NULL, *ingetcpmptr = NULL;
 
     /* If reading code interactively, work out the prompts. */
     if (interact && isset(SHINSTDIN))
 	if (!isfirstln)
-	    ingetcpmptl = (unsigned char *)prompt2;
+	    ingetcpmptl = prompt2;
 	else {
-	    ingetcpmptl = (unsigned char *)prompt;
+	    ingetcpmptl = prompt;
 	    if (rprompt)
-		ingetcpmptr = (unsigned char *)rprompt;
+		ingetcpmptr = rprompt;
 	}
     if (!(interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE))) {
 	/*
@@ -221,13 +243,13 @@ inputline(void)
 	     */
 	    char *pptbuf;
 	    int pptlen;
-	    pptbuf = putprompt((char *)ingetcpmptl, &pptlen, NULL, 1);
+	    pptbuf = unmetafy(promptexpand(ingetcpmptl, 0, NULL, NULL), &pptlen);
 	    write(2, (WRITE_ARG_2_T)pptbuf, pptlen);
 	    free(pptbuf);
 	}
-	ingetcline = (unsigned char *)shingetline();
+	ingetcline = shingetline();
     } else
-	ingetcline = zleread((char *)ingetcpmptl, (char *)ingetcpmptr, 1);
+	ingetcline = (char *)zleread(ingetcpmptl, ingetcpmptr, 1);
     if (!ingetcline) {
 	return lexstop = 1;
     }
@@ -240,18 +262,18 @@ inputline(void)
 	spaceflag = *ingetcline == ' ';
     if (isset(VERBOSE)) {
 	/* Output the whole line read so far. */
-	zputs((char *)ingetcline, stderr);
+	zputs(ingetcline, stderr);
 	fflush(stderr);
     }
-    if (*ingetcline && ingetcline[strlen((char *)ingetcline) - 1] == '\n') {
+    if (*ingetcline && ingetcline[strlen(ingetcline) - 1] == '\n') {
 	/* We've now read a complete line. */
 	lineno++;
 	if (interact && isset(SUNKEYBOARDHACK) && isset(SHINSTDIN) &&
 	    SHTTY != -1 && *ingetcline && ingetcline[1] &&
-	    ingetcline[strlen((char *)ingetcline) - 2] == '`') {
+	    ingetcline[strlen(ingetcline) - 2] == '`') {
 	    /* Junk an unmatched "`" at the end of the line. */
 	    int ct;
-	    unsigned char *ptr;
+	    char *ptr;
 
 	    for (ct = 0, ptr = ingetcline; *ptr; ptr++)
 		if (*ptr == '`')
@@ -264,7 +286,7 @@ inputline(void)
     }
     isfirstch = 1;
     /* Put this into the input channel. */
-    inputsetline((char *)ingetcline, INP_FREE);
+    inputsetline(ingetcline, INP_FREE);
 
     return 0;
 }
