@@ -1,6 +1,4 @@
 /*
- * $Id: jobs.c,v 3.1.2.7 1997/06/03 03:52:01 hzoli Exp $
- *
  * jobs.c - job control
  *
  * This file is part of zsh, the Z shell.
@@ -29,14 +27,49 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "jobs.pro"
 
+/* the process group of the shell */
+
+/**/
+pid_t mypgrp;
+ 
+/* the job we are working on */
+ 
+/**/
+int thisjob;
+
+/* the current job (+) */
+ 
+/**/
+int curjob;
+ 
+/* the previous job (-) */
+ 
+/**/
+int prevjob;
+ 
+/* the job table */
+ 
+/**/
+struct job jobtab[MAXJOB];
+ 
+/* shell timings */
+ 
+/**/
+struct tms shtms;
+ 
+/* 1 if ttyctl -f has been executed */
+ 
+/**/
+int ttyfrozen;
+ 
 /* empty job structure for quick clearing of jobtab entries */
 
 static struct job zero;		/* static variables are initialized to zero */
 
-struct timeval dtimeval, now;
+static struct timeval dtimeval, now;
 
 /* Diff two timevals for elapsed-time computations */
 
@@ -155,7 +188,8 @@ update_job(Job jn)
 	}
     }
 
-    if (shout && !ttyfrozen && !jn->stty_in_env)
+    if (shout && !ttyfrozen && !jn->stty_in_env && !zleactive &&
+	job == thisjob && !somestopped && !(jn->stat & STAT_NOSTTY))
 	gettyinfo(&shttyinfo);
 
     if (isset(MONITOR)) {
@@ -573,7 +607,9 @@ deletejob(Job jn)
 {
     struct process *pn, *nx;
 
-    for (pn = jn->procs; pn; pn = nx) {
+    pn = jn->procs;
+    jn->procs = NULL;
+    for (; pn; pn = nx) {
 	nx = pn->next;
 	zfree(pn, sizeof(struct process));
     }
@@ -1088,6 +1124,8 @@ bin_fg(char *name, char **argv, char *ops, int func)
 	case BIN_FG:
 	case BIN_BG:
 	case BIN_WAIT:
+	    if (func == BIN_BG)
+		jobtab[job].stat |= STAT_NOSTTY;
 	    if ((stopped = (jobtab[job].stat & STAT_STOPPED)))
 		makerunning(jobtab + job);
 	    else if (func == BIN_BG) {
@@ -1139,12 +1177,8 @@ bin_fg(char *name, char **argv, char *ops, int func)
 	    printjob(job + jobtab, lng, 2);
 	    break;
 	case BIN_DISOWN:
-	    {
-		static struct job zero;
-
-		jobtab[job] = zero;
-		break;
-	    }
+	    deletejob(jobtab + job);
+	    break;
 	}
 	thisjob = ocj;
     }
