@@ -1,6 +1,4 @@
 /*
- * $Id: lex.c,v 3.1.2.7 1997/06/01 06:13:15 hzoli Exp $
- *
  * lex.c - lexical analysis
  *
  * This file is part of zsh, the Z shell.
@@ -29,15 +27,128 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "lex.pro"
 
+/* tokens */
+
+/**/
+char ztokens[] = "#$^*()$=|{}[]`<>?~`,'\"\\";
+
+/* parts of the current token */
+
+/**/
+char *yytext, *tokstr;
+/**/
+int tok, tokfd;
+
+/* lexical analyzer error flag */
+ 
+/**/
+int lexstop;
+
+/* if != 0, this is the first line of the command */
+ 
+/**/
+int isfirstln;
+ 
+/* if != 0, this is the first char of the command (not including white space) */
+ 
+/**/
+int isfirstch;
+
+/* flag that an alias should be expanded after expansion ending in space */
+
+/**/
+int inalmore;
+
+/* don't do spelling correction */
+ 
+/**/
+int nocorrect;
+
+/* the line buffer */
+
+/**/
+unsigned char *line;
+
+/* cursor position and line length */
+
+/**/
+int cs, ll;
+
+/* inwhat says what exactly we are in     *
+ * (its value is one of the IN_* things). */
+
+/**/
+int inwhat;
+
+/* 1 if x added to complete in a blank between words */
+
+/**/
+int addedx;
+
+/* 1 if aliases should not be expanded */
+ 
+/**/
+int noaliases;
+
+/* we are parsing a line sent to use by the editor */
+ 
+/**/
+int zleparse;
+ 
+/**/
+int wordbeg;
+ 
+/**/
+int parbegin;
+
+/**/
+int parend;
+ 
+/* text of puctuation tokens */
+
+static char *tokstrings[WHILE + 1] = {
+    NULL,	/* NULLTOK	  0  */
+    ";",	/* SEPER	     */
+    "\\n",	/* NEWLIN	     */
+    ";",	/* SEMI		     */
+    ";;",	/* DSEMI	     */
+    "&",	/* AMPER	  5  */
+    "(",	/* INPAR	     */
+    ")",	/* OUTPAR	     */
+    "||",	/* DBAR		     */
+    "&&",	/* DAMPER	     */
+    ")",	/* OUTANG	  10 */
+    ">|",	/* OUTANGBANG	     */
+    ">>",	/* DOUTANG	     */
+    ">>|",	/* DOUTANGBANG	     */
+    "<",	/* INANG	     */
+    "<>",	/* INOUTANG	  15 */
+    "<<",	/* DINANG	     */
+    "<<-",	/* DINANGDASH	     */
+    "<&",	/* INANGAMP	     */
+    ">&",	/* OUTANGAMP	     */
+    "&>",	/* AMPOUTANG	  20 */
+    "&>|",	/* OUTANGAMPBANG     */
+    ">>&",	/* DOUTANGAMP	     */
+    ">>&|",	/* DOUTANGAMPBANG    */
+    "<<<",	/* TRINANG	     */
+    "|",	/* BAR		  25 */
+    "|&",	/* BARAMP	     */
+    "()",	/* INOUTPAR	     */
+    "((",	/* DINPAR	     */
+    "))",	/* DOUTPAR	     */
+    "&|",	/* AMPERBANG	  30 */
+    ";&",	/* SEMIAMP	     */
+};
+
 /* lexical state */
 
 static int dbparens;
 static int len = 0, bsiz = 256;
 static char *bptr;
-extern int hwgetword;
 
 struct lexstack {
     struct lexstack *next;
@@ -74,10 +185,6 @@ struct lexstack {
 
 static struct lexstack *lstack = NULL;
 
-/* History word status variables from hist.c */
-extern short *chwords;
-extern int chwordlen, chwordpos;
-
 /* save the lexical state */
 
 /* is this a hack or what? */
@@ -287,7 +394,7 @@ ctxtlex(void)
 #define LX2_OTHER 18
 #define LX2_META 19
 
-unsigned char lexact1[256], lexact2[256], lextok2[256];
+static unsigned char lexact1[256], lexact2[256], lextok2[256];
 
 /**/
 void
@@ -1042,8 +1149,6 @@ gettokstr(int c, int sub)
     return peek;
 }
 
-extern int addedx;
-
 /**/
 static int
 dquote_parse(char endchar, int sub)
