@@ -1,6 +1,4 @@
 /*
- * $Id: math.c,v 3.1.2.1 1997/06/01 06:13:15 hzoli Exp $
- *
  * math.c - mathematical expression evaluation
  *
  * This file is part of zsh, the Z shell.
@@ -29,9 +27,19 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "math.pro"
 
+/* nonzero means we are not evaluating, just parsing */
+ 
+/**/
+int noeval;
+ 
+/* last input base we used */
+
+/**/
+int lastbase;
+ 
 static char *ptr;
 
 static long yyval;
@@ -395,8 +403,9 @@ zzlex(void)
 /* the value stack */
 
 #define STACKSZ 100
-int mtok;			/* last token */
-int sp = -1;			/* stack pointer */
+static int mtok;			/* last token */
+static int sp = -1;			/* stack pointer */
+
 struct mathvalue {
     LV lval;
     long val;
