@@ -1,6 +1,4 @@
 /*
- * $Id: mem.c,v 3.1.2.2 1997/06/01 06:13:15 hzoli Exp $
- *
  * mem.c - memory management
  *
  * This file is part of zsh, the Z shell.
@@ -29,7 +27,7 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "mem.pro"
 
 /*
@@ -46,7 +44,7 @@
 	pushheap() saves the states of all currently allocated heaps and
 	popheap() resets them to the last state saved and destroys the
 	information about that state.  If you called pushheap() and
-	allocated some meory on the heaps and then come to a place where
+	allocated some memory on the heaps and then come to a place where
 	you don't need the allocated memory anymore but you still want
 	to allocate memory on the heap, you should call freeheap().  This
 	works like popheap(), only that it doesn't free the information
@@ -69,12 +67,23 @@
 	If we use zsh's own allocator we use a simple trick to avoid that
 	the (*real*) heap fills up with empty zsh-heaps: we allocate a
 	large block of memory before allocating a heap pool, this memory
-	is freed again immediatly after the pool is allocated. If there
-	are only small blocks on the free list this guarentees that the
+	is freed again immediately after the pool is allocated. If there
+	are only small blocks on the free list this guarantees that the
 	memory for the pool is at the end of the memory which means that
-	we can give it back to the systems when the pool is freed.
+	we can give it back to the system when the pool is freed.
 */
 
+/* != 0 if we are allocating in the heaplist */
+ 
+/**/
+int useheap;
+
+/* Current allocation pointers.  ncalloc() is either zalloc() or halloc(); *
+ * alloc() is either zcalloc() or hcalloc().                               */
+
+/**/
+void *(*ncalloc) _((size_t)), *(*alloc) _((size_t));
+
 #ifdef ZSH_MEM_WARNING
 # ifndef DEBUG
 #  define DEBUG 1
@@ -83,7 +92,7 @@
 
 #if defined(ZSH_MEM) && defined(ZSH_MEM_DEBUG)
 
-int h_m[1025], h_push, h_pop, h_free;
+static int h_m[1025], h_push, h_pop, h_free;
 
 #endif
 
@@ -138,7 +147,7 @@ struct heap {
 
 /* list of zsh heap */
 
-Heap heaps;
+static Heap heaps;
 
 /* save states of zsh heaps */
 
@@ -411,9 +420,14 @@ void *
 zrealloc(void *ptr, size_t size)
 {
     if (ptr) {
-	if (size)
+	if (size) {
 	    /* Do normal realloc */
-	    return realloc(ptr, size);
+	    if (!(ptr = (void *) realloc(ptr, size))) {
+		zerr("fatal error: out of memory", NULL, 0);
+		exit(1);
+	    }
+	    return ptr;
+	}
 	else
 	    /* If ptr is not NULL, but size is zero, *
 	     * then object pointed to is freed.      */
@@ -548,17 +562,17 @@ struct m_hdr {
 /* a pointer to the last free block, a pointer to the free list (the blocks
    on this list are kept in order - lowest address first) */
 
-struct m_hdr *m_lfree, *m_free;
+static struct m_hdr *m_lfree, *m_free;
 
 /* system's pagesize */
 
-long m_pgsz = 0;
+static long m_pgsz = 0;
 
 /* the highest and the lowest valid memory addresses, kept for fast validity
    checks in free() and to find out if and when we can give memory back to
    the system */
 
-char *m_high, *m_low;
+static char *m_high, *m_low;
 
 /* Management of blocks for small blocks:
    Such blocks are kept in lists (one list for each of the sizes that are
@@ -583,14 +597,14 @@ char *m_high, *m_low;
 		     sizeof(long) - sizeof(struct m_hdr *)) / M_SNUM)
 #define M_NSMALL 8
 
-struct m_hdr *m_small[M_NSMALL];
+static struct m_hdr *m_small[M_NSMALL];
 
 #ifdef ZSH_MEM_DEBUG
 
-int m_s = 0, m_b = 0;
-int m_m[1025], m_f[1025];
+static int m_s = 0, m_b = 0;
+static int m_m[1025], m_f[1025];
 
-struct m_hdr *m_l;
+static struct m_hdr *m_l;
 
 #endif /* ZSH_MEM_DEBUG */
 
