@@ -1,6 +1,4 @@
 /*
- * $Id: options.c,v 3.1.2.4 1997/06/01 06:13:15 hzoli Exp $
- *
  * options.c - shell options
  *
  * This file is part of zsh, the Z shell.
@@ -29,11 +27,24 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "options.pro"
 
+/* current emulation (used to decide which set of option letters is used) */
+
+/**/
+int emulation;
+ 
+/* the options; e.g. if opts[SHGLOB] != 0, SH_GLOB is turned on */
+ 
+/**/
+char opts[OPT_SIZE];
+ 
 /* Option name hash table */
 
+/**/
+HashTable optiontab;
+ 
 typedef struct optname *Optname;
 
 struct optname {
@@ -43,8 +54,6 @@ struct optname {
     int optno;			/* option number */
 };
 
-HashTable optiontab;
-
 /* The canonical option name table */
 
 #define OPT_CSH		EMULATE_CSH
@@ -80,6 +89,7 @@ static struct optname optns[] = {
 {NULL, "autoresume",	      0,			 AUTORESUME},
 {NULL, "badpattern",	      OPT_EMULATE|OPT_NONBOURNE, BADPATTERN},
 {NULL, "banghist",	      OPT_EMULATE|OPT_NONBOURNE, BANGHIST},
+{NULL, "bareglobqual",        OPT_EMULATE|OPT_ZSH,       BAREGLOBQUAL},
 {NULL, "beep",		      OPT_ALL,			 BEEP},
 {NULL, "bgnice",	      OPT_EMULATE|OPT_NONBOURNE, BGNICE},
 {NULL, "braceccl",	      0,			 BRACECCL},
@@ -114,6 +124,7 @@ static struct optname optns[] = {
 {NULL, "histbeep",	      OPT_ALL,			 HISTBEEP},
 {NULL, "histignoredups",      0,			 HISTIGNOREDUPS},
 {NULL, "histignorespace",     0,			 HISTIGNORESPACE},
+{NULL, "histnofunctions",     0,			 HISTNOFUNCTIONS},
 {NULL, "histnostore",	      0,			 HISTNOSTORE},
 {NULL, "histreduceblanks",    0,			 HISTREDUCEBLANKS},
 {NULL, "histverify",	      0,			 HISTVERIFY},
@@ -124,6 +135,7 @@ static struct optname optns[] = {
 {NULL, "interactivecomments", OPT_EMULATE|OPT_BOURNE,	 INTERACTIVECOMMENTS},
 {NULL, "ksharrays",	      OPT_EMULATE|OPT_BOURNE,	 KSHARRAYS},
 {NULL, "kshautoload",	      OPT_EMULATE|OPT_BOURNE,	 KSHAUTOLOAD},
+{NULL, "kshglob",             OPT_EMULATE|OPT_KSH,       KSHGLOB},
 {NULL, "kshoptionprint",      OPT_EMULATE|OPT_KSH,	 KSHOPTIONPRINT},
 {NULL, "listambiguous",	      OPT_ALL,			 LISTAMBIGUOUS},
 {NULL, "listbeep",	      OPT_ALL,			 LISTBEEP},
@@ -144,9 +156,12 @@ static struct optname optns[] = {
 {NULL, "overstrike",	      0,			 OVERSTRIKE},
 {NULL, "pathdirs",	      0,			 PATHDIRS},
 {NULL, "posixbuiltins",	      OPT_EMULATE|OPT_BOURNE,	 POSIXBUILTINS},
+{NULL, "printeightbit",       0,                         PRINTEIGHTBIT},
 {NULL, "printexitvalue",      0,			 PRINTEXITVALUE},
 {NULL, "privileged",	      OPT_SPECIAL,		 PRIVILEGED},
+{NULL, "promptbang",	      OPT_EMULATE|OPT_KSH,	 PROMPTBANG},
 {NULL, "promptcr",	      OPT_ALL,			 PROMPTCR},
+{NULL, "promptpercent",	      OPT_EMULATE|OPT_NONBOURNE, PROMPTPERCENT},
 {NULL, "promptsubst",	      OPT_EMULATE|OPT_KSH,	 PROMPTSUBST},
 {NULL, "pushdignoredups",     0,			 PUSHDIGNOREDUPS},
 {NULL, "pushdminus",	      0,			 PUSHDMINUS},
@@ -158,6 +173,7 @@ static struct optname optns[] = {
 {NULL, "recexact",	      0,			 RECEXACT},
 {NULL, "restricted",	      OPT_SPECIAL,		 RESTRICTED},
 {NULL, "rmstarsilent",	      OPT_BOURNE,		 RMSTARSILENT},
+{NULL, "rmstarwait",	      0,			 RMSTARWAIT},
 {NULL, "shfileexpansion",     OPT_EMULATE|OPT_BOURNE,	 SHFILEEXPANSION},
 {NULL, "shglob",	      OPT_EMULATE|OPT_BOURNE,	 SHGLOB},
 {NULL, "shinstdin",	      OPT_SPECIAL,		 SHINSTDIN},
@@ -176,6 +192,7 @@ static struct optname optns[] = {
 {NULL, "hashall",	      OPT_ALIAS, /* bash */	 HASHCMDS},
 {NULL, "histappend",	      OPT_ALIAS, /* bash */	 APPENDHISTORY},
 {NULL, "histexpand",	      OPT_ALIAS, /* bash */	 BANGHIST},
+{NULL, "log",		      OPT_ALIAS, /* ksh */	 -HISTNOFUNCTIONS},
 {NULL, "mailwarn",	      OPT_ALIAS, /* bash */	 MAILWARNING},
 {NULL, "onecmd",	      OPT_ALIAS, /* bash */	 SINGLECOMMAND},
 {NULL, "physical",	      OPT_ALIAS, /* ksh/bash */	 CHASELINKS},
