@@ -1,6 +1,4 @@
 /*
- * $Id: parse.c,v 3.1.2.4 1997/06/01 06:13:15 hzoli Exp $
- *
  * parse.c - parser
  *
  * This file is part of zsh, the Z shell.
@@ -29,11 +27,52 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "parse.pro"
 
+/* != 0 if we are about to read a command word */
+ 
+/**/
+int incmdpos;
+ 
+/* != 0 if we are in the middle of a [[ ... ]] */
+ 
+/**/
+int incond;
+ 
+/* != 0 if we are after a redirection (for ctxtlex only) */
+ 
+/**/
+int inredir;
+ 
+/* != 0 if we are about to read a case pattern */
+ 
+/**/
+int incasepat;
+ 
+/* != 0 if we just read a newline */
+ 
+/**/
+int isnewlin;
+
+/* != 0 if we are after a for keyword */
+
+/**/
+int infor;
+
+/* list of here-documents */
+
+/**/
+struct heredocs *hdocs;
+ 
+/* used in arrays of lists instead of NULL pointers */
+ 
+/**/
+struct list dummy_list;
+
 #define YYERROR  { tok = LEXERR; return NULL; }
 #define YYERRORV { tok = LEXERR; return; }
+#define COND_ERROR(X,Y) do{zerr(X,Y,0);discard_input();YYERROR}while(0)
 
 #define make_list()     allocnode(N_LIST)
 #define make_sublist()  allocnode(N_SUBLIST)
@@ -104,18 +143,7 @@ par_event(void)
 	    return NULL;
 	}
 	yyerror();
-	errflag = 0;
-	if (isnewlin <= 0) {
-	    /* Discard remaining stuff after a parse error. */
-	    int c;
-
-	    hwbegin(0);
-	    while ((c = hgetc()) != '\n' && !lexstop);
-	    if (c == '\n')
-		hungetc('\n');
-	    hwend();
-	}
-	errflag = 1;
+	discard_input();
 	return NULL;
     } else {
 	l->right = par_event();
@@ -994,6 +1022,8 @@ par_simple(Cmd c)
  * condlex is yylex for normal parsing, but is altered to allow
  * the test builtin to use par_cond.
  */
+
+/**/
 void (*condlex) _((void)) = yylex;
 
 /*
@@ -1063,7 +1093,6 @@ par_cond_2(void)
     Cond c, c2;
     char *s1, *s2, *s3;
     int dble = 0;
-    extern char **testargs;
 
     if (condlex == testlex) {
 	/* See the description of test in POSIX 1003.2 */
@@ -1136,7 +1165,7 @@ par_cond_2(void)
 	c->left = (void *) s1;
 	c->right = (void *) s3;
 	c->type = (xtok == INANG) ? COND_STRLT : COND_STRGTR;
-	c->ntype = NT_SET(N_COND, 1, NT_STR, NT_STR, 0, 0);
+	c->ntype = NT_SET(N_COND, NT_STR, NT_STR, 0, 0);
 	return c;
     }
     if (tok != STRING)
@@ -1163,6 +1192,24 @@ par_cond_2(void)
  * redir	: ( OUTANG | ... | TRINANG ) STRING
  */
 
+static int redirtab[TRINANG - OUTANG + 1] = {
+    WRITE,
+    WRITENOW,
+    APP,
+    APPNOW,
+    READ,
+    READWRITE,
+    HEREDOC,
+    HEREDOCDASH,
+    MERGEIN,
+    MERGEOUT,
+    ERRWRITE,
+    ERRWRITENOW,
+    ERRAPP,
+    ERRAPPNOW,
+    HERESTR,
+};
+
 /**/
 static void
 par_redir(LinkList l)
@@ -1268,13 +1315,11 @@ par_cond_double(char *a, char *b)
 {
     Cond n = (Cond) make_cond();
 
-    if (a[0] != '-' || !a[1] || a[2]) {
-	zerr("parse error: condition expected: %s", a, 0);
-	return NULL;
-    }
+    if (a[0] != '-' || !a[1] || a[2])
+	COND_ERROR("parse error: condition expected: %s", a);
     n->left = (void *) b;
     n->type = a[1];
-    n->ntype = NT_SET(N_COND, 1, NT_STR, NT_STR, 0, 0);
+    n->ntype = NT_SET(N_COND, NT_STR, NT_STR, 0, 0);
     return n;
 }
 
@@ -1310,15 +1355,33 @@ par_cond_triple(char *a, char *b, char *c)
 	if ((t0 = get_cond_num(b + 1)) > -1)
 	    n->type = t0 + COND_NT;
 	else
-	    zerr("unrecognized condition: %s", b, 0);
+	    COND_ERROR("unrecognized condition: %s", b);
     } else
-	zerr("condition expected: %s", b, 0);
+	COND_ERROR("condition expected: %s", b);
     n->left = (void *) a;
     n->right = (void *) c;
-    n->ntype = NT_SET(N_COND, 1, NT_STR, NT_STR, 0, 0);
+    n->ntype = NT_SET(N_COND, NT_STR, NT_STR, 0, 0);
     return n;
 }
 
+/**/
+static void
+discard_input(void)
+{
+    errflag = 0;
+    if (isnewlin <= 0) {
+	/* Discard remaining stuff after a parse error. */
+	int c;
+
+	hwbegin(0);
+	while ((c = hgetc()) != '\n' && !lexstop);
+	if (c == '\n')
+	    hungetc('\n');
+	hwend();
+    }
+    errflag = 1;
+}
+
 /**/
 static void
 yyerror(void)
