@@ -1,6 +1,4 @@
 /*
- * $Id: prompt.c,v 3.1.2.4 1997/06/01 06:13:15 hzoli Exp $
- *
  * prompt.c - construct zsh prompts
  *
  * This file is part of zsh, the Z shell.
@@ -29,161 +27,139 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "prompt.pro"
 
+/* text attribute mask */
+ 
+/**/
+unsigned txtattrmask;
+
+/* text change - attribute change made by prompts */
+
+/**/
+unsigned txtchange;
+
+/* the command stack for use with %_ in prompts */
+ 
+/**/
+unsigned char *cmdstack;
+/**/
+int cmdsp;
+
+/* parser states, for %_ */
+
+static char *cmdnames[] = {
+    "for",      "while",     "repeat",    "select",
+    "until",    "if",        "then",      "else",
+    "elif",     "math",      "cond",      "cmdor",
+    "cmdand",   "pipe",      "errpipe",   "foreach",
+    "case",     "function",  "subsh",     "cursh",
+    "array",    "quote",     "dquote",    "bquote",
+    "cmdsubst", "mathsubst", "elif-then", "heredoc",
+    "heredocd", "brace",     "braceparam",
+};
+ 
+/* The buffer into which an expanded and metafied prompt is being written, *
+ * and its size.                                                           */
+
+static char *buf;
+static int bufspc;
+
+/* bp is the pointer to the current position in the buffer, where the next *
+ * character will be added.                                                */
+
 static char *bp;
+
+/* bp1 is an auxilliary pointer into the buffer, which when non-NULL is *
+ * moved whenever the buffer is reallocated.  It is used when data is   *
+ * being temporarily held in the buffer.                                */
+
+static char *bp1;
+
+/* The format string, for %-expansion. */
+
+static char *fm;
+
+/* Current truncation string (metafied), the length at which truncation *
+ * occurs, and the direction in which it occurs.                        */
+
 static char *truncstr;
-static int dontcount, lensb, trunclen;
+static int trunclen, truncatleft;
 
-/* stradd() adds a string to the prompt, in a *
- * visible representation, doing truncation.  */
+/* Current level of nesting of %{ / %} sequences. */
 
-/**/
-static void
-stradd(char *d)
-{
-    /* dlen is the full length of the string we want to add */
-    int dlen = nicestrlen(d);
-    char *ps, *pd, *pc, *t;
-    int tlen, maxlen;
-    addbufspc(dlen);
-    /* This loop puts the nice representation of the string into the prompt *
-     * buffer.  It might be modified later.  Note that bp isn't changed.    */
-    for(ps=d, pd=bp; *ps; ps++)
-	for(pc=nicechar(STOUC(*ps)); *pc; pc++)
-	    *pd++ = *pc;
-    if(!trunclen || dlen <= trunclen) {
-	/* No truncation is needed, so update bp and return, *
-	 * leaving the full string in the prompt.            */
-	bp += dlen;
-	return;
-    }
-    /* We need to truncate.  t points to the truncation string -- which is *
-     * inserted literally, without nice representation.  tlen is its       *
-     * length, and maxlen is the amout of the main string that we want to  *
-     * keep.  Note that if the truncation string is longer than the        *
-     * truncation length (tlen > trunclen), the truncation string is used  *
-     * in full.                                                            */
-    t = truncstr + 1;
-    tlen = strlen(t);
-    maxlen = tlen < trunclen ? trunclen - tlen : 0;
-    addbufspc(tlen);
-    if(*truncstr == '>') {
-	/* '>' means truncate at the right.  We just move past the first *
-	 * maxlen characters of the string, and write the truncation     *
-	 * string there.                                                 */
-	bp += maxlen;
-	while(*t)
-	    pputc(*t++);
-    } else {
-	/* Truncation at the left: ps is initialised to the start of the *
-	 * part of the string that we want to keep.  pc points to the    *
-	 * end of the string.  The truncation string is added to the     *
-	 * prompt, then the desired part of the string is copied into    *
-	 * the right place.                                              */
-	ps = bp + dlen - maxlen;
-	pc = bp + dlen;
-	while(*t)
-	    pputc(*t++);
-	while(ps < pc)
-	    *bp++ = *ps++;
-    }
-}
+static int dontcount;
 
-/**/
-static int
-putstr(int d)
-{
-    addbufspc(1);
-    *bp++ = d;
-    if (!dontcount)
-	lensb++;
-    return 0;
-}
+/* Strings to use for %r and %R (for the spelling prompt). */
 
-/* get a prompt string */
+static char *rstring, *Rstring;
 
-static char *buf, *bp1, *bl0, *fm, *pmpt;
-static int bracepos, bufspc, pmpth;
+/* If non-zero, Inpar, Outpar and Nularg can be added to the buffer. */
+
+static int nonsp;
+
+/* Perform prompt expansion on a string, putting the result in a *
+ * permanently-allocated string.  If ns is non-zero, this string *
+ * may have embedded Inpar and Outpar, which indicate a toggling *
+ * between spacing and non-spacing parts of the prompt, and      *
+ * Nularg, which (in a non-spacing sequence) indicates a         *
+ * `glitch' space.                                               */
 
 /**/
 char *
-putprompt(char *fmin, int *lenp, int *wp, int cnt)
+promptexpand(char *s, int ns, char *rs, char *Rs)
 {
-    if (!fmin) {
-	*lenp = 0;
-	if (wp)
-	    *wp = 0;
+    if(!s)
 	return ztrdup("");
-    }
-    pmpth = 1;
 
     if ((termflags & TERM_UNKNOWN) && (unset(INTERACTIVE)))
         init_term();
 
-    bracepos = -1;
-    fm = fmin;
-    lensb = 0;
-    pmpt = (dontcount = !cnt) ? NULL : fm;
-    bp = bl0 = buf = zalloc(bufspc = 256);
-    bp1 = NULL;
-
-    clearerr(stdin);
-
-    trunclen = 0;
-    putpromptchar(1, '\0');
-
-    *lenp = bp - buf;
-    if (wp) {
-	*wp = bp - bl0 - lensb;
-	if (pmpt != rpmpt) {
-	    *wp %= columns;
-	    if (*wp == columns - 1) {
-		addbufspc(1);
-		*wp = 0;
-		*bp++ = ' ';
-		++*lenp;
-		pmpth++;
-	    }
-	    if (!*wp && *lenp) {
-		addbufspc(1);
-		(*wp)++;
-	    	*bp++ = ' ';
-		++*lenp;
-	    }
-	}
-    }
-    if (pmpt == lpmpt)
-	lppth = pmpth;
-
-    return buf;
-}
-
-/**/
-static int
-putpromptchar(int doprint, int endchar)
-{
-    char buf3[PATH_MAX], *ss;
-    int t0, arg, test, sep;
-    struct tm *tm;
-    time_t timet;
-    Nameddir nd;
-
     if (isset(PROMPTSUBST)) {
 	int olderr = errflag;
 
 	HEAPALLOC {
-	    fm = dupstring(fm);
-	    if (!parsestr(fm))
-		singsub(&fm);
+	    s = dupstring(s);
+	    if (!parsestr(s))
+		singsub(&s);
 	} LASTALLOC;
 	/* Ignore errors in prompt substitution */
 	errflag = olderr;
     }
+
+    rstring = rs;
+    Rstring = Rs;
+    nonsp = ns;
+    fm = s;
+    bp = buf = zalloc(bufspc = 256);
+    bp1 = NULL;
+    trunclen = 0;
+    putpromptchar(1, '\0');
+    addbufspc(1);
+    if(dontcount)
+	*bp++ = Outpar;
+    *bp = 0;
+    return buf;
+}
+
+/* Perform %- and !-expansion as required on a section of the prompt.  The *
+ * section is ended by an instance of endchar.  If doprint is 0, the valid *
+ * % sequences are merely skipped over, and nothing is stored.             */
+
+/**/
+static int
+putpromptchar(int doprint, int endchar)
+{
+    char *ss, *tmbuf = NULL;
+    int t0, arg, test, sep;
+    struct tm *tm;
+    time_t timet;
+    Nameddir nd;
+
     for (; *fm && *fm != endchar; fm++) {
 	arg = 0;
-	if (*fm == '%') {
+	if (*fm == '%' && isset(PROMPTPERCENT)) {
 	    if (idigit(*++fm)) {
 		arg = zstrtol(fm, &fm, 10);
 	    }
@@ -406,6 +382,7 @@ putpromptchar(int doprint, int endchar)
                 else
                     trunclen = arg;
                 if (trunclen) {
+		    truncatleft = *fm && *fm != ']' && *fm++ == '<';
 		    bp1 = bp;
 		    while (*fm && *fm != ']') {
 			if (*fm == '\\' && fm[1])
@@ -433,10 +410,10 @@ putpromptchar(int doprint, int endchar)
 	    case '<':
 	    case '>':
 		if((trunclen = arg)) {
+		    char ch = *fm++;
+		    truncatleft = ch == '<';
 		    bp1 = bp;
-		    addbufspc(1);
-		    *bp++ = *fm++;
-		    while (*fm && *fm != *bp1) {
+		    while (*fm && *fm != ch) {
 			if (*fm == '\\' && fm[1])
 			    ++fm;
 			addbufspc(1);
@@ -458,14 +435,16 @@ putpromptchar(int doprint, int endchar)
 		if(!*fm)
 		    return 0;
 		break;
-	    case '{':
-		if (!dontcount++)
-		    bracepos = bp - buf;
+	    case '{': /*}*/
+		if (!dontcount++ && nonsp) {
+		    addbufspc(1);
+		    *bp++ = Inpar;
+		}
 		break;
-	    case '}':
-		if (!--dontcount && bracepos != -1) {
-		    lensb += (bp - buf) - bracepos;
-		    bracepos = -1;
+	    case /*{*/ '}':
+		if (dontcount && !--dontcount && nonsp) {
+		    addbufspc(1);
+		    *bp++ = Outpar;
 		}
 		break;
 	    case 't':
@@ -492,16 +471,20 @@ putpromptchar(int doprint, int endchar)
 			tmfmt = "%m/%d/%y";
 			break;
 		    case 'D':
-			tmfmt = "%y-%m-%d";
-			if (fm[1] == '{') {
-			    for (ss = fm + 2, dd = buf3; *ss && *ss != '}'; ++ss, ++dd)
-				*dd = *((*ss == '\\' && ss[1]) ? ++ss : ss);
-			    if (*ss == '}') {
-				*dd = '\0';
-				fm = ss;
-				tmfmt = buf3;
+			if (fm[1] == '{') /*}*/ {
+			    for (ss = fm + 2; *ss && *ss != /*{*/ '}'; ss++)
+				if(*ss == '\\' && ss[1])
+				    ss++;
+			    dd = tmfmt = tmbuf = zalloc(ss - fm);
+			    for (ss = fm + 2; *ss && *ss != /*{*/ '}'; ss++) {
+				if(*ss == '\\' && ss[1])
+				    ss++;
+				*dd++ = *ss;
 			    }
-			}
+			    *dd = 0;
+			    fm = ss - !*ss;
+			} else
+			    tmfmt = "%y-%m-%d";
 			break;
 		    default:
 			tmfmt = "%l:%M%p";
@@ -509,11 +492,14 @@ putpromptchar(int doprint, int endchar)
 		    }
 		    timet = time(NULL);
 		    tm = localtime(&timet);
-		    addbufspc(80);   /* 80 is arbitrary :-( */
-		    ztrftime(bp, 79, tmfmt, tm);
-		    if (*bp == ' ')
-			chuck(bp);
+		    for(t0=80; ; t0*=2) {
+			addbufspc(t0);
+			if(ztrftime(bp, t0, tmfmt, tm) != t0)
+			    break;
+		    }
 		    bp += strlen(bp);
+		    free(tmbuf);
+		    tmbuf = NULL;
 		    break;
 		}
 	    case 'n':
@@ -527,18 +513,20 @@ putpromptchar(int doprint, int endchar)
 		} else
 		    stradd("()");
 		break;
+	    case 'L':
+		addbufspc(DIGBUFSIZE);
+		sprintf(bp, "%ld", (long)shlvl);
+		bp += strlen(bp);
+		break;
 	    case '?':
 		addbufspc(DIGBUFSIZE);
 		sprintf(bp, "%ld", (long)lastval);
 		bp += strlen(bp);
 		break;
 	    case '%':
-		addbufspc(1);
-		*bp++ = '%';
-		break;
 	    case ')':
 		addbufspc(1);
-		*bp++ = ')';
+		*bp++ = *fm;
 		break;
 	    case '#':
 		addbufspc(1);
@@ -567,72 +555,122 @@ putpromptchar(int doprint, int endchar)
 		}
 		break;
 	    case 'r':
-		if (pmpt == sprompt) {
+		if(rstring)
 		    stradd(rstring);
-		    break;
-		}
+		break;
 	    case 'R':
-		if (pmpt == sprompt) {
+		if(Rstring)
 		    stradd(Rstring);
-		    break;
-		}
+		break;
 	    case '\0':
 		return 0;
-	    default:
-		addbufspc(2);
-		*bp++ = '%';
-		pputc(*fm);
+	    case Meta:
+		fm++;
 		break;
 	    }
-	} else if (doprint) {
-	    addbufspc(1);
-	    pputc(*fm == Meta ? *++fm ^ 32 : *fm);
+	} else if(*fm == '!' && isset(PROMPTBANG)) {
+	    if(doprint)
+		if(fm[1] == '!') {
+		    fm++;
+		    addbufspc(1);
+		    pputc('!');
+		} else {
+		    addbufspc(DIGBUFSIZE);
+		    sprintf(bp, "%d", curhist);
+		    bp += strlen(bp);
+		}
+	} else {
+	    char c = *fm == Meta ? *++fm ^ 32 : *fm;
+
+	    if (doprint) {
+		addbufspc(1);
+		pputc(c);
+	    }
 	}
     }
 
     return *fm;
 }
 
+/* pputc adds a character to the buffer, metafying.  There must *
+ * already be space.                                            */
+
 /**/
 static void
 pputc(char c)
 {
-    if (!dontcount) {
-	if (pmpt == rpmpt) {
-	    if (c == '\t' || c == '\n')
-		c = ' ';
-	} else if (c == '\t') {
-	    int t0 = 7 - (7 & (bp - bl0 - lensb));
-	    lensb -= t0;
-	} else if (c == '\n') {
-	    *bp++ = c;
-	    bl0 = bp;
-	    lensb = 0;
-	    pmpth++;
-	    return;
-	}
+    if(imeta(STOUC(c))) {
+	*bp++ = Meta;
+	c ^= 32;
     }
-    if (bp - bl0 == columns)
-	pmpth++;
     *bp++ = c;
 }
 
+/* Make sure there is room for `need' more characters in the buffer. */
+
 /**/
 static void
 addbufspc(int need)
 {
+    need *= 2;   /* for metafication */
     if((bp - buf) + need > bufspc) {
-	char *oldbuf = buf;
+	int bo = bp - buf;
+	int bo1 = bp1 ? bp1 - buf : -1;
+
 	if(need & 255)
 	    need = (need | 255) + 1;
 	buf = realloc(buf, bufspc += need);
-	bp = buf + (bp - oldbuf);
-	bl0 = buf + (bl0 - oldbuf);
-	if(bp1)
-	    bp1 = buf + (bp1 - oldbuf);
+	bp = buf + bo;
+	if(bo1 != -1)
+	    bp1 = buf + bo1;
     }
 }
 
+/* stradd() adds a metafied string to the prompt, *
+ * in a visible representation, doing truncation. */
+
+/**/
+void
+stradd(char *d)
+{
+    /* dlen is the full length of the string we want to add */
+    int dlen = niceztrlen(d);
+    char *ps, *pd, *pc, *t;
+    int tlen, maxlen;
+    addbufspc(dlen);
+    /* This loop puts the nice representation of the string into the prompt *
+     * buffer.  It might be modified later.  Note that bp isn't changed.    */
+    for(ps=d, pd=bp; *ps; ps++)
+	for(pc=nicechar(*ps == Meta ? STOUC(*++ps)^32 : STOUC(*ps)); *pc; pc++)
+	    *pd++ = *pc;
+    if(!trunclen || dlen <= trunclen) {
+	/* No truncation is needed, so update bp and return, *
+	 * leaving the full string in the prompt.            */
+	bp += dlen;
+	return;
+    }
+    /* We need to truncate.  t points to the truncation string -- which is *
+     * inserted literally, without nice representation.  tlen is its       *
+     * length, and maxlen is the amout of the main string that we want to  *
+     * keep.  Note that if the truncation string is longer than the        *
+     * truncation length (tlen > trunclen), the truncation string is used  *
+     * in full.                                                            */
+    addbufspc(tlen = ztrlen(t = truncstr));
+    maxlen = tlen < trunclen ? trunclen - tlen : 0;
+    if(truncatleft) {
+	memmove(bp + strlen(t), bp + dlen - maxlen, maxlen);
+	while(*t)
+	    *bp++ = *t++;
+	bp += maxlen;
+    } else {
+	bp += maxlen;
+	while(*t)
+	    *bp++ = *t++;
+    }
+}
+
+/* tsetcap(), among other things, can write a termcap string into the buffer. */
+
 /**/
 void
 tsetcap(int cap, int flag)
@@ -646,19 +684,25 @@ tsetcap(int cap, int flag)
 	    tputs(tcstr[cap], 1, putshout);
 	    break;
 	case 1:
-	    if (!dontcount) {
-		int  t0;
-
-		if (cap == TCSTANDOUTBEG || cap == TCSTANDOUTEND)
-		    t0 = tgetnum("sg");
-		else if (cap == TCUNDERLINEBEG || cap == TCUNDERLINEEND)
-		    t0 = tgetnum("ug");
-		else
-		    t0 = 0;
-		if (t0 > 0)
-		    lensb -= t0;
+	    if (!dontcount && nonsp) {
+		addbufspc(1);
+		*bp++ = Inpar;
 	    }
 	    tputs(tcstr[cap], 1, putstr);
+	    if (!dontcount && nonsp) {
+		int glitch = 0;
+
+		if (cap == TCSTANDOUTBEG || cap == TCSTANDOUTEND)
+		    glitch = tgetnum("sg");
+		else if (cap == TCUNDERLINEBEG || cap == TCUNDERLINEEND)
+		    glitch = tgetnum("ug");
+		if(glitch < 0)
+		    glitch = 0;
+		addbufspc(glitch + 1);
+		while(glitch--)
+		    *bp++ = Nularg;
+		*bp++ = Outpar;
+	    }
 	    break;
 	}
 
@@ -673,3 +717,50 @@ tsetcap(int cap, int flag)
 	}
     }
 }
+
+/**/
+int
+putstr(int d)
+{
+    addbufspc(1);
+    pputc(d);
+    return 0;
+}
+
+/* Count height etc. of a prompt string returned by promptexpand(). *
+ * This depends on the current terminal width, and tabs and         *
+ * newlines require nontrivial processing.                          */
+
+/**/
+void
+countprompt(char *str, int *wp, int *hp)
+{
+    int w = 0, h = 1;
+    int s = 1;
+    for(; *str; str++) {
+	if(*str == Meta)
+	    str++;
+	if(*str == Inpar)
+	    s = 0;
+	else if(*str == Outpar)
+	    s = 1;
+	else if(*str == Nularg)
+	    w++;
+	else if(s) {
+	    if(*str == '\t')
+		w = (w | 7) + 1;
+	    else if(*str == '\n')
+		w = columns;
+	    else
+		w++;
+	}
+	if(w >= columns) {
+	    w = 0;
+	    h++;
+	}
+    }
+    if(wp)
+	*wp = w;
+    if(hp)
+	*hp = h;
+}
