@@ -1,6 +1,4 @@
 /*
- * $Id: prototypes.h,v 3.1.2.5 1997/06/01 06:13:15 hzoli Exp $
- *
  * prototypes.h - prototypes header file
  *
  * This file is part of zsh, the Z shell.
@@ -29,36 +27,6 @@
  *
  */
 
-#define GLOBAL_PROTOTYPES
-
-#include "builtin.pro"
-#include "compat.pro"
-#include "cond.pro"
-#include "exec.pro"
-#include "glob.pro"
-#include "hashtable.pro"
-#include "hist.pro"
-#include "init.pro"
-#include "input.pro"
-#include "jobs.pro"
-#include "lex.pro"
-#include "linklist.pro"
-#include "loop.pro"
-#include "math.pro"
-#include "mem.pro"
-#include "options.pro"
-#include "params.pro"
-#include "parse.pro"
-#include "prompt.pro"
-#include "signals.pro"
-#include "subst.pro"
-#include "text.pro"
-#include "utils.pro"
-#include "watch.pro"
-#include "module.pro"
-
-#undef GLOBAL_PROTOTYPES
-
 #ifndef HAVE_STDLIB_H
 char *malloc _((size_t));
 char *realloc _((void *, size_t));
@@ -108,7 +76,7 @@ int select _((int nfds, fd_set * readfds, fd_set * writefds, fd_set * exceptfds,
 extern int getrlimit _((int resource, struct rlimit *rlp));
 extern int setrlimit _((int resource, const struct rlimit *rlp));
 extern int getrusage _((int who, struct rusage *rusage));
-extern int gettimeofday _((struct timeval *time_value, struct timezone *time_zone));
+extern int gettimeofday _((struct timeval *tv, struct timezone *tz));
 extern int wait3 _((union wait *wait_status, int options, struct rusage *rusage));
 extern int getdomainname _((char *name, int maxlength));
 extern int select _((int nfds, fd_set * readfds, fd_set * writefds, fd_set * exceptfds, struct timeval *timeout));
@@ -133,7 +101,7 @@ extern int gethostname _((char *name, int namelen));
 #endif
 
 #ifndef HAVE_GETTIMEOFDAY
-extern void gettimeofday _((struct timeval *tv, struct timezone *tz));
+extern int gettimeofday _((struct timeval *tv, struct timezone *tz));
 #endif
 
 #ifndef HAVE_DIFFTIME
