@@ -1,6 +1,4 @@
 /*
- * $Id: signals.c,v 3.1.2.7 1997/06/01 06:13:15 hzoli Exp $
- *
  * signals.c - signals handling code
  *
  * This file is part of zsh, the Z shell.
@@ -29,9 +27,29 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "signals.pro"
  
+/* Array describing the state of each signal: an element contains *
+ * 0 for the default action or some ZSIG_* flags ored together.   */
+
+/**/
+int sigtrapped[VSIGCOUNT];
+
+/* trap functions for each signal */
+
+/**/
+List sigfuncs[VSIGCOUNT];
+
+/* Variables used by signal queueing */
+
+/**/
+int queueing_enabled, queue_front, queue_rear;
+/**/
+int signal_queue[MAX_QUEUE_SIZE];
+/**/
+sigset_t signal_mask_queue[MAX_QUEUE_SIZE];
+
 /* This is only used on machines that don't understand signal sets.  *
  * On SYSV machines this will represent the signals that are blocked *
  * (held) using sighold.  On machines which can't block signals at   *
@@ -474,8 +492,6 @@ handler(int sig)
         if (sigtrapped[SIGINT])
             dotrap(SIGINT);
         else {
-            extern int list_pipe, simple_pline;
- 
 	    if ((isset(PRIVILEGED) || isset(RESTRICTED)) &&
 		isset(INTERACTIVE) && noerrexit < 0)
 		zexit(SIGINT, 1);
