@@ -1,6 +1,4 @@
 /*
- * $Id: subst.c,v 3.1.2.1 1997/06/01 06:13:15 hzoli Exp $
- *
  * subst.c - various substitutions
  *
  * This file is part of zsh, the Z shell.
@@ -29,9 +27,12 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "subst.pro"
 
+/**/
+char nulstring[] = {Nularg, '\0'};
+
 /* Do substitutions before fork. These are:
  *  - Process substitution: <(...), >(...), =(...)
  *  - Parameter substitution
@@ -421,6 +422,8 @@ strcatsub(char **d, char *pb, char *pe, char *src, int l, char *s, int glbsub)
     return dest;
 }
 
+typedef int (*CompareFn) _((const void *, const void *));
+
 /**/
 int
 strpcmp(const void *a, const void *b)
@@ -683,7 +686,7 @@ LinkNode
 paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 {
     char *aptr = *str;
-    char *s = aptr, *u, *idbeg, *idend, *ostr = (char *) getdata(n);
+    char *s = aptr, *fstr, *idbeg, *idend, *ostr = (char *) getdata(n);
     int colf;			/* != 0 means we found a colon after the name */
     int doub = 0;		/* != 0 means we have %%, not %, or ##, not # */
     int isarr = 0;
@@ -695,7 +698,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
     int vunset = 0;
     int spbreak = isset(SHWORDSPLIT) && !ssub && !qt;
     char *val = NULL, **aval = NULL;
-    int fwidth = 0;
+    unsigned int fwidth = 0;
     Value v;
     int flags = 0;
     int flnum = 0;
@@ -996,7 +999,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		fwidth = v->pm->ct ? v->pm->ct : strlen(val);
 		switch (v->pm->flags & (PM_LEFT | PM_RIGHT_B | PM_RIGHT_Z)) {
 		    char *t;
-		    int t0;
+		    unsigned int t0;
 
 		case PM_LEFT:
 		case PM_LEFT | PM_RIGHT_Z:
@@ -1065,6 +1068,30 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
     if ((colf = *s == ':'))
 	s++;
 
+
+    /* fstr is to be the text following the substitution.  If we have *
+     * braces, we look for it here, else we infer it later on.        */
+    fstr = s;
+    if (inbrace) {
+	int bct;
+	for (bct = 1;; fstr++) {
+	    if (!*fstr)
+		break;
+	    else if (*fstr == Inbrace)
+		bct++;
+	    else if (*fstr == Outbrace && !--bct)
+		break;
+	}
+
+	if (bct) {
+	noclosebrace:
+	    zerr("closing brace expected", NULL, 0);
+	    return NULL;
+	}
+	if (*fstr)
+	    *fstr++ = '\0';
+    }
+
     /* Check for ${..?..} or ${..=..} or one of those. *
      * Only works if the name is in braces.            */
 
@@ -1075,7 +1102,6 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		    *s == '%' ||
 		    *s == '#' || *s == Pound ||
 		    *s == '?' || *s == Quest)) {
-	int bct;
 
 	if (!flnum)
 	    flnum++;
@@ -1087,28 +1113,16 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	    s++;
 	    doub = 1;
 	}
-	u = s + 1;
+	s++;
 
 	flags |= (doub << 1) | (substr << 2) | (colf << 8);
 	if (!(flags & 0xf8))
 	    flags |= 16;
 
-	for (bct = 1; bct && *++s;) {
-	    if (*s == Inbrace)
-		bct++;
-	    else if (*s == Outbrace)
-		bct--;
-	}
-
-	if (bct)
-	    goto noclosebrace;
-
-	if (*s)
-	    *s++ = '\0';
 	if (colf && !vunset)
 	    vunset = (isarr) ? !*aval : !*val || (*val == Nularg && !val[1]);
 
-	switch (u[-1]) {
+	switch (s[-1]) {
 	case '+':
 	    if (vunset) {
 		val = dupstring("");
@@ -1120,16 +1134,16 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	/* Fall Through! */
 	case '-':
 	    if (vunset) {
-		val = dupstring(u);
+		val = dupstring(s);
 		multsub(&val, &aval, &isarr, NULL);
 		copied = 1;
 	    }
 	    break;
 	case ':':
-	    if (*u != '=' && *u != Equals)
+	    if (*s != '=' && *s != Equals)
 		goto noclosebrace;
 	    vunset = 1;
-	    u++;
+	    s++;
 	    /* Fall through */
 	case '=':
 	case Equals:
@@ -1138,7 +1152,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		int l;
 
 		*idend = '\0';
-		val = dupstring(u);
+		val = dupstring(s);
 		isarr = 0;
 		if (spsep || spbreak || !arrasg)
 		    multsub(&val, NULL, NULL, sep);
@@ -1186,7 +1200,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		char *msg;
 
 		*idend = '\0';
-		msg = tricat(idbeg, ": ", *u ? u : "parameter not set");
+		msg = tricat(idbeg, ": ", *s ? s : "parameter not set");
 		zerr("%s", msg, 0);
 		zsfree(msg);
 		if (!interact)
@@ -1198,26 +1212,26 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	case '#':
 	case Pound:
 	    if (qt)
-		if (parse_subst_string(u)) {
+		if (parse_subst_string(s)) {
 		    zerr("parse error in ${...%c...} substitution",
-			 NULL, u[-1]);
+			 NULL, s[-1]);
 		    return NULL;
 		}
-	    singsub(&u);
+	    singsub(&s);
 
 	    if (!vunset && isarr) {
 		char **ap = aval;
 		char **pp = aval = (char **)ncalloc(sizeof(char *) * (arrlen(aval) + 1));
 
 		while ((*pp = *ap++)) {
-		    if (getmatch(pp, u, flags, flnum))
+		    if (getmatch(pp, s, flags, flnum))
 			pp++;
 		}
 		copied = 1;
 	    } else {
 		if (vunset)
 		    val = dupstring("");
-		getmatch(&val, u, flags, flnum);
+		getmatch(&val, s, flags, flnum);
 		copied = 1;
 	    }
 	    break;
@@ -1255,7 +1269,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		    }
 		    s = ss;
 		}
-		if (inbrace && *s != Outbrace) {
+		if (inbrace && *s) {
 		    if (*s == ':' && !imeta(s[1]))
 			zerr("unrecognized modifier `%c'", NULL, s[1]);
 		    else
@@ -1264,14 +1278,8 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		}
 	    }
 	}
-	if (inbrace) {
-	    if (*s != Outbrace) {
-	      noclosebrace:
-		zerr("closing brace expected", NULL, 0);
-		return NULL;
-	    }
-	    s++;
-	}
+	if (!inbrace)
+	    fstr = s;
     }
     if (errflag)
 	return NULL;
@@ -1363,69 +1371,73 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	int i;
 	LinkNode on = n;
 
-	if (!aval[0]) {
-	    if (plan9) {
-		uremnode(l, n);
-		*str = aptr;
-		return n;
-	    }
+	if (!aval[0] && !plan9) {
 	    if (aptr > (char *) getdata(n) &&
-		aptr[-1] == Dnull && *s == Dnull)
-		*--aptr = '\0', s++;
-	    y = (char *)ncalloc((aptr - ostr) + strlen(s) + 1);
+		aptr[-1] == Dnull && *fstr == Dnull)
+		*--aptr = '\0', fstr++;
+	    y = (char *)ncalloc((aptr - ostr) + strlen(fstr) + 1);
 	    strcpy(y, ostr);
 	    *str = y + (aptr - ostr);
-	    strcpy(*str, s);
+	    strcpy(*str, fstr);
 	    setdata(n, y);
 	    return n;
 	}
-	if (sortit && !copied)
-	    aval = arrdup(aval);
-	if (sortit == 1)
-	    qsort(aval, arrlen(aval), sizeof(char *), strpcmp);
+	if (sortit) {
+	    static CompareFn sortfn[] = {
+		strpcmp, invstrpcmp, cstrpcmp, invcstrpcmp
+	    };
 
-	else if (sortit == 2)
-	    qsort(aval, arrlen(aval), sizeof(char *), invstrpcmp);
-
-	else if (sortit == 3)
-	    qsort(aval, arrlen(aval), sizeof(char *), cstrpcmp);
-
-	else if (sortit)
-	    qsort(aval, arrlen(aval), sizeof(char *), invcstrpcmp);
+	    if (!copied)
+		aval = arrdup(aval);
 
+	    i = arrlen(aval);
+	    if (i && (*aval[i-1] || --i))
+		qsort(aval, i, sizeof(char *), sortfn[sortit-1]);
+	}
 	if (plan9) {
 	    LinkList tl = newlinklist();
+	    LinkNode tn;
 
-	    addlinknode(tl, s);
+	    *--fstr = Marker;
+	    addlinknode(tl, fstr);
 	    if (!eval && !stringsubst(tl, firstnode(tl), ssub))
 		return NULL;
 	    *str = aptr;
-	    if (empty(tl)) {
-		uremnode(l, n);
-		return n;
-	    }
-	    i = 0;
-	    while (aval[i]) {
-		LinkNode tn;
-
-		x = aval[i++];
+	    tn = firstnode(tl);
+	    while ((x = *aval++)) {
 		if (prenum || postnum)
 		    x = dopadding(x, prenum, postnum, preone, postone,
 				  premul, postmul);
 		if (eval && parsestr(x))
 		    return NULL;
 		xlen = strlen(x);
-		for (tn = firstnode(tl); tn; incnode(tn)) {
-		    strcatsub(&y, ostr, aptr, x, xlen,
-			      (char *) getdata(tn), globsubst);
+		for (tn = firstnode(tl);
+		     tn && *(y = (char *) getdata(tn)) == Marker;
+		     incnode(tn)) {
+		    strcatsub(&y, ostr, aptr, x, xlen, y + 1, globsubst);
 		    if (qt && !*y && isarr != 2)
 			y = dupstring(nulstring);
-		    if (i == 1)
-			setdata(n, (void *) y);
+		    if (plan9)
+			setdata(n, (void *) y), plan9 = 0;
 		    else
 			insertlinknode(l, n, (void *) y), incnode(n);
 		}
 	    }
+	    for (; tn; incnode(tn)) {
+		y = (char *) getdata(tn);
+		if (*y == Marker)
+		    continue;
+		if (qt && !*y && isarr != 2)
+		    y = dupstring(nulstring);
+		if (plan9)
+		    setdata(n, (void *) y), plan9 = 0;
+		else
+		    insertlinknode(l, n, (void *) y), incnode(n);
+	    }
+	    if (plan9) {
+		uremnode(l, n);
+		return n;
+	    }
 	} else {
 	    x = aval[0];
 	    if (prenum || postnum)
@@ -1450,10 +1462,11 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		    return NULL;
 		if (qt && !*x && isarr != 2)
 		    y = dupstring(nulstring);
-		else if (globsubst)
-		    tokenize(y = dupstring(x));
-		else
-		    y = x;
+		else {
+		    y = dupstring(x);
+		    if (globsubst)
+			tokenize(y);
+		}
 		insertlinknode(l, n, (void *) y), incnode(n);
 	    }
 
@@ -1464,7 +1477,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	    if (eval && parsestr(x))
 		return NULL;
 	    xlen = strlen(x);
-	    *str = strcatsub(&y, aptr, aptr, x, xlen, s, globsubst);
+	    *str = strcatsub(&y, aptr, aptr, x, xlen, fstr, globsubst);
 	    if (qt && !*y && isarr != 2)
 		y = dupstring(nulstring);
 	    insertlinknode(l, n, (void *) y), incnode(n);
@@ -1483,7 +1496,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	if (eval && parsestr(x))
 	    return NULL;
 	xlen = strlen(x);
-	*str = strcatsub(&y, ostr, aptr, x, xlen, s, globsubst);
+	*str = strcatsub(&y, ostr, aptr, x, xlen, fstr, globsubst);
 	if (qt && !*y && isarr != 2)
 	    y = dupstring(nulstring);
 	setdata(n, (void *) y);
