@@ -1,6 +1,4 @@
 /*
- * $Id: text.c,v 3.1.2.3 1997/06/01 06:13:15 hzoli Exp $
- *
  * text.c - textual representations of syntax trees
  *
  * This file is part of zsh, the Z shell.
@@ -29,7 +27,7 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "text.pro"
 
 static char *tptr, *tbuf, *tlim;
