@@ -1,6 +1,4 @@
 /*
- * $Id: utils.c,v 3.1.2.15 1997/06/01 06:13:15 hzoli Exp $
- *
  * utils.c - miscellaneous utilities
  *
  * This file is part of zsh, the Z shell.
@@ -29,7 +27,7 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "utils.pro"
 
 /* Print an error */
@@ -45,6 +43,11 @@ zwarnnam(const char *cmd, const char *fmt, const char *str, int num)
     errflag = waserr;
 }
 
+/* name of script being sourced */
+
+/**/
+char *scriptname;
+ 
 /**/
 void
 zerr(const char *fmt, const char *str, int num)
@@ -175,6 +178,8 @@ nicechar(int c)
     if (isprint(c))
 	goto done;
     if (c & 0x80) {
+	if (isset(PRINTEIGHTBIT))
+	    goto done;
 	*s++ = '\\';
 	*s++ = 'M';
 	*s++ = '-';
@@ -355,7 +360,7 @@ xsymlink(char *s)
     if (*s != '/')
 	return NULL;
     *xbuf = '\0';
-    if (xsymlinks(s + 1, 1))
+    if (!xsymlinks(s + 1, 1))
 	return ztrdup(s);
     if (!*xbuf)
 	return ztrdup("/");
@@ -399,6 +404,11 @@ fprintdir(char *s, FILE *f)
  * and uid to try to avoid requerying the password files *
  * or NIS/NIS+ database.                                 */
 
+/**/
+uid_t cached_uid;
+/**/
+char *cached_username;
+
 /**/
 char *
 get_username(void)
@@ -575,12 +585,28 @@ dircmp(char *s, char *t)
     return 1;
 }
 
+/* extra functions to call before displaying the prompt */
+
+/**/
+LinkList prepromptfns;
+
+/* the last time we checked mail */
+ 
+/**/
+time_t lastmailcheck;
+ 
+/* the last time we checked the people in the WATCH variable */
+ 
+/**/
+time_t lastwatch;
+
 /* do pre-prompt stuff */
 
 /**/
 void
 preprompt(void)
 {
+    static time_t lastperiodic;
     LinkNode ln;
     List list;
     int period = getiparam("PERIOD");
@@ -786,6 +812,23 @@ settyinfo(struct ttyinfo *ti)
     }
 }
 
+/* the default tty state */
+ 
+/**/
+struct ttyinfo shttyinfo;
+
+/* != 0 if we need to call resetvideo() */
+
+/**/
+int resetneeded;
+
+#ifdef TIOCGWINSZ
+/* window size changed */
+
+/**/
+int winchanged;
+#endif
+ 
 /* check the size of the window and adjust if necessary */
 
 /**/
@@ -1035,22 +1078,28 @@ zstrtol(const char *s, char **t, int base)
 int
 checkrmall(char *s)
 {
-    fflush(stdin);
     fprintf(shout, "zsh: sure you want to delete all the files in ");
     if (*s != '/') {
 	nicezputs(pwd[1] ? unmeta(pwd) : "", shout);
 	fputc('/', shout);
     }
     nicezputs(s, shout);
+    if(isset(RMSTARWAIT)) {
+	fputs("? (waiting ten seconds)", shout);
+	fflush(shout);
+	beep();
+	sleep(10);
+	fputc('\n', shout);
+    }
     fputs(" [yn]? ", shout);
     fflush(shout);
     beep();
-    return (getquery("ny") == 'y');
+    return (getquery("ny", 1) == 'y');
 }
 
 /**/
 int
-getquery(char *valid_chars)
+getquery(char *valid_chars, int purge)
 {
     char c, d;
     int isem = !strcmp(term, "emacs");
@@ -1065,7 +1114,10 @@ getquery(char *valid_chars)
 
 #ifdef FIONREAD
     ioctl(SHTTY, FIONREAD, (char *)&val);
-    if (val) {
+    if(purge) {
+	while(val--)
+	    read(SHTTY, &c, 1);
+    } else if (val) {
 	if (!isem)
 	    settyinfo(&shttyinfo);
 	write(SHTTY, "n\n", 2);
@@ -1229,15 +1281,12 @@ spckword(char **s, int hist, int cmd, int ask)
 	}
 	if (ask) {
 	    char *pptbuf;
-	    int pptlen;
-	    rstring = best;
-	    Rstring = guess;
-	    pptbuf = putprompt(sprompt, &pptlen, NULL, 1);
-	    fwrite(pptbuf, pptlen, 1, shout);
+	    pptbuf = promptexpand(sprompt, 0, best, guess);
+	    zputs(pptbuf, shout);
 	    free(pptbuf);
 	    fflush(shout);
 	    beep();
-	    x = getquery("nyae ");
+	    x = getquery("nyae ", 0);
 	} else
 	    x = 'y';
 	if (x == 'y' || x == ' ') {
@@ -1258,6 +1307,7 @@ spckword(char **s, int hist, int cmd, int ask)
 int
 ztrftime(char *buf, int bufsize, char *fmt, struct tm *tm)
 {
+    int hr12;
 #ifndef HAVE_STRFTIME
     static char *astr[] =
     {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
@@ -1267,9 +1317,6 @@ ztrftime(char *buf, int bufsize, char *fmt, struct tm *tm)
 #else
     char *origbuf = buf;
 #endif
-    static char *lstr[] =
-    {"12", " 1", " 2", " 3", " 4", " 5", " 6", " 7", " 8", " 9",
-     "10", "11"};
     char tmp[3];
 
 
@@ -1300,7 +1347,15 @@ ztrftime(char *buf, int bufsize, char *fmt, struct tm *tm)
 		*buf++ = '0' + tm->tm_hour % 10;
 		break;
 	    case 'l':
-		strucpy(&buf, lstr[tm->tm_hour % 12]);
+	    case 'L':
+		hr12 = tm->tm_hour % 12;
+		if (hr12 == 0)
+		    hr12 = 12;
+	        if (hr12 > 9)
+		  *buf++ = '1';
+		else if (fmt[-1] == 'l')
+		  *buf++ = ' ';
+		*buf++ = '0' + (hr12 % 10);
 		break;
 	    case 'm':
 		*buf++ = '0' + (tm->tm_mon + 1) / 10;
@@ -1639,8 +1694,7 @@ getshfunc(char *nam)
 
 /* allocate a tree element */
 
-static int sizetab[N_COUNT] =
-{
+static int sizetab[N_COUNT] = {
     sizeof(struct list),
     sizeof(struct sublist),
     sizeof(struct pline),
@@ -1655,20 +1709,34 @@ static int sizetab[N_COUNT] =
     sizeof(struct autofn),
 };
 
-static int flagtab[N_COUNT] =
-{
-    NT_SET(N_LIST, 1, NT_NODE, NT_NODE, 0, 0),
-    NT_SET(N_SUBLIST, 2, NT_NODE, NT_NODE, 0, 0),
-    NT_SET(N_PLINE, 1, NT_NODE, NT_NODE, 0, 0),
-    NT_SET(N_CMD, 3, NT_NODE, NT_STR | NT_LIST, NT_NODE | NT_LIST, NT_NODE | NT_LIST),
-    NT_SET(N_REDIR, 3, NT_STR, 0, 0, 0),
-    NT_SET(N_COND, 1, NT_NODE, NT_NODE, 0, 0),
-    NT_SET(N_FOR, 1, NT_STR, NT_STR, NT_STR, NT_NODE),
-    NT_SET(N_CASE, 0, NT_STR | NT_ARR, NT_NODE | NT_ARR, 0, 0),
-    NT_SET(N_IF, 0, NT_NODE | NT_ARR, NT_NODE | NT_ARR, 0, 0),
-    NT_SET(N_WHILE, 1, NT_NODE, NT_NODE, 0, 0),
-    NT_SET(N_VARASG, 1, NT_STR, NT_STR, NT_STR | NT_LIST, 0),
-    NT_SET(N_AUTOFN, 1, 0, 0, 0, 0),
+static int offstab[N_COUNT] = {
+    offsetof(struct list, left),
+    offsetof(struct sublist, left),
+    offsetof(struct pline, left),
+    offsetof(struct cmd, u),
+    offsetof(struct redir, name),
+    offsetof(struct cond, left),
+    offsetof(struct forcmd, name),
+    offsetof(struct casecmd, pats),
+    offsetof(struct ifcmd, ifls),
+    offsetof(struct whilecmd, cont),
+    offsetof(struct varasg, name),
+    sizeof(struct autofn),
+};
+
+static int flagtab[N_COUNT] = {
+    NT_SET(N_LIST, NT_NODE, NT_NODE, 0, 0),
+    NT_SET(N_SUBLIST, NT_NODE, NT_NODE, 0, 0),
+    NT_SET(N_PLINE, NT_NODE, NT_NODE, 0, 0),
+    NT_SET(N_CMD, NT_NODE, NT_STR | NT_LIST, NT_NODE | NT_LIST, NT_NODE | NT_LIST),
+    NT_SET(N_REDIR, NT_STR, 0, 0, 0),
+    NT_SET(N_COND, NT_NODE, NT_NODE, 0, 0),
+    NT_SET(N_FOR, NT_STR, NT_STR, NT_STR, NT_NODE),
+    NT_SET(N_CASE, NT_STR | NT_ARR, NT_NODE | NT_ARR, 0, 0),
+    NT_SET(N_IF, NT_NODE | NT_ARR, NT_NODE | NT_ARR, 0, 0),
+    NT_SET(N_WHILE, NT_NODE, NT_NODE, 0, 0),
+    NT_SET(N_VARASG, NT_STR, NT_STR, NT_STR | NT_LIST, 0),
+    NT_SET(N_AUTOFN, 0, 0, 0, 0),
 };
 
 /**/
@@ -1998,25 +2066,18 @@ expandstruct(struct node *n, int exp)
 static void *
 dupstruct2(void *a)
 {
-    int *oints, *nints;
     void **onodes, **nnodes, *ret, *n, *on;
-    int type, heap, i;
+    int type, heap;
+    size_t nodeoffs;
 
     if (!a || ((List) a) == &dummy_list)
 	return a;
-    oints = (int *) a;
-    type = *oints;
-    ret = nints = (int *) alloc(sizetab[NT_TYPE(type)]);
-    *nints = (type & ~NT_HEAP);
-    if (useheap)
-	*nints |= NT_HEAP;
-    i = NT_NUM(type);
-    while ((i + 1) & (sizeof(void *)/sizeof(int) - 1))
-	i++;
-    while (i--)
-	*++nints = *++oints;
-    onodes = (void **) (oints + 1);
-    nnodes = (void **) (nints + 1);
+    type = *(int *)a;
+    ret = alloc(sizetab[NT_TYPE(type)]);
+    memcpy(ret, a, nodeoffs = offstab[NT_TYPE(type)]);
+    *(int*)ret = (type & ~NT_HEAP) | (useheap ? NT_HEAP : 0);
+    onodes = (void **) ((char *)a + nodeoffs);
+    nnodes = (void **) ((char *)ret + nodeoffs);
     heap = type & NT_HEAP;
     for (type = (type & 0xffff00) >> 4; (type >>= 4); *nnodes++ = n) {
 	if (!(on = *onodes++))
@@ -2079,10 +2140,7 @@ freestruct(void *a)
 	return;
 
     type = * (int *) a;
-    size = NT_NUM(type);
-    while ((size + 1) & (sizeof(void *)/sizeof(int) - 1))
-	size++;
-    nodes = (void **) ((int *) a + size + 1);
+    nodes = (void **) ((char *)a + offstab[NT_TYPE(type)]);
     size = sizetab[NT_TYPE(type)];
     for (type = (type & 0xffff00) >> 4; (type >>= 4);) {
 	if ((n = *nodes++)) {
@@ -2238,6 +2296,11 @@ simplifyright(List l)
     return;
 }
 
+/* the ztypes table */
+
+/**/
+short int typtab[256];
+
 /* initialize the ztypes table */
 
 /**/
@@ -2264,7 +2327,8 @@ inittyptab(void)
     typtab['\t'] |= IBLANK | INBLANK;
     typtab['\n'] |= INBLANK;
     typtab['\0'] |= IMETA;
-    typtab[STOUC(Meta)] |= IMETA;
+    typtab[STOUC(Meta)  ] |= IMETA;
+    typtab[STOUC(Marker)] |= IMETA;
     for (t0 = (int)STOUC(Pound); t0 <= (int)STOUC(Nularg); t0++)
 	typtab[t0] |= ITOK | IMETA;
     for (s = ifs ? ifs : DEFAULT_IFS; *s; s++) {
@@ -2865,7 +2929,7 @@ zreaddir(DIR *dir, int ignoredots)
     return metafy(de->d_name, -1, META_STATIC);
 }
 
-/* Unmetafy and output a string. */
+/* Unmetafy and output a string.  Tokens are skipped. */
 
 /**/
 int
@@ -2876,7 +2940,10 @@ zputs(char const *s, FILE *stream)
     while (*s) {
 	if (*s == Meta)
 	    c = *++s ^ 32;
-	else
+	else if(itok(*s)) {
+	    s++;
+	    continue;
+	} else
 	    c = *s;
 	s++;
 	if (fputc(c, stream) < 0)
@@ -3565,4 +3632,38 @@ dputs(char *message)
     fprintf(stderr, "%s\n", message);
     fflush(stderr);
 }
-#endif
+
+#endif /* DEBUG */
+
+/**/
+int
+mode_to_octal(mode_t mode)
+{
+    int m = 0;
+
+    if(mode & S_ISUID)
+	m |= 04000;
+    if(mode & S_ISGID)
+	m |= 02000;
+    if(mode & S_ISVTX)
+	m |= 01000;
+    if(mode & S_IRUSR)
+	m |= 00400;
+    if(mode & S_IWUSR)
+	m |= 00200;
+    if(mode & S_IXUSR)
+	m |= 00100;
+    if(mode & S_IRGRP)
+	m |= 00040;
+    if(mode & S_IWGRP)
+	m |= 00020;
+    if(mode & S_IXGRP)
+	m |= 00010;
+    if(mode & S_IROTH)
+	m |= 00004;
+    if(mode & S_IWOTH)
+	m |= 00002;
+    if(mode & S_IXOTH)
+	m |= 00001;
+    return m;
+}
