@@ -1,6 +1,4 @@
 /*
- * $Id: watch.c,v 3.1.2.2 1997/06/01 06:13:15 hzoli Exp $
- *
  * watch.c - login/logout watching
  *
  * This file is part of zsh, the Z shell.
@@ -29,11 +27,11 @@
  *
  */
 
-#include "zsh.h"
+#include "zsh.mdh"
 #include "watch.pro"
 
 static int wtabsz;
-STRUCT_UTMP *wtab;
+static STRUCT_UTMP *wtab;
 static time_t lastutmpcheck;
 
 /* get the time of login/logout for WATCH */
