@@ -1,6 +1,4 @@
 /*
- * $Id: zsh.h,v 3.1.2.9 1997/05/06 05:03:03 hzoli Exp $
- *
  * zsh.h - standard header file
  *
  * This file is part of zsh, the Z shell.
@@ -29,26 +27,13 @@
  *
  */
 
-#ifndef _ZSH_H
-#define _ZSH_H
+#define trashzle()      trashzleptr()
+#define zleread(X,Y,H)  zlereadptr(X,Y,H)
+#define spaceinline(X)  spaceinlineptr(X)
+#define gotword()       gotwordptr()
+#define refresh()       refreshptr()
 
-/* We use <config.h> instead of "config.h" so that a compilation        *
- * using -I. -I$srcdir will use ./config.h rather than $srcdir/config.h *
- * (which it would do because it found this file in $srcdir).           */
-#include <config.h>
-#include <system.h>
-
-#ifndef IN_ZLE
-# define trashzle()      trashzleptr()
-# define zleread(X,Y,H)  zlereadptr(X,Y,H)
-# define spaceinline(X)  spaceinlineptr(X)
-# define gotword()       gotwordptr()
-# define refresh()       refreshptr()
-#endif /* !IN_ZLE */
-
-#ifndef IN_COMP
-# define compctlread(N,A,O,R) compctlreadptr(N,A,O,R)
-#endif /* !IN_COMP */
+#define compctlread(N,A,O,R) compctlreadptr(N,A,O,R)
 
 /* A few typical macros */
 #define minimum(a,b)  ((a) < (b) ? (a) : (b))
@@ -69,6 +54,12 @@ typedef int LV;
 # define STOUC(X)	((unsigned char)(X))
 #endif
 
+/* Meta together with the character following Meta denotes the character *
+ * which is the exclusive or of 32 and the character following Meta.     *
+ * This is used to represent characters which otherwise has special      *
+ * meaning for zsh.  These are the characters for which the imeta() test *
+ * is true: the null character, and the characters from Meta to Marker.  */
+
 #define Meta		((char) 0x83)
 
 /* Note that the fourth character in DEFAULT_IFS is Meta *
@@ -104,6 +95,9 @@ typedef int LV;
 
 #define INULL(x)	(((x) & 0xfc) == 0x98)
 
+/* Marker used in paramsubst for rc_expand_param */
+#define Marker		((char) 0x9c)
+
 /* chars that need to be quoted if meant literally */
 
 #define SPECCHARS "#$^*()=|{}[]`<>?~;&\n\t \\\'\""
@@ -176,7 +170,7 @@ enum {
 };
 
 /* Redirection types.  If you modify this, you may also have to modify *
- * redirtab in globals.h and getredirs() in text.c and the IS_* macros *
+ * redirtab in parse.c and getredirs() in text.c and the IS_* macros   *
  * below.                                                              */
 
 enum {
@@ -337,10 +331,8 @@ struct node {
 
 #define NT_TYPE(T) ((T) & 0xff)
 #define NT_N(T, N) (((T) >> (8 + (N) * 4)) & 0xf)
-#define NT_SET(T0, N, T1, T2, T3, T4) \
-    ((T0) | ((N) << 24) | \
-     ((T1) << 8) | ((T2) << 12) | ((T3) << 16) | ((T4) << 20))
-#define NT_NUM(T) (((T) >> 24) & 7)
+#define NT_SET(T0, T1, T2, T3, T4) \
+    ((T0) | ((T1) << 8) | ((T2) << 12) | ((T3) << 16) | ((T4) << 20))
 #define NT_HEAP   (1 << 30)
 
 /* tree element for lists */
@@ -592,6 +584,8 @@ struct job {
 #define STAT_SUPERJOB	(1<<7)	/* job has a subjob                     */
 #define STAT_SUBJOB	(1<<8)	/* job is a subjob                      */
 #define STAT_CURSH	(1<<9)	/* last command is in current shell     */
+#define STAT_NOSTTY	(1<<10)	/* the tty settings are not inherited   */
+				/* from this job when it exits.         */
 
 #define SP_RUNNING -1		/* fake status for jobs currently running */
 
@@ -758,6 +752,7 @@ struct shfunc {
 /* node in builtin command hash table (builtintab) */
 
 typedef int (*HandlerFunc) _((char *, char **, char *, int));
+#define NULLBINCMD ((HandlerFunc) 0)
 
 struct builtin {
     HashNode next;		/* next in hash chain                                 */
@@ -771,11 +766,17 @@ struct builtin {
     char *defopts;		/* options set by default for overloaded handlerfuncs */
 };
 
+#define BUILTIN(name, flags, handler, min, max, funcid, optstr, defopts) \
+    { NULL, name, flags, handler, min, max, funcid, optstr, defopts }
+#define BIN_PREFIX(name, flags) \
+    BUILTIN(name, flags | BINF_PREFIX, NULLBINCMD, 0, 0, 0, NULL, NULL)
+
 /* builtin flags */
 /* DISABLE IS DEFINED AS (1<<0) */
 #define BINF_PLUSOPTS		(1<<1)	/* +xyz legal */
 #define BINF_R			(1<<2)	/* this is the builtin `r' (fc -e -) */
 #define BINF_PRINTOPTS		(1<<3)
+#define BINF_ADDED		(1<<4)	/* is in the builtins hash table */
 #define BINF_FCOPTS		(1<<5)
 #define BINF_TYPEOPT		(1<<6)
 #define BINF_ECHOPTS		(1<<7)
@@ -969,6 +970,7 @@ enum {
     AUTORESUME,
     BADPATTERN,
     BANGHIST,
+    BAREGLOBQUAL,
     BEEP,
     BGNICE,
     BRACECCL,
@@ -1003,6 +1005,7 @@ enum {
     HISTBEEP,
     HISTIGNOREDUPS,
     HISTIGNORESPACE,
+    HISTNOFUNCTIONS,
     HISTNOSTORE,
     HISTREDUCEBLANKS,
     HISTVERIFY,
@@ -1013,6 +1016,7 @@ enum {
     INTERACTIVECOMMENTS,
     KSHARRAYS,
     KSHAUTOLOAD,
+    KSHGLOB,
     KSHOPTIONPRINT,
     LISTAMBIGUOUS,
     LISTBEEP,
@@ -1033,9 +1037,12 @@ enum {
     OVERSTRIKE,
     PATHDIRS,
     POSIXBUILTINS,
+    PRINTEIGHTBIT,
     PRINTEXITVALUE,
     PRIVILEGED,
+    PROMPTBANG,
     PROMPTCR,
+    PROMPTPERCENT,
     PROMPTSUBST,
     PUSHDIGNOREDUPS,
     PUSHDMINUS,
@@ -1047,6 +1054,7 @@ enum {
     RECEXACT,
     RESTRICTED,
     RMSTARSILENT,
+    RMSTARWAIT,
     SHFILEEXPANSION,
     SHGLOB,
     SHINSTDIN,
@@ -1169,8 +1177,6 @@ struct ttyinfo {
 /* Definitions for the %_ prompt escape */
 /****************************************/
 
-#include "ztype.h"
-
 #define cmdpush(X) if (!(cmdsp >= 0 && cmdsp < 256)) {;} else cmdstack[cmdsp++]=(X)
 #ifdef DEBUG
 # define cmdpop()  if (cmdsp <= 0) { \
@@ -1271,30 +1277,16 @@ struct ttyinfo {
 #define ZSIG_IGNORED	(1<<1)
 #define ZSIG_FUNC	(1<<2)
 
-/*********************************/
-/* Dynamic builtin table munging */
-/*********************************/
+/****************/
+/* Entry points */
+/****************/
 
-typedef struct binlist *BinList;
+/* compctl entry point pointers */
 
-#define BINL_ADDED (1<<0)
+typedef int (*CompctlReadFn) _((char *, char **, char *, char *));
 
-struct binlist {
-    char *nam;
-    int binf;
-    HandlerFunc func;
-    int minargs, maxargs, funcid;
-    char *optstr, *defopts;
-    int flags;
-};
+/* ZLE entry point pointers */
 
-/***********************/
-/* Shared header files */
-/***********************/
-
-#include "signals.h"
-#include "prototypes.h"
-#include "globals.h"
-#include "hashtable.h"
-
-#endif /*!_ZSH_H*/
+typedef void (*ZleVoidFn) _((void));
+typedef void (*ZleVoidIntFn) _((int));
+typedef unsigned char * (*ZleReadFn) _((char *, char *, int));
